/* ========================================================================
* PROJECT: ARToolKitPlus
* ========================================================================
* This work is based on the original ARToolKit developed by
*   Hirokazu Kato
*   Mark Billinghurst
*   HITLab, University of Washington, Seattle
* http://www.hitl.washington.edu/artoolkit/
*
* Copyright of the derived and new portions of this work
*     (C) 2006 Graz University of Technology
*
* This framework is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This framework is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this framework; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* For further information please contact 
*   Dieter Schmalstieg
*   <schmalstieg@icg.tu-graz.ac.at>
*   Graz University of Technology, 
*   Institut for Computer Graphics and Vision,
*   Inffeldgasse 16a, 8010 Graz, Austria.
* ========================================================================
** @author   Daniel Wagner
*
* $Id: TrackerImpl.h 172 2006-07-25 14:05:47Z daniel $
* @file
* ======================================================================== */


#ifndef __ARTOOLKIT_TRACKERIMPL_HEADERFILE__
#define __ARTOOLKIT_TRACKERIMPL_HEADERFILE__


#include <ARToolKitPlus/ar.h>
#include <ARToolKitPlus/arMulti.h>
#include <ARToolKitPlus/matrix.h>
#include <ARToolKitPlus/Tracker.h>
#include <ARToolKitPlus/MemoryManager.h>
#include <ARToolKitPlus/Camera.h>
#include <ARToolKitPlus/CameraFactory.h>
#include <ARToolKitPlus/extra/BCH.h>


#if defined(_MSC_VER)
#  if _MSC_VER<1300
#    define _OLD_MSCOMPILER_
#    pragma message (">>> Compiling for old MS Compiler")
#  endif
#endif


#define AR_TEMPL_FUNC template <int __PATTERN_SIZE_X, int __PATTERN_SIZE_Y, int __PATTERN_SAMPLE_NUM, int __MAX_LOAD_PATTERNS, int __MAX_IMAGE_PATTERNS>
#define AR_TEMPL_TRACKER TrackerImpl<__PATTERN_SIZE_X, __PATTERN_SIZE_Y, __PATTERN_SAMPLE_NUM, __MAX_LOAD_PATTERNS, __MAX_IMAGE_PATTERNS>


namespace ARToolKitPlus {


// Compile time information...
//
static bool usesSinglePrecision();				/// Returns whether single or double precision is used
//const char* getDescriptionString();		/// Returns a short description string about compile time settings

#ifndef _ARTKP_NO_MEMORYMANAGER_
extern MemoryManager* memManager;
#endif //_ARTKP_NO_MEMORYMANAGER_

void artkp_Free(void* nRawMemory);

template<class T> T* artkp_Alloc(size_t size)
{
#ifndef _ARTKP_NO_MEMORYMANAGER_
	if(memManager)
		return (T*)memManager->getMemory(size*sizeof(T));
	else
#endif //_ARTKP_NO_MEMORYMANAGER_
		return (T*)::malloc(size*sizeof(T));
}


/// TrackerImpl implements the Tracker interface
template <int __PATTERN_SIZE_X, int __PATTERN_SIZE_Y, int __PATTERN_SAMPLE_NUM, int __MAX_LOAD_PATTERNS, int __MAX_IMAGE_PATTERNS>
class TrackerImpl : public Tracker
{
public:
	enum {
		PATTERN_WIDTH = __PATTERN_SIZE_X,
		PATTERN_HEIGHT = __PATTERN_SIZE_Y,
		PATTERN_SAMPLE_NUM = __PATTERN_SAMPLE_NUM,

		MAX_LOAD_PATTERNS = __MAX_LOAD_PATTERNS,
		MAX_IMAGE_PATTERNS = __MAX_IMAGE_PATTERNS,
		WORK_SIZE = 1024*MAX_IMAGE_PATTERNS,

#ifdef SMALL_LUM8_TABLE
		LUM_TABLE_SIZE = (0xffff >> 6) + 1,
#else
		LUM_TABLE_SIZE = 0xffff + 1,
#endif
	};


	TrackerImpl();
	virtual ~TrackerImpl();

	/// does final clean up (memory deallocation)
	virtual void cleanup();


	/// Sets the pixel format of the camera image
	/**
	 *  Default format is RGB888 (PIXEL_FORMAT_RGB)
	 */
	virtual bool setPixelFormat(PIXEL_FORMAT nFormat);

	/// Loads a camera calibration file and stores data internally
	/**
	 *  To prevent memory leaks, this method internally deletes an existing camera.
	 *  If you want to use more than one camera, retrieve the existing camera using getCamera()
	 *  and call setCamera(NULL); before loading another camera file.
	 *  On destruction, ARToolKitPlus will only destroy the currently set camera. All other
	 *  cameras have to be destroyed manually.
	 */
	virtual bool loadCameraFile(const char* nCamParamFile, ARFloat nNearClip, ARFloat nFarClip);

	virtual void setLoadUndistLUT(bool nSet)  {  loadCachedUndist = nSet;  }

	/// sets an instance which implements the ARToolKit::Logger interface
	virtual void setLogger(ARToolKitPlus::Logger* nLogger)  {  logger = nLogger;  }

	/// marker detection using tracking history
	virtual int arDetectMarker(ARUint8 *dataPtr, int thresh, ARMarkerInfo **marker_info, int *marker_num);

	/// marker detection without using tracking history
	virtual int arDetectMarkerLite(ARUint8 *dataPtr, int thresh, ARMarkerInfo **marker_info, int *marker_num);

	/// calculates the transformation matrix between camera and the given multi-marker config
	virtual ARFloat arMultiGetTransMat(ARMarkerInfo *marker_info, int marker_num, ARMultiMarkerInfoT *config);

	/// calculates the transformation matrix between camera and the given marker
	virtual ARFloat arGetTransMat(ARMarkerInfo *marker_info, ARFloat center[2], ARFloat width, ARFloat conv[3][4]);

	virtual ARFloat arGetTransMatCont(ARMarkerInfo *marker_info, ARFloat prev_conv[3][4], ARFloat center[2], ARFloat width, ARFloat conv[3][4]);

	// RPP integration -- [t.pintaric]
	virtual ARFloat rppMultiGetTransMat(ARMarkerInfo *marker_info, int marker_num, ARMultiMarkerInfoT *config);
	virtual ARFloat rppGetTransMat(ARMarkerInfo *marker_info, ARFloat center[2], ARFloat width, ARFloat conv[3][4]);

	/// loads a pattern from a file
	virtual int arLoadPatt(char *filename);

	/// frees a pattern from memory
	virtual int arFreePatt(int patno);

	virtual int arMultiFreeConfig( ARMultiMarkerInfoT *config );

	virtual ARMultiMarkerInfoT *arMultiReadConfigFile(const char *filename);

	virtual void activateBinaryMarker(int nThreshold)  {  binaryMarkerThreshold = nThreshold;  }

	/// Activate the usage of id-based markers rather than template based markers
	/**
	 *  id-based markers directly encode the marker id in the image.
	 *  see arBitFieldPattern.h for more information
	 */
	virtual void setMarkerMode(MARKER_MODE nMarkerMode);


	/// activates the complensation of brightness falloff in the corners of the camera image
	/**
	 *  some cameras have a falloff in brightness at the border of the image, which creates
	 *  problems with thresholding the image. use this function to set a (linear) adapted
	 *  threshold value. the threshold value will stay exactly the same at the center but
	 *  will deviate near to the border. all values specify a difference, not absolute values!
	 *  nCorners define the falloff a all four corners. nLeftRight defines the falloff
	 *  at the half y-position at the left and right side of the image. nTopBottom defines the falloff
	 *  at the half x-position at the top and bottom side of the image.
	 *  all values between these 9 points (center, 4 corners, left, right, top, bottom) will
	 *  be interpolated.
	 */
	virtual void activateVignettingCompensation(bool nEnable, int nCorners=0, int nLeftRight=0, int nTopBottom=0);


	/// Calculates the camera matrix from an ARToolKit camera file.
	/**
	 * This method retrieves the OpenGL projection matrix that is stored
	 * in an ARToolKit camera calibration file.
	 * Returns true if loading of the camera file succeeded.
	 */
	bool calcCameraMatrix(const char* nCamParamFile, int nWidth, int nHeight,
								 ARFloat nNear, ARFloat nFar, ARFloat *nMatrix);

	
	/// Changes the resolution of the camera after the camerafile was already loaded
	virtual void changeCameraSize(int nWidth, int nHeight);


	/// Changes the undistortion mode
	/**
	 * Default value is UNDIST_STD which means that
	 * artoolkit's standard undistortion method is used.
	 */
	virtual void setUndistortionMode(UNDIST_MODE nMode);

	/// Changes the Pose Estimation Algorithm
	/**
	* POSE_ESTIMATOR_ORIGINAL (default): arGetTransMat()
	* POSE_ESTIMATOR_RPP: "Robust Pose Estimation from a Planar Target"
	*/
	virtual bool setPoseEstimator(POSE_ESTIMATOR nMethod);

	/// Sets a new relative border width. ARToolKit's default value is 0.25
	/**
	 * Take caution that the markers need of course really have thiner borders.
	 * Values other than 0.25 have not been tested for regular pattern-based matching,
	 * but only for id-encoded markers. It might be that the pattern creation process
	 * needs to be updated too.
	 */
	virtual void setBorderWidth(ARFloat nFraction)  {  relBorderWidth = nFraction;  }


	/// Sets the threshold value that is used for black/white conversion
	virtual void setThreshold(int nValue)  {  thresh = nValue;  }


	/// Returns the current threshold value.
	virtual int getThreshold() const  {  return thresh;  }


	/// Turns automatic threshold calculation on/off
	virtual void activateAutoThreshold(bool nEnable)  {  autoThreshold.enable = nEnable;  }


	/// Returns true if automatic threshold detection is enabled
	virtual bool isAutoThresholdActivated() const  {  return autoThreshold.enable;  }

	/// Sets the number of times the threshold is randomized in case no marker was visible (Minimum: 1, Default: 2)
	/**
	 *  Autothreshold requires a visible marker to estime the optimal thresholding value. If
	 *  no marker is visible ARToolKitPlus randomizes the thresholding value until a marker is
	 *  found. This function sets the number of times ARToolKitPlus will randomize the threshold
	 *  value and research for a marker per calc() invokation until it gives up.
	 *  A value of 2 means that ARToolKitPlus will analyze the image a second time with an other treshold value
	 *  if it does not find a marker the first time. Each unsuccessful try uses less processing power
	 *  than a single full successful position estimation.
	 */
	virtual void setNumAutoThresholdRetries(int nNumRetries)  {  autoThreshold.numRandomRetries = nNumRetries>=1 ? nNumRetries : 1;  }


	/// Sets an image processing mode (half or full resolution)
	/**
	 *  Half resolution is faster but less accurate. When using
	 *  full resolution smaller markers will be detected at a
	 *  higher accuracy (or even detected at all).
	 */
	virtual void setImageProcessingMode(IMAGE_PROC_MODE nMode)  {  arImageProcMode = (nMode==IMAGE_HALF_RES ? AR_IMAGE_PROC_IN_HALF : AR_IMAGE_PROC_IN_FULL);  }


	/// Returns an opengl-style modelview transformation matrix
	virtual const ARFloat* getModelViewMatrix() const  {  return gl_para;  }


	/// Returns an opengl-style projection transformation matrix
	virtual const ARFloat* getProjectionMatrix() const  {  return gl_cpara;  }


	/// Returns a short description with compiled-in settings
	virtual const char* getDescription();


	/// Returns the compiled-in pixel format
	virtual PIXEL_FORMAT getPixelFormat() const  {  return static_cast<PIXEL_FORMAT>(pixelFormat);  }


	/// Returns the numbber of bits per pixel for the compiled-in pixel format
	virtual int getBitsPerPixel() const  {  return pixelSize*8;  }


	/// Returns the maximum number of patterns that can be loaded
	virtual int getNumLoadablePatterns() const  {  return MAX_LOAD_PATTERNS;  }


	/// Returns the current camera
	virtual Camera* getCamera()  {  return arCamera;  }


	/// Sets a new camera without specifying new near and far clip values
	virtual void setCamera(Camera* nCamera);


	/// Sets a new camera including specifying new near and far clip values
	virtual void setCamera(Camera* nCamera, ARFloat nNearClip, ARFloat nFarClip);


	virtual ARFloat calcOpenGLMatrixFromMarker(ARMarkerInfo* nMarkerInfo, ARFloat nPatternCenter[2], ARFloat nPatternSize, ARFloat *nOpenGLMatrix);


	virtual ARFloat executeSingleMarkerPoseEstimator(ARMarkerInfo *marker_info, ARFloat center[2], ARFloat width, ARFloat conv[3][4]);


	virtual ARFloat executeMultiMarkerPoseEstimator(ARMarkerInfo *marker_info, int marker_num, ARMultiMarkerInfoT *config);



protected:
	bool checkPixelFormat();

	void checkImageBuffer();

	//static int arParamChangeSize( ARParam *source, int xsize, int ysize, ARParam *newparam );

	//static int arParamSave( char *filename, int num, ARParam *param, ...);
	//static int arParamLoad( char *filename, int num, ARParam *param, ...);

	// converts an ARToolKit transformation matrix for usage with OpenGL
	void convertTransformationMatrixToOpenGLStyle(ARFloat para[3][4], ARFloat gl_para[16]);

	// converts an ARToolKit projection matrix for usage with OpenGL
	static bool convertProjectionMatrixToOpenGLStyle(ARParam *param, ARFloat gnear, ARFloat gfar, ARFloat m[16]);
	static bool convertProjectionMatrixToOpenGLStyle2(ARFloat cparam[3][4], int width, int height, ARFloat gnear, ARFloat gfar, ARFloat m[16]);


	ARMarkerInfo2* arDetectMarker2(ARInt16 *limage, int label_num, int *label_ref,
								   int *warea, ARFloat *wpos, int *wclip,
								   int area_max, int area_min, ARFloat factor, int *marker_num);

	int arGetContour(ARInt16 *limage, int *label_ref, int label, int clip[4], ARMarkerInfo2 *marker_infoTWO);

	int check_square(int area, ARMarkerInfo2 *marker_infoTWO, ARFloat factor);

	int arGetCode(ARUint8 *image, int *x_coord, int *y_coord, int *vertex,
				  int *code, int *dir, ARFloat *cf, int thresh);

	int arGetPatt(ARUint8 *image, int *x_coord, int *y_coord, int *vertex,
				  ARUint8 ext_pat[PATTERN_HEIGHT][PATTERN_WIDTH][3]);

	int pattern_match( ARUint8 *data, int *code, int *dir, ARFloat *cf);

	int downsamplePattern(ARUint8* data, unsigned char* imgPtr);

	int bitfield_check_simple(ARUint8 *data, int *code, int *dir, ARFloat *cf, int thresh);

	int bitfield_check_BCH(ARUint8 *data, int *code, int *dir, ARFloat *cf, int thresh);

	void gen_evec(void);

	ARMarkerInfo* arGetMarkerInfo(ARUint8 *image, ARMarkerInfo2 *marker_info2, int *marker_num, int thresh);

	ARFloat arGetTransMat2(ARFloat rot[3][3], ARFloat ppos2d[][2], ARFloat ppos3d[][2], int num, ARFloat conv[3][4]);


	ARFloat arGetTransMat4(ARFloat rot[3][3], ARFloat ppos2d[][2], ARFloat ppos3d[][3], int num, ARFloat conv[3][4]);

	ARFloat arGetTransMat5(ARFloat rot[3][3], ARFloat ppos2d[][2],
						   ARFloat ppos3d[][3], int num, ARFloat conv[3][4],
						   Camera *pCam);
						   //ARFloat *dist_factor, ARFloat cpara[3][4]);

	ARFloat arGetTransMatSub(ARFloat rot[3][3], ARFloat ppos2d[][2],
							 ARFloat pos3d[][3], int num, ARFloat conv[3][4],
							 Camera *pCam);
							 //ARFloat *dist_factor, ARFloat cpara[3][4] );

	ARFloat arModifyMatrix(ARFloat rot[3][3], ARFloat trans[3], ARFloat cpara[3][4],
								  ARFloat vertex[][3], ARFloat pos2d[][2], int num);

	ARFloat arModifyMatrix2(ARFloat rot[3][3], ARFloat trans[3], ARFloat cpara[3][4],
								   ARFloat vertex[][3], ARFloat pos2d[][2], int num);

	int arGetAngle(ARFloat rot[3][3], ARFloat *wa, ARFloat *wb, ARFloat *wc);

	int arGetRot(ARFloat a, ARFloat b, ARFloat c, ARFloat rot[3][3]);

	int arGetNewMatrix(ARFloat a, ARFloat b, ARFloat c,
							  ARFloat trans[3], ARFloat trans2[3][4],
							  ARFloat cpara[3][4], ARFloat ret[3][4]);

	int arGetInitRot(ARMarkerInfo *marker_info, ARFloat cpara[3][4], ARFloat rot[3][3]);


	ARFloat arGetTransMatCont2(ARMarkerInfo *marker_info, ARFloat center[2], ARFloat width, ARFloat conv[3][4]);

	ARFloat arGetTransMatContSub(ARMarkerInfo *marker_info, ARFloat prev_conv[3][4], ARFloat center[2], ARFloat width, ARFloat conv[3][4]);



	ARInt16* arLabeling(ARUint8 *image, int thresh,int *label_num, int **area,
						ARFloat **pos, int **clip, int **label_ref );


	ARInt16* arLabeling_ABGR(ARUint8 *image, int thresh,int *label_num, int **area, ARFloat **pos, int **clip, int **label_ref);
	ARInt16* arLabeling_BGR(ARUint8 *image, int thresh,int *label_num, int **area, ARFloat **pos, int **clip, int **label_ref);
	ARInt16* arLabeling_RGB(ARUint8 *image, int thresh,int *label_num, int **area, ARFloat **pos, int **clip, int **label_ref);
	ARInt16* arLabeling_RGB565(ARUint8 *image, int thresh,int *label_num, int **area, ARFloat **pos, int **clip, int **label_ref);
	ARInt16* arLabeling_LUM(ARUint8 *image, int thresh,int *label_num, int **area, ARFloat **pos, int **clip, int **label_ref);

	//ARInt16* labeling2(ARUint8 *image, int thresh,int *label_num, int **area,
	//				   ARFloat **pos, int **clip, int **label_ref, int LorR );

	//ARInt16* labeling3(ARUint8 *image, int thresh, int *label_num, int **area,
	//				   ARFloat **pos, int **clip, int **label_ref, int LorR );


	int arActivatePatt(int patno);

	int arDeactivatePatt(int patno);

	int arMultiActivate(ARMultiMarkerInfoT *config);

	int arMultiDeactivate( ARMultiMarkerInfoT *config );

	int verify_markers(ARMarkerInfo *marker_info, int marker_num, ARMultiMarkerInfoT *config);

	int arInitCparam( Camera *pCam );

	int arGetLine(int x_coord[], int y_coord[], int coord_num, int vertex[], ARFloat line[4][3], ARFloat v[4][2]);

	int arGetLine2(int x_coord[], int y_coord[], int coord_num, int vertex[], ARFloat line[4][3], ARFloat v[4][2], Camera *pCam);

	static int arUtilMatMul(ARFloat s1[3][4], ARFloat s2[3][4], ARFloat d[3][4]);

	static int arUtilMatInv(ARFloat s[3][4], ARFloat d[3][4]);

	static int arMatrixPCA(ARMat *input, ARMat *evec, ARVec *ev, ARVec *mean);

	static int arMatrixPCA2(ARMat *input, ARMat *evec, ARVec *ev);

	static int arParamSaveDouble(char *filename, int num, ARParamDouble *param, ...);

	static int arParamLoadDouble(char *filename, int num, ARParamDouble *param, ...);

	static int arParamDecomp(ARParam *source, ARParam *icpara, ARFloat trans[3][4]);

	static int arParamDecompMat(ARFloat source[3][4], ARFloat cpara[3][4], ARFloat trans[3][4]);

	int arParamObserv2Ideal_none(Camera* pCam, ARFloat ox, ARFloat oy, ARFloat *ix, ARFloat *iy);

	int arParamObserv2Ideal_LUT(Camera* pCam, ARFloat ox, ARFloat oy, ARFloat *ix, ARFloat *iy);

	int arParamObserv2Ideal_std(Camera* pCam, ARFloat ox, ARFloat oy, ARFloat *ix, ARFloat *iy);
	int arParamIdeal2Observ_std(Camera* pCam, ARFloat ix, ARFloat iy, ARFloat *ox, ARFloat *oy);

	typedef int (TrackerImpl::* ARPARAM_UNDIST_FUNC)(Camera* pCam, ARFloat ox, ARFloat oy, ARFloat *ix, ARFloat *iy);

	typedef ARFloat (TrackerImpl::* POSE_ESTIMATOR_FUNC)(ARMarkerInfo *marker_info, ARFloat center[2], ARFloat width, ARFloat conv[3][4]);
	typedef ARFloat (TrackerImpl::* MULTI_POSE_ESTIMATOR_FUNC)(ARMarkerInfo *marker_info, int marker_num, ARMultiMarkerInfoT *config);

	void buildUndistO2ITable(Camera* pCam);

	void checkRGB565LUT();

	// calculates amount of data that will be allocated via artkp_Alloc()
	static size_t getDynamicMemoryRequirements();

	Profiler& getProfiler()  {  return profiler;  }


public:
	static void* operator new(size_t size);

	static void operator delete(void *rawMemory);


	// required for calib camera, should otherwise not be used directly
	//
	void setFittingMode(int nWhich)  {  arFittingMode = nWhich;  }

	ARFloat arGetTransMat3(ARFloat rot[3][3], ARFloat ppos2d[][2],
						   ARFloat ppos3d[][2], int num, ARFloat conv[3][4],
						   Camera *pCam);

	static int arParamObserv2Ideal(Camera *pCam, ARFloat ox, ARFloat oy, ARFloat *ix, ARFloat *iy);
	static int arParamIdeal2Observ(Camera *pCam, ARFloat ix, ARFloat iy, ARFloat *ox, ARFloat *oy);


protected:
	//int refineCorner(ARFloat &edge_x, ARFloat &edge_y, const ARFloat old_edge_x, const ARFloat old_edge_y, const int roi_radius,
	//				 const void* img_ptr, const int img_width, const int img_height);

	struct AutoThreshold {
		enum {
			MINLUM0 = 255,
			MAXLUM0 = 0
		};

		void reset()
		{
			minLum = MINLUM0;  maxLum = MAXLUM0;
		}

		void addValue(int nRed, int nGreen, int nBlue, int nPixelFormat)
		{
			int lum;

			// in RGB565 and LUM8 all three values are simply the grey value...
			if(nPixelFormat==PIXEL_FORMAT_RGB565 || nPixelFormat==PIXEL_FORMAT_LUM)
				lum = nRed;
			else
				lum = (nRed + (nGreen<<1) + nBlue)>>2;

			if(lum<minLum)
				minLum = lum;
			if(lum>maxLum)
				maxLum = lum;
		}

		int calc()
		{
			return (minLum+maxLum)/2;
		}

		bool enable;
		int minLum,maxLum;
		int numRandomRetries;
	} autoThreshold;


	PIXEL_FORMAT			pixelFormat;
	int						pixelSize;

	int						binaryMarkerThreshold;

	// arDetectMarker.cpp
	//
	ARMarkerInfo2			*marker_info2;
	ARMarkerInfo			*wmarker_info;
	int						wmarker_num;

	arPrevInfo				prev_info[MAX_IMAGE_PATTERNS];
	int						prev_num;

	arPrevInfo				sprev_info[2][MAX_IMAGE_PATTERNS];
	int						sprev_num[2];

	// arDetectMarker2.cpp
	//
	ARMarkerInfo2			*marker_infoTWO;		// CAUTION: this member has to be manually allocated!
													//          see TrackerSingleMarker for more info on this.

    int						arGetContour_wx[AR_CHAIN_MAX];
    int						arGetContour_wy[AR_CHAIN_MAX];


	// arGetCode.cpp
	int    pattern_num;
	int    patf[MAX_LOAD_PATTERNS];
	int    pat[MAX_LOAD_PATTERNS][4][PATTERN_HEIGHT*PATTERN_WIDTH*3];
	ARFloat patpow[MAX_LOAD_PATTERNS][4];
	int    patBW[MAX_LOAD_PATTERNS][4][PATTERN_HEIGHT*PATTERN_WIDTH*3];
	ARFloat patpowBW[MAX_LOAD_PATTERNS][4];

	ARFloat evec[EVEC_MAX][PATTERN_HEIGHT*PATTERN_WIDTH*3];
	ARFloat epat[MAX_LOAD_PATTERNS][4][EVEC_MAX];
	int    evec_dim;
	int    evecf;
	ARFloat evecBW[EVEC_MAX][PATTERN_HEIGHT*PATTERN_WIDTH*3];
	ARFloat epatBW[MAX_LOAD_PATTERNS][4][EVEC_MAX];
	int    evec_dimBW;
	int    evecBWf;

	// arGetMarkerInfo.cpp
	//
	ARMarkerInfo    marker_infoL[MAX_IMAGE_PATTERNS];

	// arGetTransMat.cpp
	//
	ARFloat  pos2d[P_MAX][2];
	ARFloat  pos3d[P_MAX][3];

	// arLabeling.cpp
	//
	ARInt16      *l_imageL; //[HARDCODED_BUFFER_WIDTH*HARDCODED_BUFFER_HEIGHT];		// dyna
	ARInt16      *l_imageR;
	int			 l_imageL_size;

	int          *workL;  //[WORK_SIZE];											// dyna
	int          *work2L; //[WORK_SIZE*7];											// dyna

	int          *workR;
	int          *work2R;
	int          *wareaR;
	int          *wclipR;
	ARFloat      *wposR;

	int          wlabel_numL;
	int          wlabel_numR;
	int          *wareaL;  //[WORK_SIZE];											// dyna
	int          *wclipL;  //[WORK_SIZE*4];											// dyna
	ARFloat       *wposL;  //[WORK_SIZE*2];											// dyna

	int        arFittingMode;
	int        arImageProcMode;
	Camera	   *arCamera;
	bool		loadCachedUndist;
	int        arImXsize, arImYsize;
	int        arTemplateMatchingMode;
	int        arMatchingPCAMode;

	ARUint8*   arImageL;

	MARKER_MODE		markerMode;

	unsigned char *RGB565_to_LUM8_LUT;		// lookup table for RGB565 to LUM8 conversion


	// camera distortion addon by Daniel
	//
	UNDIST_MODE		undistMode;
	unsigned int	*undistO2ITable;
	//unsigned int	*undistI2OTable;


	ARFloat			relBorderWidth;

	ARPARAM_UNDIST_FUNC arParamObserv2Ideal_func;
	//ARPARAM_UNDIST_FUNC arParamIdeal2Observ_func;

	// RPP integration -- [t.pintaric]
	POSE_ESTIMATOR  poseEstimator;
	//POSE_ESTIMATOR_FUNC poseEstimator_func;
	//MULTI_POSE_ESTIMATOR_FUNC multiPoseEstimator_func;

	ARToolKitPlus::Logger	*logger;

	int						screenWidth, screenHeight;
	int						thresh;

	ARParam					cparam;

	ARFloat					gl_para[16];
	ARFloat					gl_cpara[16];

	char					*descriptionString;

	struct {
		bool enabled;
		int corners, leftright, bottomtop;
	} vignetting;

	BCH						*bchProcessor;
	Profiler				profiler;
};


}	// namespace ARToolKitPlus



// this is templated code, so we need to include all this here...
//
// Included ../../src/extra/FixedPoint.h
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id$
 * @file
 * ======================================================================== */


#ifndef __FIXEDPOINT_HEADERFILE__
#define __FIXEDPOINT_HEADERFILE__


namespace ARToolKitPlus {


#define FIXED_Float_To_Fixed_n(x, n)       ((I32)(x * (float)(1 << n) + 0.5f))
#define FIXED_Fixed_n_To_Float(x, n)	((float)x / (float)(1 << n))	



#ifdef _USEGPP_
#  include <gpp.h>
#  define _USE_GPP_ARITHMETIC_
#  define _USE_GPP_TRIGONOMETRIC_
#  define _USE_GPP_VECTOR_
#  define _FIXEDPOINT_MATH_ACTIVATED_

// autolink against Intel GPP
#  ifdef DEBUG
#    pragma comment( lib, "gpp_WMMX40_d.lib" )
#  else
#    pragma comment( lib, "gpp_WMMX40_r.lib" )
#  endif
#endif


#ifdef _USEFIXED_
#  if defined(_WIN32) || defined(WIN32)
	typedef unsigned __int64	U64;
	typedef __int64				I64;
	typedef unsigned int U32;
	typedef int I32;
#  endif
#  define _USE_GENERIC_ARITHMETIC_
#  define _USE_GENERIC_TRIGONOMETRIC_
#  define _USE_GENERIC_VECTOR_
#  define _FIXEDPOINT_MATH_ACTIVATED_
#endif


#ifdef _USE_GPP_ARITHMETIC_

#define FIXED_MUL2(a,b, res, bits)  \
	gppMul_n_32s((a), (b), &res, bits);

#define FIXED_MUL3(a,b,c, res, bits)  \
	gppMul_n_32s((a), (b), &_tmp1, bits);  \
	gppMul_n_32s(_tmp1, (c), &res, bits);

#define FIXED_MUL4(a,b,c,d, res, bits)  \
	gppMul_n_32s((a), (b), &_tmp1, bits);  \
	gppMul_n_32s(_tmp1, (c), &_tmp2, bits);  \
	gppMul_n_32s(_tmp2, (d), &res, bits);

#define FIXED_DIV2(a,b, res, bits)  \
	gppDiv_n_32s((a), (b), &res, bits);

#endif //_USE_GPP_ARITHMETIC_


#ifdef _USE_GPP_TRIGONOMETRIC_

#define FIXED_SIN(theta, sin_theta, n) \
	gppSinHP_n_32s((theta), sin_theta, n);

#define FIXED_COS(theta, cos_theta, n) \
	gppCosHP_n_32s((theta), cos_theta, n);

#define FIXED_SINCOS(theta, sin_theta, cos_theta, n) \
	gppSinCosHP_n_32s((theta), sin_theta, cos_theta, n);

#endif // _USE_GPP_TRIGONOMETRIC_


#ifdef _USE_GPP_VECTOR_

typedef GPP_VEC3D FIXED_VEC3D;

#define FIXED_VEC3_DOT(vec1, vec2, res, n) \
	gppVec3DDot_n_32s((vec1), (vec2), (res), (n));

#define FIXED_VEC3_SUB(vec1, vec2, res) \
	gppVec3DSub_n_32s((vec1), (vec2), (res));

#define FIXED_VEC3_LENGTH_SQ(vec, res, n) \
	gppVec3DLengthSq_n_32s((vec), (res), (n));

#endif // _USE_GPP_VECTOR_



#ifdef _USE_GENERIC_ARITHMETIC_

#  define FIXED_MUL2(a,b, res, bits)  \
	res = ((I32) (((I64)a * (I64)b)  >> bits));

#  define FIXED_MUL3(a,b,c, res, bits)  \
	FIXED_MUL2((a), (b), (res), bits);  \
	FIXED_MUL2((res), (c), (res), bits);

#  define FIXED_MUL4(a,b,c,d, res, bits)  \
	FIXED_MUL2((a), (b), (res), bits);  \
	FIXED_MUL2((res), (c), (res), bits);  \
	FIXED_MUL2((res), (d), (res), bits);

#  define FIXED_DIV2(a,b, res, bits)  \
	res = (I32) ((((I64)a)<<bits)/(I64)b);

#endif //_USE_GENERIC_ARITHMETIC_


#ifdef _USE_GENERIC_TRIGONOMETRIC_

void Fixed28_Init();
void Fixed28_Deinit();

inline void Fixed28_SinCos(I32 phi, I32 &sin, I32 &cos);

#define FIXED_SINCOS(theta, sin_theta, cos_theta, n) \
	Fixed28_SinCos((theta), *(sin_theta), *(cos_theta));

#endif //_USE_GENERIC_TRIGONOMETRIC_


#ifdef _USE_GENERIC_VECTOR_

typedef struct{
	I32 	x, y, z;
} FIXED_VEC3D;

inline void FIXED_VEC3_DOT(FIXED_VEC3D* vec1, FIXED_VEC3D* vec2, I32* res, I32 n)
{
	I32 x,y,z;
	FIXED_MUL2(vec1->x, vec2->x, x, n);
	FIXED_MUL2(vec1->y, vec2->y, y, n);
	FIXED_MUL2(vec1->z, vec2->z, z, n);

	*res = x+y+z;
}

inline void FIXED_VEC3_SUB(FIXED_VEC3D* vec1, FIXED_VEC3D* vec2, FIXED_VEC3D* res)
{
	res->x = vec1->x - vec2->x;
	res->y = vec1->y - vec2->y;
	res->z = vec1->z - vec2->z;
}

inline void FIXED_VEC3_LENGTH_SQ(FIXED_VEC3D* vec, U32* res, I32 n)
{
	I32 x,y,z;
	FIXED_MUL2(vec->x, vec->x, x, n);
	FIXED_MUL2(vec->y, vec->y, y, n);
	FIXED_MUL2(vec->z, vec->z, z, n);

	*res = x+y+z;
}

#endif //_USE_GENERIC_VECTOR_


}  // namespace ARToolKitPlus


#endif //__FIXEDPOINT_HEADERFILE__
// End of ../../src/extra/FixedPoint.h
// Included ../../src/core/arBitFieldPattern.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id: arBitFieldPattern.cxx 162 2006-04-19 21:28:10Z grabner $
 * @file
 * ======================================================================== */


#include <ARToolKitPlus/Tracker.h>
#include <ARToolKitPlus/arBitFieldPattern.h>

#include <assert.h>
#include <memory.h>


namespace ARToolKitPlus {


static void
applyMaskSimple(IDPATTERN& nPattern)
{
	nPattern ^= fullMask;
}


static void
applyMaskBCH(IDPATTERN& nPattern)
{
	nPattern ^= bchMask;
}


static bool
isBitSet(IDPATTERN pat, int which)
{
	return ((pat>>which)&1) != 0;
}

/*
static void
setBit(IDPATTERN& pat, int which)
{
	static IDPATTERN one = 1;
	pat |= (one<<which);
}
*/

/*
static void
clearBit(IDPATTERN& pat, int which)
{
	static IDPATTERN one = 1;
	pat &= ~(one<<which);
}
*/

static void
rotate90CW(IDPATTERN& nPattern)
{
	IDPATTERN one=1, tmpPat = nPattern;
	nPattern = 0;

	for(int i=0; i<pattBits; i++)
	{
		if(isBitSet(tmpPat, rotate90[i]))
			nPattern |= (one<<i);
	}
}


static void
generatePatternSimple(int nID, IDPATTERN& nPattern)
{
	IDPATTERN tmpPat = nID & idMask;

	nPattern = (tmpPat<<posMask0) | (tmpPat<<posMask1) | (tmpPat<<posMask2) | (tmpPat<<posMask3);
	applyMaskSimple(nPattern);
}


static void
generatePatternBCH(int nID, IDPATTERN& nPattern)
{
	BCH bchProcessor;
	_64bits encoded;

	bchProcessor.encode(encoded, nID);

	nPattern = encoded;
	applyMaskBCH(nPattern);
}


static float
checkPatternBitSimple(IDPATTERN nPattern, int nBit, int& nValue)
{
	int b0 = (int)(nPattern>>(posMask0+nBit))&1,
		b1 = (int)(nPattern>>(posMask1+nBit))&1,
		b2 = (int)(nPattern>>(posMask2+nBit))&1,
		b3 = (int)(nPattern>>(posMask3+nBit))&1,
		sum = b0+b1+b2+b3;

	switch(sum)
	{
	case 0:
		nValue = 0;
		return 1.00f;
	case 1:
		nValue = 0;
		return 0.50f;
	case 2:
		nValue = 0;			// opt for zero since one means white area which can happen due to reflectance
		return 0.00f;
	case 3:
		nValue = 1;
		return 0.50f;
	case 4:
		nValue = 1;
		return 1.00f;
	}

	// should never come here...
	//assert(false);
	return 0.0f;
}


static void
checkPatternSimple(IDPATTERN nPattern, int& nID, float& nProp)
{
	nProp = 0.0f;
	nID = 0;

	applyMaskSimple(nPattern);

	for(int i=0; i<idBits; i++)
	{
		int bitvalue = 0;
		nProp += checkPatternBitSimple(nPattern, i, bitvalue);

		nID |= (bitvalue<<i);
	}

	nProp /= (float)idBits;

	if(nProp<0.9f)
		nProp = 0.0f;
}


static void
checkPatternBCH(IDPATTERN nPattern, int& nID, float& nProp, BCH* nProcessor)
{
	int err = -1;
	_64bits decodedPattern = 0;
	unsigned int andMask = (1<<bchBits)-1;

	nProp = 0.0f;
	nID = 0;

	applyMaskBCH(nPattern);

	nProcessor->decode(err, decodedPattern, nPattern);

	nID = (int)(decodedPattern & andMask);

	switch(err)
	{
	case 0:
		nProp = 1.00f;
		break;

	case 1:
		nProp = 0.75f;
		break;

	case 2:
		nProp = 0.50f;
		break;

	case 3:
		nProp = 0.25f;
		break;

	default:
		nProp = 0.0f;
		break;
	}
}


AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::downsamplePattern(ARUint8* data, unsigned char* imgPtr)
{
	int x,y;

	if(PATTERN_WIDTH==18 && PATTERN_HEIGHT==18)
	{
		// this code simply takes the center pixel of each 3x3 cell
		// this is the fastest method for 18x18 to convert down to 6x6 but gives worst results
		// even more: this method gives no advantage over directly using a pattern size of 6x6.
		//
/*
		for(y=1; y<PATTERN_HEIGHT; y+=3)
			for(x=1; x<PATTERN_WIDTH; x+=3)
			{
				int idx = (y*PATTERN_WIDTH+x)*3;
				*imgPtr++ = (data[idx+0]+(data[idx+1]<<1)+data[idx+2])>>2;
			}
*/

		// this code downsamples the 18x18 RGB24 image to 6x6 LUM8 by
		// averaging each 3x3 cell into a single pixel
		//
		for(y=0; y<PATTERN_HEIGHT; y+=3)
			for(x=0; x<PATTERN_WIDTH; x+=3)
			{
				int idx = (y*PATTERN_WIDTH+x)*3, val=0;
				val = (data[idx+0]+(data[idx+1]<<1)+data[idx+2])>>2;

				idx += 3;
				val += (data[idx+0]+(data[idx+1]<<1)+data[idx+2])>>2;

				idx += 3;
				val += (data[idx+0]+(data[idx+1]<<1)+data[idx+2])>>2;

				idx += PATTERN_WIDTH*3 - 6;
				val += (data[idx+0]+(data[idx+1]<<1)+data[idx+2])>>2;

				idx += 3;
				val += (data[idx+0]+(data[idx+1]<<1)+data[idx+2])>>2;

				idx += 3;
				val += (data[idx+0]+(data[idx+1]<<1)+data[idx+2])>>2;

				idx += PATTERN_WIDTH*3 - 6;
				val += (data[idx+0]+(data[idx+1]<<1)+data[idx+2])>>2;

				idx += 3;
				val += (data[idx+0]+(data[idx+1]<<1)+data[idx+2])>>2;

				idx += 3;
				val += (data[idx+0]+(data[idx+1]<<1)+data[idx+2])>>2;

				*imgPtr++ = val/9;
			}

/*
		FILE* fp = fopen("dump18x18.raw", "wb");
		fwrite(data, PATTERN_HEIGHT*PATTERN_WIDTH*3, 1, fp);
		fclose(fp);

		fp = fopen("dump6x6.raw", "wb");
		fwrite(patimg, 6*6, 1, fp);
		fclose(fp);
*/
	}
	else
	if(PATTERN_WIDTH==12 && PATTERN_HEIGHT==12)
	{
		// this code downsamples the 12x12 RGB24 image to 6x6 LUM8 by
		// averaging each 2x2 cell into a single pixel
		//
		for(y=0; y<PATTERN_HEIGHT; y+=2)
			for(x=0; x<PATTERN_WIDTH; x+=2)
			{
				int idx = (y*PATTERN_WIDTH+x)*3, val=0;
				val = (data[idx+0]+(data[idx+1]<<1)+data[idx+2])>>2;

				idx += 3;
				val += (data[idx+0]+(data[idx+1]<<1)+data[idx+2])>>2;

				idx += PATTERN_WIDTH*3 - 3;
				val += (data[idx+0]+(data[idx+1]<<1)+data[idx+2])>>2;

				idx += 3;
				val += (data[idx+0]+(data[idx+1]<<1)+data[idx+2])>>2;

				*imgPtr++ = val/4;
			}

/*
		FILE* fp = fopen("dump12x12.raw", "wb");
		fwrite(data, PATTERN_HEIGHT*PATTERN_WIDTH*3, 1, fp);
		fclose(fp);

		fp = fopen("dump6x6.raw", "wb");
		fwrite(patimg, 6*6, 1, fp);
		fclose(fp);
*/
	}
	else
	if(PATTERN_WIDTH==6 && PATTERN_HEIGHT==6)
	{
		// this code simply converts the 6x6 RGB pattern into a
		// 6x6 greyscale image
		//
		for(int idx=0; idx<PATTERN_WIDTH*PATTERN_HEIGHT*3; idx+=3)
			*imgPtr++ = (data[idx+0]+(data[idx+1]<<1)+data[idx+2])>>2;

/*
	FILE* fp = fopen("dump6x6.raw", "wb");
	fwrite(data, PATTERN_HEIGHT*PATTERN_WIDTH, 1, fp);
	fclose(fp);
*/
	}
	else
	{
		// the pattern size has to be 18x18, 12x12 or 6x6
		// for performance reasons generic downsampling is not supported
		assert((PATTERN_WIDTH==18 && PATTERN_HEIGHT==18) || (PATTERN_WIDTH==12 && PATTERN_HEIGHT==12) || (PATTERN_WIDTH==6 && PATTERN_HEIGHT==6));
		return -1;
	}

	return 0;
}



AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::bitfield_check_simple( ARUint8 *data, int *code, int *dir, ARFloat *cf, int thresh)
{
	assert(sizeof(IDPATTERN)>=8 && "IDPATTERN must be at least 64-bit integer");

	unsigned char patimg[idPattWidth*idPattHeight], *imgPtr=patimg;
	int i;

	// first step is to reduce the pattern to 6x6. only the following
	// three resolutions are supported:
	//   - 18x18 (downsampling of each 3x3 cell)
	//   - 12x12 (downsampling of each 2x2 cell)
	//   -  6x6  (no downsampling)
	//
	if(downsamplePattern(data, imgPtr)==-1)
	{
		*code = 0;
		*dir = 0;
		*cf = -1.0f;
		return -1;
	}


	// now we do a thresholding and create the IDPATTERN bitfield
	//
	IDPATTERN pat=0, one=1;

	for(i=0; i<pattBits; i++)
		if(patimg[pattBits-1-i]>thresh)
			pat |= one<<i;


	// finally we check all four rotations and take the best one
	// if it is good enough
	//
	IDPATTERN	pat0, pat90, pat180, pat270;
	int			id0=-1,id90=-1,id180=-1,id270=-1;
	float		prop0=0.0f,prop90=0.0f,prop180=0.0f,prop270=0.0f;

	pat0 = pat;
	checkPatternSimple(pat0, id0, prop0);

	pat90 = pat0;
	rotate90CW(pat90);
	checkPatternSimple(pat90, id90, prop90);

	pat180 = pat90;
	rotate90CW(pat180);
	checkPatternSimple(pat180, id180, prop180);

	pat270 = pat180;
	rotate90CW(pat270);
	checkPatternSimple(pat270, id270, prop270);

	if(prop0>=prop90 && prop0>=prop180 && prop0>=prop270)		// is prop0 maximum?
	{
		*dir = 0;
		*cf = prop0;
		*code = id0;
	}
	else
	if(prop90>=prop0 && prop90>=prop180 && prop90>=prop270)		// is prop90 maximum?
	{
		*dir = 1;
		*cf = prop90;
		*code = id90;
	}
	else
	if(prop180>=prop0 && prop180>=prop90 && prop180>=prop270)	// is prop180 maximum?
	{
		*dir = 2;
		*cf = prop180;
		*code = id180;
	}
	else
	if(prop270>=prop0 && prop270>=prop90 && prop270>=prop180)	// is prop270 maximum?
	{
		*dir = 3;
		*cf = prop270;
		*code = id270;
	}
	else
	{
		assert(false);
	}
	
	return 0;
}


AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::bitfield_check_BCH( ARUint8 *data, int *code, int *dir, ARFloat *cf, int thresh)
{
	assert(sizeof(IDPATTERN)>=8 && "IDPATTERN must be at least 64-bit integer");

	unsigned char patimg[idPattWidth*idPattHeight], *imgPtr=patimg;
	int i;


	// first step is to reduce the pattern to 6x6. only the following
	// three resolutions are supported:
	//   - 18x18 (downsampling of each 3x3 cell)
	//   - 12x12 (downsampling of each 2x2 cell)
	//   -  6x6  (no downsampling)
	//
	if(downsamplePattern(data, imgPtr)==-1)
	{
		*code = 0;
		*dir = 0;
		*cf = -1.0f;
		return -1;
	}


	// now we do a thresholding and create the IDPATTERN bitfield
	//
	IDPATTERN pat=0, one=1;

	for(i=0; i<pattBits; i++)
		if(patimg[pattBits-1-i]>thresh)
			pat |= one<<i;


	// finally we check all four rotations and take the best one
	// if it is good enough
	//
	IDPATTERN	pat0, pat90, pat180, pat270;
	int			id0=-1,id90=-1,id180=-1,id270=-1;
	float		prop0=0.0f,prop90=0.0f,prop180=0.0f,prop270=0.0f;

	if(bchProcessor==NULL)
		bchProcessor = new BCH;

	pat0 = pat;
	checkPatternBCH(pat0, id0, prop0, bchProcessor);

	pat90 = pat0;
	rotate90CW(pat90);
	checkPatternBCH(pat90, id90, prop90, bchProcessor);

	pat180 = pat90;
	rotate90CW(pat180);
	checkPatternBCH(pat180, id180, prop180, bchProcessor);

	pat270 = pat180;
	rotate90CW(pat270);
	checkPatternBCH(pat270, id270, prop270, bchProcessor);

	if(prop0>=prop90 && prop0>=prop180 && prop0>=prop270)		// is prop0 maximum?
	{
		*dir = 0;
		*cf = prop0;
		*code = id0;
	}
	else
	if(prop90>=prop0 && prop90>=prop180 && prop90>=prop270)		// is prop90 maximum?
	{
		*dir = 1;
		*cf = prop90;
		*code = id90;
	}
	else
	if(prop180>=prop0 && prop180>=prop90 && prop180>=prop270)	// is prop180 maximum?
	{
		*dir = 2;
		*cf = prop180;
		*code = id180;
	}
	else
	if(prop270>=prop0 && prop270>=prop90 && prop270>=prop180)	// is prop270 maximum?
	{
		*dir = 3;
		*cf = prop270;
		*code = id270;
	}
	else
	{
		assert(false);
	}
	
	return 0;
}



}  // namespace ARToolKitPlus
// End of ../../src/core/arBitFieldPattern.cxx
// Included ../../src/core/arDetectMarker.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id: arDetectMarker.cxx 162 2006-04-19 21:28:10Z grabner $
 * @file
 * ======================================================================== */


#include <ARToolKitPlus/Tracker.h>
#include <stdio.h>


namespace ARToolKitPlus {


// marker detection using tracking history
//
AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arDetectMarker(ARUint8 *dataPtr, int _thresh, ARMarkerInfo **marker_info, int *marker_num)
{
    ARInt16                *limage=NULL;
    int                    label_num;
    int                    *area, *clip, *label_ref;
    ARFloat                 *pos;
    ARFloat                 rarea, rlen, rlenmin;
    ARFloat                 diff, diffmin;
    int                    cid, cdir;
    int                    i, j, k;

	autoThreshold.reset();
	checkImageBuffer();

//	FILE* fp = fopen("imgdump.raw", "wb");
//	fwrite(dataPtr, 1, 320*240*2, fp);
//	fclose(fp);

    *marker_num = 0;

	for(int numTries = 0;;)
	{
		limage = arLabeling(dataPtr, _thresh, &label_num, &area, &pos, &clip, &label_ref);
		if(limage)
		{
			marker_info2 = arDetectMarker2(limage, label_num, label_ref, area, pos, clip, AR_AREA_MAX, AR_AREA_MIN, 1.0, &wmarker_num);
			assert(wmarker_num <= MAX_IMAGE_PATTERNS);
			if(marker_info2)
			{
				wmarker_info = arGetMarkerInfo(dataPtr, marker_info2, &wmarker_num, _thresh);
				assert(wmarker_num <= MAX_IMAGE_PATTERNS);
				if(wmarker_info && wmarker_num>0)
					break;
			}
		}

		if(!autoThreshold.enable)
			break;
		else
		{
			_thresh = thresh = (rand() % 230) + 10;
			if(++numTries>autoThreshold.numRandomRetries)
				break;
		}

	}

	if(!limage || !marker_info2 || !wmarker_info)
		return -1;

    for( i = 0; i < prev_num; i++ ) {
        rlenmin = 10.0;
        cid = -1;
        for( j = 0; j < wmarker_num; j++ ) {
            rarea = (ARFloat)prev_info[i].marker.area / (ARFloat)wmarker_info[j].area;
            if( rarea < 0.7 || rarea > 1.43 ) continue;
            rlen = ( (wmarker_info[j].pos[0] - prev_info[i].marker.pos[0])
                   * (wmarker_info[j].pos[0] - prev_info[i].marker.pos[0])
                   + (wmarker_info[j].pos[1] - prev_info[i].marker.pos[1])
                   * (wmarker_info[j].pos[1] - prev_info[i].marker.pos[1]) ) / wmarker_info[j].area;
            if( rlen < 0.5 && rlen < rlenmin ) {
                rlenmin = rlen;
                cid = j;
            }
        }
        if( cid >= 0 && wmarker_info[cid].cf < prev_info[i].marker.cf ) {
            wmarker_info[cid].cf = prev_info[i].marker.cf;
            wmarker_info[cid].id = prev_info[i].marker.id;
            diffmin = 10000.0 * 10000.0;
            cdir = -1;
            for( j = 0; j < 4; j++ ) {
                diff = 0;
                for( k = 0; k < 4; k++ ) {
                    diff += (prev_info[i].marker.vertex[k][0] - wmarker_info[cid].vertex[(j+k)%4][0])
                          * (prev_info[i].marker.vertex[k][0] - wmarker_info[cid].vertex[(j+k)%4][0])
                          + (prev_info[i].marker.vertex[k][1] - wmarker_info[cid].vertex[(j+k)%4][1])
                          * (prev_info[i].marker.vertex[k][1] - wmarker_info[cid].vertex[(j+k)%4][1]);
                }
                if( diff < diffmin ) {
                    diffmin = diff;
                    cdir = (prev_info[i].marker.dir - j + 4) % 4;
                }
            }
            wmarker_info[cid].dir = cdir;
        }
    }

    for( i = 0; i < wmarker_num; i++ ) {
        if( wmarker_info[i].cf < 0.5 ) wmarker_info[i].id = -1;
   }


/*------------------------------------------------------------*/

    for( i = j = 0; i < prev_num; i++ ) {
        prev_info[i].count++;
	if( (prev_info[i].count) < 4 ) {
            prev_info[j] = prev_info[i];
            j++;
        }
    }
    prev_num = j;

    for( i = 0; i < wmarker_num; i++ ) {
        if( wmarker_info[i].id < 0 )
			continue;

        for( j = 0; j < prev_num; j++ ) {
            if( prev_info[j].marker.id == wmarker_info[i].id )
				break;
        }
		if(j<MAX_IMAGE_PATTERNS)
		{
			prev_info[j].marker = wmarker_info[i];
			prev_info[j].count  = 1;
			if( j == prev_num )
				prev_num++;
		}
    }

    for( i = 0; i < prev_num; i++ ) {
        for( j = 0; j < wmarker_num; j++ ) {
            rarea = (ARFloat)prev_info[i].marker.area / (ARFloat)wmarker_info[j].area;
            if( rarea < 0.7 || rarea > 1.43 ) continue;
            rlen = ( (wmarker_info[j].pos[0] - prev_info[i].marker.pos[0])
                   * (wmarker_info[j].pos[0] - prev_info[i].marker.pos[0])
                   + (wmarker_info[j].pos[1] - prev_info[i].marker.pos[1])
                   * (wmarker_info[j].pos[1] - prev_info[i].marker.pos[1]) ) / wmarker_info[j].area;
            if( rlen < 0.5 ) break;
        }
        if(j==wmarker_num && wmarker_num<MAX_IMAGE_PATTERNS) {
            wmarker_info[wmarker_num] = prev_info[i].marker;
            wmarker_num++;
			assert(wmarker_num <= MAX_IMAGE_PATTERNS);
        }
    }


    *marker_num  = wmarker_num;
    *marker_info = wmarker_info;

	assert(*marker_num <= MAX_IMAGE_PATTERNS);

	if(autoThreshold.enable)
		thresh = autoThreshold.calc();

    return 0;
}


// marker detection without using tracking history
//
AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arDetectMarkerLite(ARUint8 *dataPtr, int _thresh, ARMarkerInfo **marker_info, int *marker_num)
{
    ARInt16                *limage = NULL;
    int                    label_num;
    int                    *area, *clip, *label_ref;
    ARFloat                 *pos;
    int                    i;

	autoThreshold.reset();
	checkImageBuffer();

    *marker_num = 0;

	for(int numTries = 0;;)
	{
		limage = arLabeling(dataPtr, _thresh, &label_num, &area, &pos, &clip, &label_ref);
		if(limage)
		{
			marker_info2 = arDetectMarker2(limage, label_num, label_ref, area, pos, clip, AR_AREA_MAX, AR_AREA_MIN, 1.0, &wmarker_num);
			if(marker_info2)
			{
				wmarker_info = arGetMarkerInfo(dataPtr, marker_info2, &wmarker_num, _thresh);
				if(wmarker_info && wmarker_num>0)
					break;
			}
		}

		if(!autoThreshold.enable)
			break;
		else
		{
			_thresh = thresh = (rand() % 230) + 10;
			if(++numTries>autoThreshold.numRandomRetries)
				break;
		}

	}

	if(!limage || !marker_info2 || !wmarker_info)
		return -1;


/*
    limage = arLabeling(dataPtr, _thresh, &label_num, &area, &pos, &clip, &label_ref);
    if( limage == 0 )    return -1;

    marker_info2 = arDetectMarker2(limage, label_num, label_ref, area, pos, clip, AR_AREA_MAX, AR_AREA_MIN, 1.0, &wmarker_num);
    if( marker_info2 == 0 ) return -1;

    wmarker_info = arGetMarkerInfo(dataPtr, marker_info2, &wmarker_num, _thresh);
    if( wmarker_info == 0 ) return -1;
*/

    for( i = 0; i < wmarker_num; i++ )
        if( wmarker_info[i].cf < 0.5 )
			wmarker_info[i].id = -1;


    *marker_num  = wmarker_num;
    *marker_info = wmarker_info;

	if(autoThreshold.enable)
		thresh = autoThreshold.calc();

    return 0;
}


}	// namespace ARToolKitPlus
// End of ../../src/core/arDetectMarker.cxx
// Included ../../src/core/arDetectMarker2.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id: arDetectMarker2.cxx 162 2006-04-19 21:28:10Z grabner $
 * @file
 * ======================================================================== */


#include <ARToolKitPlus/Tracker.h>


namespace ARToolKitPlus {

static int get_vertex( int x_coord[], int y_coord[], int st,  int ed,
                       ARFloat thresh, int vertex[], int *vnum);


AR_TEMPL_FUNC ARMarkerInfo2*
AR_TEMPL_TRACKER::arDetectMarker2(ARInt16 *limage, int label_num, int *label_ref,
                    int *warea, ARFloat *wpos, int *wclip,
                    int area_max, int area_min, ARFloat factor, int *marker_num)
{
    ARMarkerInfo2     *pm;
    int               xsize, ysize;
    int               marker_num2;
    int               i, j, ret;
    ARFloat            d;

	PROFILE_BEGINSEC(profiler, DETECTMARKER2)

    if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) {
        area_min /= 4;
        area_max /= 4;
        xsize = arImXsize / 2;
        ysize = arImYsize / 2;
    }
    else {
        xsize = arImXsize;
        ysize = arImYsize;
    }
    marker_num2 = 0;
    for(i=0; i<label_num; i++ ) {
        if( warea[i] < area_min || warea[i] > area_max ) continue;
        if( wclip[i*4+0] == 1 || wclip[i*4+1] == xsize-2 ) continue;
        if( wclip[i*4+2] == 1 || wclip[i*4+3] == ysize-2 ) continue;

        ret = arGetContour( limage, label_ref, i+1,
                            &(wclip[i*4]), &(marker_infoTWO[marker_num2]));
        if( ret < 0 ) continue;

        ret = check_square( warea[i], &(marker_infoTWO[marker_num2]), factor );
        if( ret < 0 ) continue;

        marker_infoTWO[marker_num2].area   = warea[i];
        marker_infoTWO[marker_num2].pos[0] = wpos[i*2+0];
        marker_infoTWO[marker_num2].pos[1] = wpos[i*2+1];
        marker_num2++;
        if(marker_num2==MAX_IMAGE_PATTERNS)
			break;
    }

    for( i=0; i < marker_num2; i++ ) {
        for( j=i+1; j < marker_num2; j++ ) {
            d = (marker_infoTWO[i].pos[0] - marker_infoTWO[j].pos[0])
              * (marker_infoTWO[i].pos[0] - marker_infoTWO[j].pos[0])
              + (marker_infoTWO[i].pos[1] - marker_infoTWO[j].pos[1])
              * (marker_infoTWO[i].pos[1] - marker_infoTWO[j].pos[1]);
            if( marker_infoTWO[i].area > marker_infoTWO[j].area ) {
                if( d < marker_infoTWO[i].area / 4 ) {
                    marker_infoTWO[j].area = 0;
                }
            }
            else {
                if( d < marker_infoTWO[j].area / 4 ) {
                    marker_infoTWO[i].area = 0;
                }
            }
        }
    }
    for( i=0; i < marker_num2; i++ ) {
        if( marker_infoTWO[i].area == 0.0 ) {
            for( j=i+1; j < marker_num2; j++ ) {
                marker_infoTWO[j-1] = marker_infoTWO[j];
            }
            marker_num2--;
        }
    }

    if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) {
        pm = &(marker_infoTWO[0]);
        for( i = 0; i < marker_num2; i++ ) {
            pm->area *= 4;
            pm->pos[0] *= 2.0;
            pm->pos[1] *= 2.0;
            for( j = 0; j< pm->coord_num; j++ ) {
                pm->x_coord[j] *= 2;
                pm->y_coord[j] *= 2;
            }
            pm++;
        }
    }

	PROFILE_ENDSEC(profiler, DETECTMARKER2)

    *marker_num = marker_num2;
    return( &(marker_infoTWO[0]) );
}


AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arGetContour(ARInt16 *limage, int *label_ref, int label, int clip[4], ARMarkerInfo2 *marker_infoTWO)
{
    static const int      xdir[8] = { 0, 1, 1, 1, 0,-1,-1,-1};
    static const int      ydir[8] = {-1,-1, 0, 1, 1, 1, 0,-1};
    //static int      wx[AR_CHAIN_MAX];
    //static int      wy[AR_CHAIN_MAX];
    ARInt16         *p1;
    int             xsize, ysize;
    int             sx, sy, dir;
    int             dmax, d, v1 = 0;
    int             i, j;

    if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) {
        xsize = arImXsize / 2;
        ysize = arImYsize / 2;
    }
    else {
        xsize = arImXsize;
        ysize = arImYsize;
    }
    j = clip[2];
    p1 = &(limage[j*xsize+clip[0]]);
    for( i = clip[0]; i <= clip[1]; i++, p1++ ) {
        if( *p1 > 0 && label_ref[(*p1)-1] == label ) {
            sx = i; sy = j; break;
        }
    }
    if( i > clip[1] ) {
        printf("??? 1\n"); return(-1);
    }

    marker_infoTWO->coord_num = 1;
    marker_infoTWO->x_coord[0] = sx;
    marker_infoTWO->y_coord[0] = sy;
    dir = 5;
    for(;;) {
        p1 = &(limage[marker_infoTWO->y_coord[marker_infoTWO->coord_num-1] * xsize
                    + marker_infoTWO->x_coord[marker_infoTWO->coord_num-1]]);
        dir = (dir+5)%8;
        for(i=0;i<8;i++) {
            if( p1[ydir[dir]*xsize+xdir[dir]] > 0 ) break;
            dir = (dir+1)%8;
        }
        if( i == 8 ) {
            printf("??? 2\n"); return(-1);
        }
        marker_infoTWO->x_coord[marker_infoTWO->coord_num]
            = marker_infoTWO->x_coord[marker_infoTWO->coord_num-1] + xdir[dir];
        marker_infoTWO->y_coord[marker_infoTWO->coord_num]
            = marker_infoTWO->y_coord[marker_infoTWO->coord_num-1] + ydir[dir];
        if( marker_infoTWO->x_coord[marker_infoTWO->coord_num] == sx
         && marker_infoTWO->y_coord[marker_infoTWO->coord_num] == sy ) break;
        marker_infoTWO->coord_num++;
        if( marker_infoTWO->coord_num == AR_CHAIN_MAX-1 ) {
            printf("??? 3\n"); return(-1);
        }
    }

    dmax = 0;
    for(i=1;i<marker_infoTWO->coord_num;i++) {
        d = (marker_infoTWO->x_coord[i]-sx)*(marker_infoTWO->x_coord[i]-sx)
          + (marker_infoTWO->y_coord[i]-sy)*(marker_infoTWO->y_coord[i]-sy);
        if( d > dmax ) {
            dmax = d;
            v1 = i;
        }
    }

    for(i=0;i<v1;i++) {
        arGetContour_wx[i] = marker_infoTWO->x_coord[i];
        arGetContour_wy[i] = marker_infoTWO->y_coord[i];
    }
    for(i=v1;i<marker_infoTWO->coord_num;i++) {
        marker_infoTWO->x_coord[i-v1] = marker_infoTWO->x_coord[i];
        marker_infoTWO->y_coord[i-v1] = marker_infoTWO->y_coord[i];
    }
    for(i=0;i<v1;i++) {
        marker_infoTWO->x_coord[i-v1+marker_infoTWO->coord_num] = arGetContour_wx[i];
        marker_infoTWO->y_coord[i-v1+marker_infoTWO->coord_num] = arGetContour_wy[i];
    }
    marker_infoTWO->x_coord[marker_infoTWO->coord_num] = marker_infoTWO->x_coord[0];
    marker_infoTWO->y_coord[marker_infoTWO->coord_num] = marker_infoTWO->y_coord[0];
    marker_infoTWO->coord_num++;

    return 0;
}


AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::check_square(int area, ARMarkerInfo2 *marker_infoTWO, ARFloat factor)
{
    int             sx, sy;
    int             dmax, d, v1;
    int             vertex[10], vnum;
    int             wv1[10], wvnum1, wv2[10], wvnum2, v2;
    ARFloat          thresh;
    int             i;


    dmax = 0;
    v1 = 0;
    sx = marker_infoTWO->x_coord[0];
    sy = marker_infoTWO->y_coord[0];
    for(i=1;i<marker_infoTWO->coord_num-1;i++) {
        d = (marker_infoTWO->x_coord[i]-sx)*(marker_infoTWO->x_coord[i]-sx)
          + (marker_infoTWO->y_coord[i]-sy)*(marker_infoTWO->y_coord[i]-sy);
        if( d > dmax ) {
            dmax = d;
            v1 = i;
        }
    }

    thresh = (ARFloat)(area/0.75) * (ARFloat)0.01 * factor;
    vnum = 1;
    vertex[0] = 0;
    wvnum1 = 0;
    wvnum2 = 0;
    if( get_vertex(marker_infoTWO->x_coord, marker_infoTWO->y_coord, 0,  v1,
                   thresh, wv1, &wvnum1) < 0 ) {
        return(-1);
    }
    if( get_vertex(marker_infoTWO->x_coord, marker_infoTWO->y_coord,
                   v1,  marker_infoTWO->coord_num-1, thresh, wv2, &wvnum2) < 0 ) {
        return(-1);
    }

    if( wvnum1 == 1 && wvnum2 == 1 ) {
        vertex[1] = wv1[0];
        vertex[2] = v1;
        vertex[3] = wv2[0];
    }
    else if( wvnum1 > 1 && wvnum2 == 0 ) {
        v2 = v1 / 2;
        wvnum1 = wvnum2 = 0;
        if( get_vertex(marker_infoTWO->x_coord, marker_infoTWO->y_coord,
                       0,  v2, thresh, wv1, &wvnum1) < 0 ) {
            return(-1);
        }
        if( get_vertex(marker_infoTWO->x_coord, marker_infoTWO->y_coord,
                       v2,  v1, thresh, wv2, &wvnum2) < 0 ) {
            return(-1);
        }
        if( wvnum1 == 1 && wvnum2 == 1 ) {
            vertex[1] = wv1[0];
            vertex[2] = wv2[0];
            vertex[3] = v1;
        }
        else {
            return(-1);
        }
    }
    else if( wvnum1 == 0 && wvnum2 > 1 ) {
        v2 = (v1 + marker_infoTWO->coord_num-1) / 2;
        wvnum1 = wvnum2 = 0;
        if( get_vertex(marker_infoTWO->x_coord, marker_infoTWO->y_coord,
                   v1, v2, thresh, wv1, &wvnum1) < 0 ) {
            return(-1);
        }
        if( get_vertex(marker_infoTWO->x_coord, marker_infoTWO->y_coord,
                   v2, marker_infoTWO->coord_num-1, thresh, wv2, &wvnum2) < 0 ) {
            return(-1);
        }
        if( wvnum1 == 1 && wvnum2 == 1 ) {
            vertex[1] = v1;
            vertex[2] = wv1[0];
            vertex[3] = wv2[0];
        }
        else {
            return(-1);
        }
    }
    else {
        return(-1);
    }

    marker_infoTWO->vertex[0] = vertex[0];
    marker_infoTWO->vertex[1] = vertex[1];
    marker_infoTWO->vertex[2] = vertex[2];
    marker_infoTWO->vertex[3] = vertex[3];
    marker_infoTWO->vertex[4] = marker_infoTWO->coord_num-1;

    return(0);
}

static int
get_vertex( int x_coord[], int y_coord[], int st,  int ed, ARFloat thresh, int vertex[], int *vnum)
{
    ARFloat   d, dmax;
    ARFloat   a, b, c;
    int      i, v1 = 0;

    a = (ARFloat)(y_coord[ed] - y_coord[st]);
    b = (ARFloat)(x_coord[st] - x_coord[ed]);
    c = (ARFloat)(x_coord[ed]*y_coord[st] - y_coord[ed]*x_coord[st]);
    dmax = 0;
    for(i=st+1;i<ed;i++) {
        d = a*x_coord[i] + b*y_coord[i] + c;
        if( d*d > dmax ) {
            dmax = d*d;
            v1 = i;
        }
    }
    if( dmax/(a*a+b*b) > thresh ) {
        if( get_vertex(x_coord, y_coord, st,  v1, thresh, vertex, vnum) < 0 )
            return(-1);

        if( (*vnum) > 5 ) return(-1);
        vertex[(*vnum)] = v1;
        (*vnum)++;

        if( get_vertex(x_coord, y_coord, v1,  ed, thresh, vertex, vnum) < 0 )
            return(-1);
    }

    return(0);
}


}  // namespace ARToolKitPlus
// End of ../../src/core/arDetectMarker2.cxx
// Included ../../src/core/arGetCode.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id: arGetCode.cxx 162 2006-04-19 21:28:10Z grabner $
 * @file
 * ======================================================================== */


#include <stdio.h>
#include <math.h>

#include <ARToolKitPlus/Tracker.h>
#include <ARToolKitPlus/ar.h>
#include <ARToolKitPlus/matrix.h>


namespace ARToolKitPlus {
	

static void get_cpara(ARFloat world[4][2], ARFloat vertex[4][2], ARFloat para[3][3]);
static void put_zero(ARUint8 *p, int size);


AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arLoadPatt(char *filename)
{
    FILE    *fp;
    int     patno;
    int     h, i, j, l, m;
    int     i1, i2, i3;

    if(pattern_num == -1 ) {
        for( i = 0; i < MAX_LOAD_PATTERNS; i++ ) patf[i] = 0;
        pattern_num = 0;
    }

    for( i = 0; i < MAX_LOAD_PATTERNS; i++ ) {
        if(patf[i] == 0) break;
    }
    if( i == MAX_LOAD_PATTERNS ) return -1;
    patno = i;

    if( (fp=fopen(filename, "r")) == NULL ) {
        printf("\"%s\" not found!!\n", filename);
        return(-1);
    }

    for( h=0; h<4; h++ ) {
        l = 0;
        for( i3 = 0; i3 < 3; i3++ ) {
            for( i2 = 0; i2 < PATTERN_HEIGHT; i2++ ) {
                for( i1 = 0; i1 < PATTERN_WIDTH; i1++ ) {
                    if( fscanf(fp, "%d", &j) != 1 ) {
                        printf("Pattern Data read error!!\n");
                        return -1;
                    }
					if(binaryMarkerThreshold!=-1)
						j = (j<binaryMarkerThreshold) ? 0 : 255;
                    j = 255-j;
                    pat[patno][h][(i2*PATTERN_WIDTH+i1)*3+i3] = j;
                    if( i3 == 0 ) patBW[patno][h][i2*PATTERN_WIDTH+i1]  = j;
                    else          patBW[patno][h][i2*PATTERN_WIDTH+i1] += j;
                    if( i3 == 2 ) patBW[patno][h][i2*PATTERN_WIDTH+i1] /= 3;
                    l += j;
                }
            }
        }
        l /= (PATTERN_HEIGHT*PATTERN_WIDTH*3);

        m = 0;
        for( i = 0; i < PATTERN_HEIGHT*PATTERN_WIDTH*3; i++ ) {
            pat[patno][h][i] -= l;
            m += (pat[patno][h][i]*pat[patno][h][i]);
        }
        patpow[patno][h] = (ARFloat)sqrt((ARFloat)m);
        if( patpow[patno][h] == 0.0 ) patpow[patno][h] = (ARFloat)0.0000001;

        m = 0;
        for( i = 0; i < PATTERN_HEIGHT*PATTERN_WIDTH; i++ ) {
            patBW[patno][h][i] -= l;
            m += (patBW[patno][h][i]*patBW[patno][h][i]);
        }
        patpowBW[patno][h] = (ARFloat)sqrt((ARFloat)m);
        if( patpowBW[patno][h] == 0.0 ) patpowBW[patno][h] = (ARFloat)0.0000001;
    }
    fclose(fp);

    patf[patno] = 1;
    pattern_num++;

/*
    gen_evec();
*/

    return( patno );
}

AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arFreePatt( int patno )
{
    if( patf[patno] == 0 ) return -1;

    patf[patno] = 0;
    pattern_num--;

    gen_evec();

    return 1;
}

AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arActivatePatt( int patno )
{
    if( patf[patno] == 0 ) return -1;

    patf[patno] = 1;

    return 1;
}

AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arDeactivatePatt( int patno )
{
    if( patf[patno] == 0 ) return -1;

    patf[patno] = 2;

    return 1;
}


AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arGetCode(ARUint8 *image, int *x_coord, int *y_coord, int *vertex,
				   int *code, int *dir, ARFloat *cf, int thresh)
{
    ARUint8 ext_pat[PATTERN_HEIGHT][PATTERN_WIDTH][3];

    arGetPatt(image, x_coord, y_coord, vertex, ext_pat);

	if(autoThreshold.enable)
	{
		int x,y;

		for(y=0; y<PATTERN_HEIGHT; y++)
			for(x=0; x<PATTERN_WIDTH; x++)
				//autoThreshold.addValue(ext_pat[y][x][0], ext_pat[y][x][1], ext_pat[y][x][2]);
				autoThreshold.addValue(ext_pat[y][x][0], ext_pat[y][x][1], ext_pat[y][x][2], pixelFormat);
	}


//#pragma message (">>> WARNING: compiling with marker content dumping. performance will be very low !!!")
//	FILE* fp = fopen("dump.raw", "wb");
//	fwrite(ext_pat, PATTERN_HEIGHT*PATTERN_WIDTH*3, 1, fp);
//	fclose(fp);

	switch(markerMode)
	{
	case MARKER_TEMPLATE:
		pattern_match((ARUint8 *)ext_pat, code, dir, cf);
		break;

	case MARKER_ID_SIMPLE:
		bitfield_check_simple((ARUint8 *)ext_pat, code, dir, cf, thresh);
		break;

	case MARKER_ID_BCH:
		bitfield_check_BCH((ARUint8 *)ext_pat, code, dir, cf, thresh);
		break;
	}

	/*if(useBitFieldMarkers)
	{
		if(idMarkerMode==IDMARKER_SIMPLE)
			bitfield_check_simple((ARUint8 *)ext_pat, code, dir, cf, thresh);
		else
			bitfield_check_BCH((ARUint8 *)ext_pat, code, dir, cf, thresh);
	}
	else
		pattern_match((ARUint8 *)ext_pat, code, dir, cf);*/

    return(0);
}

//#if 1
AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arGetPatt(ARUint8 *image, int *x_coord, int *y_coord, int *vertex,
						    ARUint8 ext_pat[PATTERN_HEIGHT][PATTERN_WIDTH][3])
{
    ARUint32  ext_pat2[PATTERN_HEIGHT][PATTERN_WIDTH][3];
    ARFloat    world[4][2];
    ARFloat    local[4][2];
    ARFloat    para[3][3];
    ARFloat    d, xw, yw;
    int       xc, yc;
    int       xdiv, ydiv;
    int       xdiv2, ydiv2;
    int       lx1, lx2, ly1, ly2;
    int       i, j;
	ARUint8		col8;
    // int       k1, k2, k3; // unreferenced

	unsigned short* image16 = (unsigned short*)image;

    world[0][0] = 100.0;
    world[0][1] = 100.0;
    world[1][0] = 100.0 + 10.0;
    world[1][1] = 100.0;
    world[2][0] = 100.0 + 10.0;
    world[2][1] = 100.0 + 10.0;
    world[3][0] = 100.0;
    world[3][1] = 100.0 + 10.0;
    for( i = 0; i < 4; i++ ) {
        local[i][0] = (ARFloat)x_coord[vertex[i]];
        local[i][1] = (ARFloat)y_coord[vertex[i]];
    }
    get_cpara( world, local, para );

    lx1 = (int)((local[0][0] - local[1][0])*(local[0][0] - local[1][0])
              + (local[0][1] - local[1][1])*(local[0][1] - local[1][1]));
    lx2 = (int)((local[2][0] - local[3][0])*(local[2][0] - local[3][0])
              + (local[2][1] - local[3][1])*(local[2][1] - local[3][1]));
    ly1 = (int)((local[1][0] - local[2][0])*(local[1][0] - local[2][0])
              + (local[1][1] - local[2][1])*(local[1][1] - local[2][1]));
    ly2 = (int)((local[3][0] - local[0][0])*(local[3][0] - local[0][0])
              + (local[3][1] - local[0][1])*(local[3][1] - local[0][1]));
    if( lx2 > lx1 ) lx1 = lx2;
    if( ly2 > ly1 ) ly1 = ly2;
    xdiv2 = PATTERN_WIDTH;
    ydiv2 = PATTERN_HEIGHT;
    if( arImageProcMode == AR_IMAGE_PROC_IN_FULL ) {
        while( xdiv2*xdiv2 < lx1/4 ) xdiv2*=2;
        while( ydiv2*ydiv2 < ly1/4 ) ydiv2*=2;
    }
    else {
        while( xdiv2*xdiv2*4 < lx1/4 ) xdiv2*=2;
        while( ydiv2*ydiv2*4 < ly1/4 ) ydiv2*=2;
    }
    if( xdiv2 > PATTERN_SAMPLE_NUM ) xdiv2 = PATTERN_SAMPLE_NUM;
    if( ydiv2 > PATTERN_SAMPLE_NUM ) ydiv2 = PATTERN_SAMPLE_NUM;

    xdiv = xdiv2/PATTERN_WIDTH;
    ydiv = ydiv2/PATTERN_HEIGHT;
/*
printf("%3d(%f), %3d(%f)\n", xdiv2, sqrt(lx1), ydiv2, sqrt(ly1));
*/


	// special case xdiv==1 and ydiv==1, so we can remove all divides and multiplies for indexing
	//
	if(xdiv==1 && ydiv==1)
	{
		ARFloat border = relBorderWidth * 10.0f;
		ARFloat xyFrom = 100.0f + border,
				xyTo = 110.0f - border,
				xyStep = xyTo-xyFrom;
		ARFloat steps[PATTERN_WIDTH];

		for( i = 0; i < xdiv2; i++ )
			steps[i] = xyFrom + xyStep * (ARFloat)(i+0.5f) / (ARFloat)xdiv2;

		for( j = 0; j < ydiv2; j++ ) {
			yw = steps[j];
			for( i = 0; i < xdiv2; i++ ) {
				xw = steps[i];
				d = para[2][0]*xw + para[2][1]*yw + para[2][2];
				if( d == 0 ) return(-1);
				xc = (int)((para[0][0]*xw + para[0][1]*yw + para[0][2])/d);
				yc = (int)((para[1][0]*xw + para[1][1]*yw + para[1][2])/d);
				//das hier sollte noch getestet werden...
				/*
				if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) {
					xc = ((xc+1)/2)*2;
					yc = ((yc+1)/2)*2;
				}*/

				if( xc >= 0 && xc < arImXsize && yc >= 0 && yc < arImYsize )
				{
					switch(pixelFormat)
					{
					case PIXEL_FORMAT_ABGR:
						ext_pat[j][i][0] = image[(yc*arImXsize+xc)*4+1];
						ext_pat[j][i][1] = image[(yc*arImXsize+xc)*4+2];
						ext_pat[j][i][2] = image[(yc*arImXsize+xc)*4+3];
						break;
					
					case PIXEL_FORMAT_BGRA:
						ext_pat[j][i][0] = image[(yc*arImXsize+xc)*4+0];
						ext_pat[j][i][1] = image[(yc*arImXsize+xc)*4+1];
						ext_pat[j][i][2] = image[(yc*arImXsize+xc)*4+2];
						break;

					case PIXEL_FORMAT_BGR:
						ext_pat[j][i][0] = image[(yc*arImXsize+xc)*3+0];
						ext_pat[j][i][1] = image[(yc*arImXsize+xc)*3+1];
						ext_pat[j][i][2] = image[(yc*arImXsize+xc)*3+2];
						break;

					case PIXEL_FORMAT_RGBA:
						ext_pat[j][i][0] = image[(yc*arImXsize+xc)*4+2];
						ext_pat[j][i][1] = image[(yc*arImXsize+xc)*4+1];
						ext_pat[j][i][2] = image[(yc*arImXsize+xc)*4+0];

					case PIXEL_FORMAT_RGB:
						ext_pat[j][i][0] = image[(yc*arImXsize+xc)*3+2];
						ext_pat[j][i][1] = image[(yc*arImXsize+xc)*3+1];
						ext_pat[j][i][2] = image[(yc*arImXsize+xc)*3+0];
						break;

					case PIXEL_FORMAT_RGB565:
						col8 = getLUM8_from_RGB565(image16+yc*arImXsize+xc);
						ext_pat[j][i][0] = col8;
						ext_pat[j][i][1] = col8;
						ext_pat[j][i][2] = col8;
						break;

					case PIXEL_FORMAT_LUM:
						col8 = image[(yc*arImXsize+xc)];
						ext_pat[j][i][0] = col8;
						ext_pat[j][i][1] = col8;
						ext_pat[j][i][2] = col8;
						break;
					}
				}
			}
		}
	}
	else
	// general case: xdiv!=1 or ydiv!=1
	//
	{
		ARFloat border = relBorderWidth * 10.0f;
		ARFloat xyFrom = 100.0f + border,
				xyTo = 110.0f - border,
				xyStep = xyTo-xyFrom;
		int jy,ix;
		ARUint8 col8;

		put_zero( (ARUint8 *)ext_pat2, PATTERN_HEIGHT*PATTERN_WIDTH*3*sizeof(ARUint32) );

		for( j = 0; j < ydiv2; j++ ) {
			//yw = (ARFloat)(102.5) + (ARFloat)(5.0) * (ARFloat)(j+0.5) / (ARFloat)ydiv2;
			yw = xyFrom + xyStep * (ARFloat)(j+0.5f) / (ARFloat)ydiv2;
			for( i = 0; i < xdiv2; i++ ) {
				//xw = (ARFloat)(102.5) + (ARFloat)(5.0) * (ARFloat)(i+0.5) / (ARFloat)xdiv2;
				xw = xyFrom + xyStep * (ARFloat)(i+0.5f) / (ARFloat)xdiv2;
				d = para[2][0]*xw + para[2][1]*yw + para[2][2];
				if( d == 0 ) return(-1);
				xc = (int)((para[0][0]*xw + para[0][1]*yw + para[0][2])/d);
				yc = (int)((para[1][0]*xw + para[1][1]*yw + para[1][2])/d);
				/*
				if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) {
					xc = ((xc+1)/2)*2;
					yc = ((yc+1)/2)*2;
				}*/
				if( xc >= 0 && xc < arImXsize && yc >= 0 && yc < arImYsize )
				{
					/*if(PIX_FORMAT==PIXEL_FORMAT_ABGR) {
						ext_pat2[j/ydiv][i/xdiv][0] += image[(yc*arImXsize+xc)*PIX_SIZE+1];
						ext_pat2[j/ydiv][i/xdiv][1] += image[(yc*arImXsize+xc)*PIX_SIZE+2];
						ext_pat2[j/ydiv][i/xdiv][2] += image[(yc*arImXsize+xc)*PIX_SIZE+3];
					}
					if(PIX_FORMAT==PIXEL_FORMAT_BGRA) {
						ext_pat2[j/ydiv][i/xdiv][0] += image[(yc*arImXsize+xc)*PIX_SIZE+0];
						ext_pat2[j/ydiv][i/xdiv][1] += image[(yc*arImXsize+xc)*PIX_SIZE+1];
						ext_pat2[j/ydiv][i/xdiv][2] += image[(yc*arImXsize+xc)*PIX_SIZE+2];
					}
					if(PIX_FORMAT==PIXEL_FORMAT_BGR) {
						ext_pat2[j/ydiv][i/xdiv][0] += image[(yc*arImXsize+xc)*PIX_SIZE+0];
						ext_pat2[j/ydiv][i/xdiv][1] += image[(yc*arImXsize+xc)*PIX_SIZE+1];
						ext_pat2[j/ydiv][i/xdiv][2] += image[(yc*arImXsize+xc)*PIX_SIZE+2];
					}
					if(PIX_FORMAT==PIXEL_FORMAT_RGBA) {
						ext_pat2[j/ydiv][i/xdiv][0] += image[(yc*arImXsize+xc)*PIX_SIZE+2];
						ext_pat2[j/ydiv][i/xdiv][1] += image[(yc*arImXsize+xc)*PIX_SIZE+1];
						ext_pat2[j/ydiv][i/xdiv][2] += image[(yc*arImXsize+xc)*PIX_SIZE+0];
					}
					if(PIX_FORMAT==PIXEL_FORMAT_RGB) {
						ext_pat2[j/ydiv][i/xdiv][0] += image[(yc*arImXsize+xc)*PIX_SIZE+2];
						ext_pat2[j/ydiv][i/xdiv][1] += image[(yc*arImXsize+xc)*PIX_SIZE+1];
						ext_pat2[j/ydiv][i/xdiv][2] += image[(yc*arImXsize+xc)*PIX_SIZE+0];
					}
					if(PIX_FORMAT==PIXEL_FORMAT_RGB565) {
						int jy=j/ydiv, ix=i/xdiv;
						//ARUint8 col = RGB565_to_LUM8_LUT[image16[yc*arImXsize+xc]];
						ARUint8 col = getLUM8_from_RGB565(image16+yc*arImXsize+xc);
						ext_pat2[jy][ix][0] += col;
						ext_pat2[jy][ix][1] += col;
						ext_pat2[jy][ix][2] += col;
					}
					if(PIX_FORMAT==PIXEL_FORMAT_LUM) {
						int jy=j/ydiv, ix=i/xdiv;
						ARUint8 col = image[(yc*arImXsize+xc)*PIX_SIZE];
						ext_pat2[jy][ix][0] += col;
						ext_pat2[jy][ix][1] += col;
						ext_pat2[jy][ix][2] += col;
					}*/

					switch(pixelFormat)
					{
					case PIXEL_FORMAT_ABGR:
						ext_pat2[j/ydiv][i/xdiv][0] += image[(yc*arImXsize+xc)*4+1];
						ext_pat2[j/ydiv][i/xdiv][1] += image[(yc*arImXsize+xc)*4+2];
						ext_pat2[j/ydiv][i/xdiv][2] += image[(yc*arImXsize+xc)*4+3];
						break;

					case PIXEL_FORMAT_BGRA:
						ext_pat2[j/ydiv][i/xdiv][0] += image[(yc*arImXsize+xc)*4+0];
						ext_pat2[j/ydiv][i/xdiv][1] += image[(yc*arImXsize+xc)*4+1];
						ext_pat2[j/ydiv][i/xdiv][2] += image[(yc*arImXsize+xc)*4+2];
						break;

					case PIXEL_FORMAT_BGR:
						ext_pat2[j/ydiv][i/xdiv][0] += image[(yc*arImXsize+xc)*3+0];
						ext_pat2[j/ydiv][i/xdiv][1] += image[(yc*arImXsize+xc)*3+1];
						ext_pat2[j/ydiv][i/xdiv][2] += image[(yc*arImXsize+xc)*3+2];
						break;

					case PIXEL_FORMAT_RGBA:
						ext_pat2[j/ydiv][i/xdiv][0] += image[(yc*arImXsize+xc)*4+2];
						ext_pat2[j/ydiv][i/xdiv][1] += image[(yc*arImXsize+xc)*4+1];
						ext_pat2[j/ydiv][i/xdiv][2] += image[(yc*arImXsize+xc)*4+0];
						break;

					case PIXEL_FORMAT_RGB:
						ext_pat2[j/ydiv][i/xdiv][0] += image[(yc*arImXsize+xc)*3+2];
						ext_pat2[j/ydiv][i/xdiv][1] += image[(yc*arImXsize+xc)*3+1];
						ext_pat2[j/ydiv][i/xdiv][2] += image[(yc*arImXsize+xc)*3+0];
						break;

					case PIXEL_FORMAT_RGB565:
						jy=j/ydiv; ix=i/xdiv;
						col8 = getLUM8_from_RGB565(image16+yc*arImXsize+xc);
						ext_pat2[jy][ix][0] += col8;
						ext_pat2[jy][ix][1] += col8;
						ext_pat2[jy][ix][2] += col8;
						break;

					case PIXEL_FORMAT_LUM:
						jy=j/ydiv; ix=i/xdiv;
						col8 = image[yc*arImXsize+xc];
						ext_pat2[jy][ix][0] += col8;
						ext_pat2[jy][ix][1] += col8;
						ext_pat2[jy][ix][2] += col8;
						break;
					}
				}
			}
		}

		for( j = 0; j < PATTERN_HEIGHT; j++ ) {
			for( i = 0; i < PATTERN_HEIGHT; i++ ) {
				ext_pat[j][i][0] = ext_pat2[j][i][0] / (xdiv*ydiv);
				ext_pat[j][i][1] = ext_pat2[j][i][1] / (xdiv*ydiv);
				ext_pat[j][i][2] = ext_pat2[j][i][2] / (xdiv*ydiv);
			}
		}
	}

    return(0);
}
//#else
/*
int arGetPatt( ARUint8 *image, int *x_coord, int *y_coord, int *vertex,
               ARUint8 ext_pat[PATTERN_HEIGHT][PATTERN_WIDTH][3] )
{
    ARFloat  world[4][2];
    ARFloat  local[4][2];
    ARFloat  para[3][3];
    ARFloat  d, xw, yw;
    int     xc, yc;
    int     i, j;
    int     k1, k2, k3;

    world[0][0] = 100.0;
    world[0][1] = 100.0;
    world[1][0] = 100.0 + 10.0;
    world[1][1] = 100.0;
    world[2][0] = 100.0 + 10.0;
    world[2][1] = 100.0 + 10.0;
    world[3][0] = 100.0;
    world[3][1] = 100.0 + 10.0;
    for( i = 0; i < 4; i++ ) {
        local[i][0] = x_coord[vertex[i]];
        local[i][1] = y_coord[vertex[i]];
    }
    get_cpara( world, local, para );

    put_zero( (ARUint8 *)ext_pat, PATTERN_HEIGHT*PATTERN_WIDTH*3 );
    for( j = 0; j < AR_PATT_SAMPLE_NUM; j++ ) {
        yw = 102.5 + 5.0 * (j+0.5) / (ARFloat)AR_PATT_SAMPLE_NUM;
        for( i = 0; i < AR_PATT_SAMPLE_NUM; i++ ) {
            xw = 102.5 + 5.0 * (i+0.5) / (ARFloat)AR_PATT_SAMPLE_NUM;
            d = para[2][0]*xw + para[2][1]*yw + para[2][2];
            if( d == 0 ) return(-1);
            xc = (int)((para[0][0]*xw + para[0][1]*yw + para[0][2])/d);
            yc = (int)((para[1][0]*xw + para[1][1]*yw + para[1][2])/d);
            if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) {
                xc = ((xc+1)/2)*2;
                yc = ((yc+1)/2)*2;
            }
            if( xc >= 0 && xc < arImXsize && yc >= 0 && yc < arImYsize ) {
#ifdef AR_PIX_FORMAT_ABGR
                k1 = image[(yc*arImXsize+xc)*AR_PIX_SIZE+1];
                k1 = ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][0]
                   + k1*(PATTERN_HEIGHT*PATTERN_WIDTH)/(AR_PATT_SAMPLE_NUM*AR_PATT_SAMPLE_NUM);
                ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][0] = (k1 > 255)? 255: k1;
                k2 = image[(yc*arImXsize+xc)*AR_PIX_SIZE+2];
                k2 = ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][1]
                   + k2*(PATTERN_HEIGHT*PATTERN_WIDTH)/(AR_PATT_SAMPLE_NUM*AR_PATT_SAMPLE_NUM);
                ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][1] = (k2 > 255)? 255: k2;
                k3 = image[(yc*arImXsize+xc)*AR_PIX_SIZE+3];
                k3 = ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][2]
                   + k3*(PATTERN_HEIGHT*PATTERN_WIDTH)/(AR_PATT_SAMPLE_NUM*AR_PATT_SAMPLE_NUM);
                ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][2] = (k3 > 255)? 255: k3;
#endif
#ifdef AR_PIX_FORMAT_BGRA
                k1 = image[(yc*arImXsize+xc)*AR_PIX_SIZE+0];
                k1 = ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][0]
                   + k1*(PATTERN_HEIGHT*PATTERN_WIDTH)/(AR_PATT_SAMPLE_NUM*AR_PATT_SAMPLE_NUM);
                ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][0] = (k1 > 255)? 255: k1;
                k2 = image[(yc*arImXsize+xc)*AR_PIX_SIZE+1];
                k2 = ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][1]
                   + k2*(PATTERN_HEIGHT*PATTERN_WIDTH)/(AR_PATT_SAMPLE_NUM*AR_PATT_SAMPLE_NUM);
                ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][1] = (k2 > 255)? 255: k2;
                k3 = image[(yc*arImXsize+xc)*AR_PIX_SIZE+2];
                k3 = ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][2]
                   + k3*(PATTERN_HEIGHT*PATTERN_WIDTH)/(AR_PATT_SAMPLE_NUM*AR_PATT_SAMPLE_NUM);
                ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][2] = (k3 > 255)? 255: k3;
#endif
#ifdef AR_PIX_FORMAT_BGR
                k1 = image[(yc*arImXsize+xc)*AR_PIX_SIZE+0];
                k1 = ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][0]
                   + k1*(PATTERN_HEIGHT*PATTERN_WIDTH)/(AR_PATT_SAMPLE_NUM*AR_PATT_SAMPLE_NUM);
                ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][0] = (k1 > 255)? 255: k1;
                k2 = image[(yc*arImXsize+xc)*AR_PIX_SIZE+1];
                k2 = ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][1]
                   + k2*(PATTERN_HEIGHT*PATTERN_WIDTH)/(AR_PATT_SAMPLE_NUM*AR_PATT_SAMPLE_NUM);
                ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][1] = (k2 > 255)? 255: k2;
                k3 = image[(yc*arImXsize+xc)*AR_PIX_SIZE+2];
                k3 = ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][2]
                   + k3*(PATTERN_HEIGHT*PATTERN_WIDTH)/(AR_PATT_SAMPLE_NUM*AR_PATT_SAMPLE_NUM);
                ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][2] = (k3 > 255)? 255: k3;
#endif
#ifdef AR_PIX_FORMAT_RGBA
                k1 = image[(yc*arImXsize+xc)*AR_PIX_SIZE+2];
                k1 = ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][0]
                   + k1*(PATTERN_HEIGHT*PATTERN_WIDTH)/(AR_PATT_SAMPLE_NUM*AR_PATT_SAMPLE_NUM);
                ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][0] = (k1 > 255)? 255: k1;
                k2 = image[(yc*arImXsize+xc)*AR_PIX_SIZE+1];
                k2 = ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][1]
                   + k2*(PATTERN_HEIGHT*PATTERN_WIDTH)/(AR_PATT_SAMPLE_NUM*AR_PATT_SAMPLE_NUM);
                ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][1] = (k2 > 255)? 255: k2;
                k3 = image[(yc*arImXsize+xc)*AR_PIX_SIZE+0];
                k3 = ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][2]
                   + k3*(PATTERN_HEIGHT*PATTERN_WIDTH)/(AR_PATT_SAMPLE_NUM*AR_PATT_SAMPLE_NUM);
                ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][2] = (k3 > 255)? 255: k3;
#endif
#ifdef AR_PIX_FORMAT_RGB
                k1 = image[(yc*arImXsize+xc)*AR_PIX_SIZE+2];
                k1 = ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][0]
                   + k1*(PATTERN_HEIGHT*PATTERN_WIDTH)/(AR_PATT_SAMPLE_NUM*AR_PATT_SAMPLE_NUM);
                ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][0] = (k1 > 255)? 255: k1;
                k2 = image[(yc*arImXsize+xc)*AR_PIX_SIZE+1];
                k2 = ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][1]
                   + k2*(PATTERN_HEIGHT*PATTERN_WIDTH)/(AR_PATT_SAMPLE_NUM*AR_PATT_SAMPLE_NUM);
                ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][1] = (k2 > 255)? 255: k2;
                k3 = image[(yc*arImXsize+xc)*AR_PIX_SIZE+0];
                k3 = ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][2]
                   + k3*(PATTERN_HEIGHT*PATTERN_WIDTH)/(AR_PATT_SAMPLE_NUM*AR_PATT_SAMPLE_NUM);
                ext_pat[j*PATTERN_HEIGHT/AR_PATT_SAMPLE_NUM][i*PATTERN_WIDTH/AR_PATT_SAMPLE_NUM][2] = (k3 > 255)? 255: k3;
#endif
            }
        }
    }

    return(0);
}
//#endif
*/


AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::pattern_match( ARUint8 *data, int *code, int *dir, ARFloat *cf)
{
    ARFloat invec[EVEC_MAX];
    int    input[PATTERN_HEIGHT*PATTERN_WIDTH*3];
    int    i, j, l;
    int    k = 0; // fix VC7 compiler warning: uninitialized variable
    int    ave, sum, res, res2;
    ARFloat datapow, sum2, min;
    ARFloat max = 0.0; // fix VC7 compiler warning: uninitialized variable

	// uncomment to dump the unprojected content of the marker that artoolkit found in the image
/*		if(FILE* fp = fopen("dump.raw", "wb"))
		{
			fwrite(data, 1, PATTERN_HEIGHT*PATTERN_WIDTH*3, fp);
			fclose(fp);
		}*/

    sum = ave = 0;
    for(i=0;i<PATTERN_HEIGHT*PATTERN_WIDTH*3;i++) {
        ave += (255-data[i]);
    }
    ave /= (PATTERN_HEIGHT*PATTERN_WIDTH*3);

    if( arTemplateMatchingMode == AR_TEMPLATE_MATCHING_COLOR ) {
        for(i=0;i<PATTERN_HEIGHT*PATTERN_WIDTH*3;i++) {
            input[i] = (255-data[i]) - ave;
            sum += input[i]*input[i];
        }
    }
    else {
        for(i=0;i<PATTERN_HEIGHT*PATTERN_WIDTH;i++) {
            input[i] = ((255-data[i*3+0]) + (255-data[i*3+1]) + (255-data[i*3+02]))/3 - ave;
            sum += input[i]*input[i];
        }
    }

    datapow = (ARFloat)sqrt( (ARFloat)sum );
    if( datapow == 0.0 ) {
        *code = 0;
        *dir  = 0;
        *cf   = -1.0;
        return -1;
    }

    res = res2 = -1;
    if( arTemplateMatchingMode == AR_TEMPLATE_MATCHING_COLOR ) {
        if( arMatchingPCAMode == AR_MATCHING_WITH_PCA && evecf ) {

            for( i = 0; i < evec_dim; i++ ) {
                invec[i] = 0.0;
                for( j = 0; j < PATTERN_HEIGHT*PATTERN_WIDTH*3; j++ ) {
                    invec[i] += evec[i][j] * input[j];
                }
                invec[i] /= datapow;
            }

            min = 10000.0;
            k = -1;
            for( l = 0; l < pattern_num; l++ ) {
                k++;
                while( patf[k] == 0 ) k++;
                if( patf[k] == 2 ) continue;
#ifdef ARTK_DEBUG
                printf("%3d: ", k);
#endif
                for( j = 0; j < 4; j++ ) {
                    sum2 = 0;
                    for(i = 0; i < evec_dim; i++ ) {
                        sum2 += (invec[i] - epat[k][j][i]) * (invec[i] - epat[k][j][i]);
                    }
#ifdef ARTK_DEBUG
                    printf("%10.7f ", sum2);
#endif
                    if( sum2 < min ) { min = sum2; res = j; res2 = k; }
                }
#ifdef ARTK_DEBUG
                printf("\n");
#endif
            }
            sum = 0;
            for(i=0;i<PATTERN_HEIGHT*PATTERN_WIDTH*3;i++) sum += input[i]*pat[res2][res][i];
            max = sum / patpow[res2][res] / datapow;
        }
        else {
            k = -1;
            max = 0.0;
            for( l = 0; l < pattern_num; l++ ) {
                k++;
                while( patf[k] == 0 ) k++;
                if( patf[k] == 2 ) continue;
                for( j = 0; j < 4; j++ ) {
                    sum = 0;
                    for(i=0;i<PATTERN_HEIGHT*PATTERN_WIDTH*3;i++) sum += input[i]*pat[k][j][i];
                    sum2 = sum / patpow[k][j] / datapow;
                    if( sum2 > max ) { max = sum2; res = j; res2 = k; }
                }
            }
        }
    }
    else {
        for( l = 0; l < pattern_num; l++ ) {
            k++;
            while( patf[k] == 0 ) k++;
            if( patf[k] == 2 ) continue;
            for( j = 0; j < 4; j++ ) {
                sum = 0;
                for(i=0;i<PATTERN_HEIGHT*PATTERN_WIDTH;i++) sum += input[i]*patBW[k][j][i];
                sum2 = sum / patpowBW[k][j] / datapow;
                if( sum2 > max ) { max = sum2; res = j; res2 = k; }
            }
        }
    }

    *code = res2;
    *dir  = res;
    *cf   = max;

#ifdef ARTK_DEBUG
    printf("%d %d %f\n", res2, res, max);
#endif

    return 0;
}


AR_TEMPL_FUNC void
AR_TEMPL_TRACKER::gen_evec(void)
{
    int    i, j, k, ii, jj;
    ARMat  *input, *wevec;
    ARVec  *wev;
    ARFloat sum, sum2;
    int    dim;

    if( pattern_num < 4 ) {
        evecf   = 0;
        evecBWf = 0;
        return;
    }

#ifdef ARTK_DEBUG
    printf("------------------------------------------\n");
#endif

    dim = (pattern_num*4 < PATTERN_HEIGHT*PATTERN_WIDTH*3)? pattern_num*4: PATTERN_HEIGHT*PATTERN_WIDTH*3;
    input  = Matrix::alloc( pattern_num*4, PATTERN_HEIGHT*PATTERN_WIDTH*3 );
    wevec   = Matrix::alloc( dim, PATTERN_HEIGHT*PATTERN_WIDTH*3 );
    wev     = Vector::alloc( dim );

    for( j = jj = 0; jj < MAX_LOAD_PATTERNS; jj++ ) {
        if( patf[jj] == 0 ) continue;
        for( k = 0; k < 4; k++ ) {
            for( i = 0; i < PATTERN_HEIGHT*PATTERN_WIDTH*3; i++ ) {
                input->m[(j*4+k)*PATTERN_HEIGHT*PATTERN_WIDTH*3+i] = pat[j][k][i] / patpow[j][k];
            }
        }
        j++;
    }

    if( arMatrixPCA2(input, wevec, wev) < 0 ) {
        Matrix::free( input );
        Matrix::free( wevec );
        Vector::free( wev );
        evecf   = 0;
        evecBWf = 0;
        return;
    }

    sum = 0.0;
    for( i = 0; i < dim; i++ ) {
        sum += wev->v[i];
#ifdef ARTK_DEBUG
        printf("%2d(%10.7f): \n", i+1, sum);
#endif
        if( sum > 0.90 ) break;
        if( i == EVEC_MAX-1 ) break;
    }
    evec_dim = i+1;

    for( j = 0; j < evec_dim; j++ ) {
        for( i = 0; i < PATTERN_HEIGHT*PATTERN_WIDTH*3; i++ ) {
            evec[j][i] = wevec->m[j*PATTERN_HEIGHT*PATTERN_WIDTH*3+i];
        }
    }
    
    for( i = 0; i < MAX_LOAD_PATTERNS; i++ ) {
        if(patf[i] == 0) continue;
        for( j = 0; j < 4; j++ ) {
#ifdef ARTK_DEBUG
            printf("%2d[%d]: ", i+1, j+1);
#endif
            sum2 = 0.0;
            for( k = 0; k < evec_dim; k++ ) {
                sum = 0.0;
                for(ii=0;ii<PATTERN_HEIGHT*PATTERN_WIDTH*3;ii++) {
                    sum += evec[k][ii] * pat[i][j][ii] / patpow[i][j];
                }
#ifdef ARTK_DEBUG
                printf("%10.7f ", sum);
#endif
                epat[i][j][k] = sum;
                sum2 += sum*sum;
            }
#ifdef ARTK_DEBUG
            printf(":: %10.7f\n", sqrt(sum2));
#endif
        }
    }

    Matrix::free( input );
    Matrix::free( wevec );
    Vector::free( wev );

    evecf   = 1;
    evecBWf = 0;

    return;
}


static void
get_cpara( ARFloat world[4][2], ARFloat vertex[4][2], ARFloat para[3][3] )
{
    ARMat   *a, *b, *c;
    int     i;

    a = Matrix::alloc( 8, 8 );
    b = Matrix::alloc( 8, 1 );
    c = Matrix::alloc( 8, 1 );
    for( i = 0; i < 4; i++ ) {
        a->m[i*16+0]  = world[i][0];
        a->m[i*16+1]  = world[i][1];
        a->m[i*16+2]  = 1.0;
        a->m[i*16+3]  = 0.0;
        a->m[i*16+4]  = 0.0;
        a->m[i*16+5]  = 0.0;
        a->m[i*16+6]  = -world[i][0] * vertex[i][0];
        a->m[i*16+7]  = -world[i][1] * vertex[i][0];
        a->m[i*16+8]  = 0.0;
        a->m[i*16+9]  = 0.0;
        a->m[i*16+10] = 0.0;
        a->m[i*16+11] = world[i][0];
        a->m[i*16+12] = world[i][1];
        a->m[i*16+13] = 1.0;
        a->m[i*16+14] = -world[i][0] * vertex[i][1];
        a->m[i*16+15] = -world[i][1] * vertex[i][1];
        b->m[i*2+0] = vertex[i][0];
        b->m[i*2+1] = vertex[i][1];
    }
    Matrix::selfInv( a );
    Matrix::mul( c, a, b );
    for( i = 0; i < 2; i++ ) {
        para[i][0] = c->m[i*3+0];
        para[i][1] = c->m[i*3+1];
        para[i][2] = c->m[i*3+2];
    }
    para[2][0] = c->m[2*3+0];
    para[2][1] = c->m[2*3+1];
    para[2][2] = 1.0;
    Matrix::free( a );
    Matrix::free( b );
    Matrix::free( c );
}


/*
static void   put_zero( ARUint8 *p, int size )
{
    while( (size--) > 0 ) *(p++) = 0;
}
*/

}  // namespace ARToolKitPlus
// End of ../../src/core/arGetCode.cxx
// Included ../../src/core/arGetMarkerInfo.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id: arGetMarkerInfo.cxx 162 2006-04-19 21:28:10Z grabner $
 * @file
 * ======================================================================== */


#include <ARToolKitPlus/Tracker.h>


namespace ARToolKitPlus {


AR_TEMPL_FUNC ARMarkerInfo*
AR_TEMPL_TRACKER::arGetMarkerInfo(ARUint8 *image, ARMarkerInfo2 *marker_info2, int *marker_num, int thresh)
{
    int            id, dir;
    ARFloat         cf;
    int            i, j;

	PROFILE_BEGINSEC(profiler, GETMARKERINFO)

    for( i = j = 0; i < *marker_num; i++ ) {
        marker_infoL[j].area   = marker_info2[i].area;
        marker_infoL[j].pos[0] = marker_info2[i].pos[0];
        marker_infoL[j].pos[1] = marker_info2[i].pos[1];

        if( arGetLine(marker_info2[i].x_coord, marker_info2[i].y_coord,
                      marker_info2[i].coord_num, marker_info2[i].vertex,
                      marker_infoL[j].line, marker_infoL[j].vertex) < 0 ) continue;

        arGetCode( image,
                   marker_info2[i].x_coord, marker_info2[i].y_coord,
                   marker_info2[i].vertex, &id, &dir, &cf, thresh);

        marker_infoL[j].id  = id;
        marker_infoL[j].dir = dir;
        marker_infoL[j].cf  = cf;

        j++;
    }
    *marker_num = j;

	PROFILE_ENDSEC(profiler, GETMARKERINFO)

    return( marker_infoL );
}


}  // namespace ARToolKitPlus
// End of ../../src/core/arGetMarkerInfo.cxx
// Included ../../src/core/arGetTransMat.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id: arGetTransMat.cxx 162 2006-04-19 21:28:10Z grabner $
 * @file
 * ======================================================================== */


#include <stdlib.h>
#include <math.h>

#include <ARToolKitPlus/Tracker.h>
#include <ARToolKitPlus/matrix.h>


namespace ARToolKitPlus {


AR_TEMPL_FUNC ARFloat
AR_TEMPL_TRACKER::arGetTransMat(ARMarkerInfo *marker_info, ARFloat center[2], ARFloat width, ARFloat conv[3][4])
{
    ARFloat  rot[3][3];
    ARFloat  ppos2d[4][2];
    ARFloat  ppos3d[4][2];
    int     dir;
    ARFloat  err;
    int     i;

	PROFILE_BEGINSEC(profiler, GETTRANSMAT)

	if( arGetInitRot( marker_info, arCamera->mat, rot ) < 0 )
	{
		PROFILE_ENDSEC(profiler, GETTRANSMAT)
		return -1;
	}

    dir = marker_info->dir;
    ppos2d[0][0] = marker_info->vertex[(4-dir)%4][0];
    ppos2d[0][1] = marker_info->vertex[(4-dir)%4][1];
    ppos2d[1][0] = marker_info->vertex[(5-dir)%4][0];
    ppos2d[1][1] = marker_info->vertex[(5-dir)%4][1];
    ppos2d[2][0] = marker_info->vertex[(6-dir)%4][0];
    ppos2d[2][1] = marker_info->vertex[(6-dir)%4][1];
    ppos2d[3][0] = marker_info->vertex[(7-dir)%4][0];
    ppos2d[3][1] = marker_info->vertex[(7-dir)%4][1];
    ppos3d[0][0] = center[0] - width*(ARFloat)0.5;
    ppos3d[0][1] = center[1] + width*(ARFloat)0.5;
    ppos3d[1][0] = center[0] + width*(ARFloat)0.5;
    ppos3d[1][1] = center[1] + width*(ARFloat)0.5;
    ppos3d[2][0] = center[0] + width*(ARFloat)0.5;
    ppos3d[2][1] = center[1] - width*(ARFloat)0.5;
    ppos3d[3][0] = center[0] - width*(ARFloat)0.5;
    ppos3d[3][1] = center[1] - width*(ARFloat)0.5;

    for( i = 0; i < AR_GET_TRANS_MAT_MAX_LOOP_COUNT; i++ ) {
		err = arGetTransMat3( rot, ppos2d, ppos3d, 4, conv, arCamera);
        if( err < AR_GET_TRANS_MAT_MAX_FIT_ERROR ) break;
    }

	PROFILE_ENDSEC(profiler, GETTRANSMAT)
    return err;
}

AR_TEMPL_FUNC ARFloat
AR_TEMPL_TRACKER::arGetTransMat2(ARFloat rot[3][3], ARFloat ppos2d[][2], ARFloat ppos3d[][2], int num, ARFloat conv[3][4])
{
	return arGetTransMat3( rot, ppos2d, ppos3d, num, conv, arCamera);
}


AR_TEMPL_FUNC ARFloat
AR_TEMPL_TRACKER::arGetTransMat3(ARFloat rot[3][3], ARFloat ppos2d[][2],
                   ARFloat ppos3d[][2], int num, ARFloat conv[3][4],
				   Camera *pCam )
                   //ARFloat *dist_factor, ARFloat cpara[3][4] )
{
    ARFloat  off[3], pmax[3], pmin[3];
    ARFloat  ret;
    int     i;

	PROFILE_BEGINSEC(profiler, GETTRANSMAT3)

    pmax[0]=pmax[1]=pmax[2] = -10000000000.0;
    pmin[0]=pmin[1]=pmin[2] =  10000000000.0;
    for( i = 0; i < num; i++ ) {
        if( ppos3d[i][0] > pmax[0] ) pmax[0] = ppos3d[i][0];
        if( ppos3d[i][0] < pmin[0] ) pmin[0] = ppos3d[i][0];
        if( ppos3d[i][1] > pmax[1] ) pmax[1] = ppos3d[i][1];
        if( ppos3d[i][1] < pmin[1] ) pmin[1] = ppos3d[i][1];
/*
        if( ppos3d[i][2] > pmax[2] ) pmax[2] = ppos3d[i][2];
        if( ppos3d[i][2] < pmin[2] ) pmin[2] = ppos3d[i][2];
*/
    }
    off[0] = -(pmax[0] + pmin[0])  * (ARFloat)0.5;
    off[1] = -(pmax[1] + pmin[1])  * (ARFloat)0.5;
    off[2] = -(pmax[2] + pmin[2])  * (ARFloat)0.5;
    for( i = 0; i < num; i++ ) {
        pos3d[i][0] = ppos3d[i][0] + off[0];
        pos3d[i][1] = ppos3d[i][1] + off[1];
/*
        pos3d[i][2] = ppos3d[i][2] + off[2];
*/
        pos3d[i][2] = 0.0;
    }

    ret = arGetTransMatSub( rot, ppos2d, pos3d, num, conv, pCam);
                            //dist_factor, cpara );

    conv[0][3] = conv[0][0]*off[0] + conv[0][1]*off[1] + conv[0][2]*off[2] + conv[0][3];
    conv[1][3] = conv[1][0]*off[0] + conv[1][1]*off[1] + conv[1][2]*off[2] + conv[1][3];
    conv[2][3] = conv[2][0]*off[0] + conv[2][1]*off[1] + conv[2][2]*off[2] + conv[2][3];

	PROFILE_ENDSEC(profiler, GETTRANSMAT3)
    return ret;
}

AR_TEMPL_FUNC ARFloat
AR_TEMPL_TRACKER::arGetTransMat4(ARFloat rot[3][3], ARFloat ppos2d[][2], ARFloat ppos3d[][3], int num, ARFloat conv[3][4])
{
    return arGetTransMat5( rot, ppos2d, ppos3d, num, conv, arCamera);
//                           arParam.dist_factor, arParam.mat );
}


AR_TEMPL_FUNC ARFloat
AR_TEMPL_TRACKER::arGetTransMat5(ARFloat rot[3][3], ARFloat ppos2d[][2],
				   ARFloat ppos3d[][3], int num, ARFloat conv[3][4],
				   Camera *pCam)
				   //ARFloat *dist_factor, ARFloat cpara[3][4])
{
    ARFloat  off[3], pmax[3], pmin[3];
    ARFloat  ret;
    int     i;

    pmax[0]=pmax[1]=pmax[2] = -10000000000.0;
    pmin[0]=pmin[1]=pmin[2] =  10000000000.0;
    for( i = 0; i < num; i++ ) {
        if( ppos3d[i][0] > pmax[0] ) pmax[0] = ppos3d[i][0];
        if( ppos3d[i][0] < pmin[0] ) pmin[0] = ppos3d[i][0];
        if( ppos3d[i][1] > pmax[1] ) pmax[1] = ppos3d[i][1];
        if( ppos3d[i][1] < pmin[1] ) pmin[1] = ppos3d[i][1];
        if( ppos3d[i][2] > pmax[2] ) pmax[2] = ppos3d[i][2];
        if( ppos3d[i][2] < pmin[2] ) pmin[2] = ppos3d[i][2];
    }
    off[0] = -(pmax[0] + pmin[0])  * (ARFloat)0.5;
    off[1] = -(pmax[1] + pmin[1])  * (ARFloat)0.5;
    off[2] = -(pmax[2] + pmin[2])  * (ARFloat)0.5;
    for( i = 0; i < num; i++ ) {
        pos3d[i][0] = ppos3d[i][0] + off[0];
        pos3d[i][1] = ppos3d[i][1] + off[1];
        pos3d[i][2] = ppos3d[i][2] + off[2];
    }

    ret = arGetTransMatSub( rot, ppos2d, pos3d, num, conv, pCam);
                            //dist_factor, cpara );

    conv[0][3] = conv[0][0]*off[0] + conv[0][1]*off[1] + conv[0][2]*off[2] + conv[0][3];
    conv[1][3] = conv[1][0]*off[0] + conv[1][1]*off[1] + conv[1][2]*off[2] + conv[1][3];
    conv[2][3] = conv[2][0]*off[0] + conv[2][1]*off[1] + conv[2][2]*off[2] + conv[2][3];

    return ret;
}

AR_TEMPL_FUNC ARFloat
AR_TEMPL_TRACKER::arGetTransMatSub(ARFloat rot[3][3], ARFloat ppos2d[][2],
                     ARFloat pos3d[][3], int num, ARFloat conv[3][4],
					 Camera *pCam )
                     //ARFloat *dist_factor, ARFloat cpara[3][4] )
{
    ARMat   *mat_a, *mat_b, *mat_c, *mat_d, *mat_e, *mat_f;
    ARFloat  trans[3];
    ARFloat  wx, wy, wz;
    ARFloat  ret;
    int     i, j;

	PROFILE_BEGINSEC(profiler, GETTRANSMATSUB)

    mat_a = Matrix::alloc( num*2, 3 );
    mat_b = Matrix::alloc( 3, num*2 );
    mat_c = Matrix::alloc( num*2, 1 );
    mat_d = Matrix::alloc( 3, 3 );
    mat_e = Matrix::alloc( 3, 1 );
    mat_f = Matrix::alloc( 3, 1 );

    if( arFittingMode == AR_FITTING_TO_INPUT ) {
        for( i = 0; i < num; i++ ) {
            arParamIdeal2Observ_std(pCam, ppos2d[i][0], ppos2d[i][1], &pos2d[i][0], &pos2d[i][1]);
        }
    }
    else {
        for( i = 0; i < num; i++ ) {
            pos2d[i][0] = ppos2d[i][0];
            pos2d[i][1] = ppos2d[i][1];
        }
    }

    for( j = 0; j < num; j++ ) {
        wx = rot[0][0] * pos3d[j][0]
           + rot[0][1] * pos3d[j][1]
           + rot[0][2] * pos3d[j][2];
        wy = rot[1][0] * pos3d[j][0]
           + rot[1][1] * pos3d[j][1]
           + rot[1][2] * pos3d[j][2];
        wz = rot[2][0] * pos3d[j][0]
           + rot[2][1] * pos3d[j][1]
           + rot[2][2] * pos3d[j][2];
        mat_a->m[j*6+0] = mat_b->m[num*0+j*2] = pCam->mat[0][0];
        mat_a->m[j*6+1] = mat_b->m[num*2+j*2] = pCam->mat[0][1];
        mat_a->m[j*6+2] = mat_b->m[num*4+j*2] = pCam->mat[0][2] - pos2d[j][0];
        mat_c->m[j*2+0] = wz * pos2d[j][0]
               - pCam->mat[0][0]*wx - pCam->mat[0][1]*wy - pCam->mat[0][2]*wz;
        mat_a->m[j*6+3] = mat_b->m[num*0+j*2+1] = 0.0;
        mat_a->m[j*6+4] = mat_b->m[num*2+j*2+1] = pCam->mat[1][1];
        mat_a->m[j*6+5] = mat_b->m[num*4+j*2+1] = pCam->mat[1][2] - pos2d[j][1];
        mat_c->m[j*2+1] = wz * pos2d[j][1]
               - pCam->mat[1][1]*wy - pCam->mat[1][2]*wz;
    }
    Matrix::mul( mat_d, mat_b, mat_a );
    Matrix::mul( mat_e, mat_b, mat_c );
    Matrix::selfInv( mat_d );
    Matrix::mul( mat_f, mat_d, mat_e );
    trans[0] = mat_f->m[0];
    trans[1] = mat_f->m[1];
    trans[2] = mat_f->m[2];

	/*trans[0] = 3.96559f;
	trans[1] = 27.0546f;
	trans[2] = 274.627f;*/

	{
		ARFloat a,b,c;
		arGetAngle( rot, &a, &b, &c );

		//trans[0] = -13.5f;
		//trans[1] = 45.7f;
		//trans[2] = 303.0f;
		//arGetRot( -90.5f*3.1415f/180.0f, 120.3f*3.1415f/180.0f, 31.2f*3.1415f/180.0f, rot );

		ret = arModifyMatrix( rot, trans, pCam->mat, pos3d, pos2d, num );

		arGetAngle( rot, &a, &b, &c );
		a=a;
	}

	// double begin
	//
    /*for( j = 0; j < num; j++ ) {
        wx = rot[0][0] * pos3d[j][0]
           + rot[0][1] * pos3d[j][1]
           + rot[0][2] * pos3d[j][2];
        wy = rot[1][0] * pos3d[j][0]
           + rot[1][1] * pos3d[j][1]
           + rot[1][2] * pos3d[j][2];
        wz = rot[2][0] * pos3d[j][0]
           + rot[2][1] * pos3d[j][1]
           + rot[2][2] * pos3d[j][2];
        mat_a->m[j*6+0] = mat_b->m[num*0+j*2] = pCam->mat[0][0];
        mat_a->m[j*6+1] = mat_b->m[num*2+j*2] = pCam->mat[0][1];
        mat_a->m[j*6+2] = mat_b->m[num*4+j*2] = pCam->mat[0][2] - pos2d[j][0];
        mat_c->m[j*2+0] = wz * pos2d[j][0]
               - pCam->mat[0][0]*wx - pCam->mat[0][1]*wy - pCam->mat[0][2]*wz;
        mat_a->m[j*6+3] = mat_b->m[num*0+j*2+1] = 0.0;
        mat_a->m[j*6+4] = mat_b->m[num*2+j*2+1] = pCam->mat[1][1];
        mat_a->m[j*6+5] = mat_b->m[num*4+j*2+1] = pCam->mat[1][2] - pos2d[j][1];
        mat_c->m[j*2+1] = wz * pos2d[j][1]
               - pCam->mat[1][1]*wy - pCam->mat[1][2]*wz;
    }
    Matrix::mul( mat_d, mat_b, mat_a );
    Matrix::mul( mat_e, mat_b, mat_c );
    Matrix::selfInv( mat_d );
    Matrix::mul( mat_f, mat_d, mat_e );
    trans[0] = mat_f->m[0];
    trans[1] = mat_f->m[1];
    trans[2] = mat_f->m[2];

    ret = arModifyMatrix( rot, trans, pCam->mat, pos3d, pos2d, num );*/
	//
	// double end

    Matrix::free( mat_a );
    Matrix::free( mat_b );
    Matrix::free( mat_c );
    Matrix::free( mat_d );
    Matrix::free( mat_e );
    Matrix::free( mat_f );

    for( j = 0; j < 3; j++ ) {
        for( i = 0; i < 3; i++ ) conv[j][i] = rot[j][i];
        conv[j][3] = trans[j];
    }

	PROFILE_ENDSEC(profiler, GETTRANSMATSUB)
    return ret;
}


}  // namespace ARToolKitPlus
// End of ../../src/core/arGetTransMat.cxx
// Included ../../src/core/arGetTransMat2.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id: arGetTransMat2.cxx 162 2006-04-19 21:28:10Z grabner $
 * @file
 * ======================================================================== */


//#include <stdlib.h>
//#include <math.h>

//#include <ARToolKitPlus/Tracker.h>
//#include <ARToolKitPlus/matrix.h>
//#include <ARToolKitPlus/extra/Profiler.h>


namespace ARToolKitPlus {


#define MD_PI         3.14159265358979323846


#ifndef _FIXEDPOINT_MATH_ACTIVATED_


AR_TEMPL_FUNC ARFloat
AR_TEMPL_TRACKER::arModifyMatrix(ARFloat rot[3][3], ARFloat trans[3], ARFloat cpara[3][4],
				   ARFloat vertex[][3], ARFloat pos2d[][2], int num)
{
    ARFloat    factor;
    ARFloat    a, b, c;
    ARFloat    a1, b1, c1;
    ARFloat    a2, b2, c2;
    ARFloat    ma = 0, mb = 0, mc = 0;
    ARFloat    combo[3][4];
    ARFloat    hx, hy, h, x, y;
    ARFloat    err, minerr;
    int        t1, t2, t3;
    int       s1 = 0, s2 = 0, s3 = 0;
    int       i, j;

	PROFILE_BEGINSEC(profiler, MODIFYMATRIX)

    arGetAngle( rot, &a, &b, &c );

    a2 = a;
    b2 = b;
    c2 = c;
    factor = (ARFloat)(10.0*MD_PI/180.0);
    for( j = 0; j < 15; j++ ) {
        minerr = 1000000000.0;
        for(t1=-1;t1<=1;t1++) {
        for(t2=-1;t2<=1;t2++) {
        for(t3=-1;t3<=1;t3++) {

            a1 = a2 + factor*t1;
            b1 = b2 + factor*t2;
            c1 = c2 + factor*t3;

            arGetNewMatrix( a1, b1, c1, trans, NULL, cpara, combo );

            err = 0.0;
            for( i = 0; i < num; i++ ) {
                hx = combo[0][0] * vertex[i][0]
                   + combo[0][1] * vertex[i][1]
                   + combo[0][2] * vertex[i][2]
                   + combo[0][3];
                hy = combo[1][0] * vertex[i][0]
                   + combo[1][1] * vertex[i][1]
                   + combo[1][2] * vertex[i][2]
                   + combo[1][3];
                h  = combo[2][0] * vertex[i][0]
                   + combo[2][1] * vertex[i][1]
                   + combo[2][2] * vertex[i][2]
                   + combo[2][3];
                x = hx / h;
                y = hy / h;

                err += (pos2d[i][0] - x) * (pos2d[i][0] - x)
                     + (pos2d[i][1] - y) * (pos2d[i][1] - y);
            }

            if( err < minerr ) {
                minerr = err;
                ma = a1;
                mb = b1;
                mc = c1;
                s1 = t1; s2 = t2; s3 = t3;
            }
        }
        }
        }

        if( s1 == 0 && s2 == 0 && s3 == 0 /*&& ss1 == 0 && ss2 == 0 && ss3 == 0*/ ) factor *= 0.5;
        a2 = ma;
        b2 = mb;
        c2 = mc;
    }

    arGetRot( ma, mb, mc, rot );

	PROFILE_ENDSEC(profiler, MODIFYMATRIX)

    return minerr/num;
}


AR_TEMPL_FUNC ARFloat
AR_TEMPL_TRACKER::arModifyMatrix2(ARFloat rot[3][3], ARFloat trans[3], ARFloat cpara[3][4],
					ARFloat vertex[][3], ARFloat pos2d[][2], int num)
{
    ARFloat    factor;
    ARFloat    a, b, c;
    ARFloat    a1, b1, c1;
    ARFloat    a2, b2, c2;
    ARFloat    ma, mb, mc;
    ARFloat    combo[3][4];
    ARFloat    hx, hy, h, x, y;
    ARFloat    err, minerr;
    int        t1, t2, t3, tt1,tt2,tt3;
	ARFloat	   tfact[5] = { 0.96f, 0.98f, 1.0f, 1.02f, 1.04f };
	ARFloat	   modtrans[3], mmodtrans[3];
    int       s1, s2, s3, ss1,ss2,ss3;
    int       i, j;

    arGetAngle( rot, &a, &b, &c );

    a2 = a;
    b2 = b;
    c2 = c;
    factor = (ARFloat)(40.0*MD_PI/180.0);
    for( j = 0; j < 15; j++ ) {
        minerr = 1000000000.0;
        for(t1=-1;t1<=1;t1++) {
        for(t2=-1;t2<=1;t2++) {
        for(t3=-1;t3<=1;t3++) {
        for(tt1=-2;tt1<=2;tt1++) {
        for(tt2=-2;tt2<=2;tt2++) {
        for(tt3=-2;tt3<=2;tt3++) {

            a1 = a2 + factor*t1;
            b1 = b2 + factor*t2;
            c1 = c2 + factor*t3;
			modtrans[0] = trans[0]*tfact[tt1+2];
			modtrans[1] = trans[1]*tfact[tt2+2];
			modtrans[2] = trans[2]*tfact[tt3+2];

            arGetNewMatrix( a1, b1, c1, modtrans, NULL, cpara, combo );

            err = 0.0;
            for( i = 0; i < num; i++ ) {
                hx = combo[0][0] * vertex[i][0]
                   + combo[0][1] * vertex[i][1]
                   + combo[0][2] * vertex[i][2]
                   + combo[0][3];
                hy = combo[1][0] * vertex[i][0]
                   + combo[1][1] * vertex[i][1]
                   + combo[1][2] * vertex[i][2]
                   + combo[1][3];
                h  = combo[2][0] * vertex[i][0]
                   + combo[2][1] * vertex[i][1]
                   + combo[2][2] * vertex[i][2]
                   + combo[2][3];
                x = hx / h;
                y = hy / h;

                err += (pos2d[i][0] - x) * (pos2d[i][0] - x)
                     + (pos2d[i][1] - y) * (pos2d[i][1] - y);
            }

            if( err < minerr ) {
                minerr = err;
                ma = a1;
                mb = b1;
                mc = c1;
				mmodtrans[0] = modtrans[0];
				mmodtrans[1] = modtrans[1];
				mmodtrans[2] = modtrans[2];

                s1 = t1; s2 = t2; s3 = t3;
				ss1 = tt1; ss2 = tt2; ss3 = tt3;
            }
        }
        }
        }
		}
		}
		}

        if( s1 == 0 && s2 == 0 && s3 == 0 /*&& ss1 == 0 && ss2 == 0 && ss3 == 0*/ ) factor *= 0.5;
        a2 = ma;
        b2 = mb;
        c2 = mc;
		trans[0] = mmodtrans[0];
		trans[1] = mmodtrans[1];
		trans[2] = mmodtrans[2];
    }

    arGetRot( ma, mb, mc, rot );

    return minerr/num;
}


#else //_FIXEDPOINT_MATH_ACTIVATED_


//////////////////////////////////////////////////////////////
//
//             FIXEDPOINT specific code starts here
//
//

#define BITS 8


int arGetNewMatrix12(I32 _a, I32 _b, I32 _c, FIXED_VEC3D _trans, ARFloat trans2[3][4],
					 FIXED_VEC3D _cpara[3], I32 _cpara3[3], FIXED_VEC3D ret[3], I32 _ret3[3], Profiler& nProfiler);



AR_TEMPL_FUNC ARFloat
AR_TEMPL_TRACKER::arModifyMatrix(ARFloat rot[3][3], ARFloat trans[3], ARFloat cpara[3][4],
                   ARFloat vertex[][3], ARFloat pos2d[][2], int num)
{
    //ARFloat		factor;
    ARFloat		a, b, c;
    //ARFloat		a1, b1, c1;
    ARFloat		a2, b2, c2;
    ARFloat		ma, mb, mc;
    int			t1, t2, t3;
    int			s1, s2, s3;
    int			i, j, k;
	ARFloat		minerr;

    I32    _hx, _hy, _h;
	I32    _err, _minerr;
	U32    _ures;
	I32	   _a1,_b1,_c1;
	I32	   _a2,_b2,_c2;
	I32		_ma, _mb, _mc;

	PROFILE_BEGINSEC(profiler, MODIFYMATRIX)

	FIXED_VEC3D	*_vertex = (FIXED_VEC3D*)malloc(num*sizeof(FIXED_VEC3D)),
				*_pos2d = (FIXED_VEC3D*)malloc(num*sizeof(FIXED_VEC3D)),
				_combo[3], _vec1, _vec2, _trans;
	I32			_combo3[3];

	FIXED_VEC3D	_cpara[3];
	I32			_cpara3[3];

	_vec1.z = 0;

	_trans.x = FIXED_Float_To_Fixed_n(trans[0], 12);
	_trans.y = FIXED_Float_To_Fixed_n(trans[1], 12);
	_trans.z = FIXED_Float_To_Fixed_n(trans[2], 12);

	for(j=0; j<3; j++)
	{
		_cpara[j].x = FIXED_Float_To_Fixed_n(cpara[j][0], 12);
		_cpara[j].y = FIXED_Float_To_Fixed_n(cpara[j][1], 12);
		_cpara[j].z = FIXED_Float_To_Fixed_n(cpara[j][2], 12);
		_cpara3[j] =  FIXED_Float_To_Fixed_n(cpara[j][3], 12);
	}

	for(j=0; j<num; j++)
	{
		_vertex[j].x = FIXED_Float_To_Fixed_n(vertex[j][0], BITS);
		_vertex[j].y = FIXED_Float_To_Fixed_n(vertex[j][1], BITS);
		_vertex[j].z = FIXED_Float_To_Fixed_n(vertex[j][2], BITS);

		_pos2d[j].x = FIXED_Float_To_Fixed_n(pos2d[j][0], BITS);
		_pos2d[j].y = FIXED_Float_To_Fixed_n(pos2d[j][1], BITS);
		_pos2d[j].z = 0;
	}


    arGetAngle( rot, &a, &b, &c );

	PROFILE_BEGINSEC(profiler, MODIFYMATRIX_LOOP)

    a2 = a;
    b2 = b;
    c2 = c;
    //factor = (ARFloat)(10.0*MD_PI/180.0);

	//I32 fix_a2, fix_b2, fix_c2;
	I32 fix_factor = FIXED_Float_To_Fixed_n((10.0*MD_PI/180.0), 12);
	I32 fix_a[3], fix_b[3], fix_c[3];

	_a2 = FIXED_Float_To_Fixed_n(a2, 12);
	_b2 = FIXED_Float_To_Fixed_n(b2, 12);
	_c2 = FIXED_Float_To_Fixed_n(c2, 12);

    for( j = 0; j < 10; j++ ) {
		_minerr = 0x40000000;		// value

		fix_a[0] = _a2 - fix_factor;   fix_a[1] = _a2;   fix_a[2] = _a2 + fix_factor;
		fix_b[0] = _b2 - fix_factor;   fix_b[1] = _b2;   fix_b[2] = _b2 + fix_factor;
		fix_c[0] = _c2 - fix_factor;   fix_c[1] = _c2;   fix_c[2] = _c2 + fix_factor;

        for(t1=-1;t1<=1;t1++) {
        for(t2=-1;t2<=1;t2++) {
        for(t3=-1;t3<=1;t3++) {
            //a1 = a2 + factor*t1;
            //b1 = b2 + factor*t2;
            //c1 = c2 + factor*t3;

			_a1 = fix_a[t1+1];
			_b1 = fix_b[t2+1];
			_c1 = fix_c[t3+1];

			PROFILE_BEGINSEC(profiler, GETNEWMATRIX)
            arGetNewMatrix12(_a1, _b1, _c1, _trans, NULL, _cpara, _cpara3, _combo, _combo3, profiler);
			PROFILE_ENDSEC(profiler, GETNEWMATRIX)

			for(k=0; k<3; k++)
			{
				_combo[k].x >>= 4;
				_combo[k].y >>= 4;
				_combo[k].z >>= 4;
				_combo3[k] >>= 4;
			}

            _err = 0;
            for( i = 0; i < num; i++ ) {
				FIXED_VEC3_DOT(_combo+0, _vertex+i, &_hx, BITS);
				_hx += _combo3[0];

				FIXED_VEC3_DOT(_combo+1, _vertex+i, &_hy, BITS);
				_hy += _combo3[1];

				FIXED_VEC3_DOT(_combo+2, _vertex+i, &_h, BITS);
				_h += _combo3[2];

				FIXED_DIV2(_hx, _h, _vec1.x, BITS);
				FIXED_DIV2(_hy, _h, _vec1.y, BITS);

				FIXED_VEC3_SUB(_pos2d+i, &_vec1, &_vec2);
				FIXED_VEC3_LENGTH_SQ(&_vec2, &_ures, BITS);
				_err += _ures;
            }

            if( _err < _minerr ) {
                _minerr = _err;
                //ma = a1;
                //mb = b1;
                //mc = c1;
				_ma = _a1;
				_mb = _b1;
				_mc = _c1;
                s1 = t1; s2 = t2; s3 = t3;
            }
        }
        }
        }

        if(s1 == 0 && s2 == 0 && s3 == 0)
		{
			//factor *= (ARFloat)0.5;
			fix_factor >>= 1;
		}
        //a2 = ma;
        //b2 = mb;
        //c2 = mc;
		_a2 = _ma;
		_b2 = _mb;
		_c2 = _mc;
    }

	PROFILE_ENDSEC(profiler, MODIFYMATRIX_LOOP)

	free(_vertex);
	free(_pos2d);

	ma = FIXED_Fixed_n_To_Float(_ma, 12);
	mb = FIXED_Fixed_n_To_Float(_mb, 12);
	mc = FIXED_Fixed_n_To_Float(_mc, 12);

    arGetRot( ma, mb, mc, rot );

	minerr = FIXED_Fixed_n_To_Float(_minerr, BITS);

	PROFILE_ENDSEC(profiler, MODIFYMATRIX)

    return minerr/num;
}

/*
AR_TEMPL_FUNC ARFloat
AR_TEMPL_TRACKER::arModifyMatrix(ARFloat rot[3][3], ARFloat trans[3], ARFloat cpara[3][4],
								 ARFloat vertex[][3], ARFloat pos2d[][2], int num)
{
	ARFloat		factor;
	ARFloat		a, b, c;
	ARFloat		a1, b1, c1;
	ARFloat		a2, b2, c2;
	ARFloat		ma, mb, mc;
	int			t1, t2, t3;
	int			s1, s2, s3;
	int			i, j, k;
	ARFloat		minerr;

	I32    _hx, _hy, _h;
	I32    _err, _minerr;
	U32    _ures;
	I32	   _a1,_b1,_c1;

	FIXED_VEC3D	*_vertex = (FIXED_VEC3D*)malloc(num*sizeof(FIXED_VEC3D)),
		*_pos2d = (FIXED_VEC3D*)malloc(num*sizeof(FIXED_VEC3D)),
		_combo[3], _vec1, _vec2, _trans;
	I32			_combo3[3];

	FIXED_VEC3D	_cpara[3];
	I32			_cpara3[3];

	_vec1.z = 0;

	_trans.x = FIXED_Float_To_Fixed_n(trans[0], 12);
	_trans.y = FIXED_Float_To_Fixed_n(trans[1], 12);
	_trans.z = FIXED_Float_To_Fixed_n(trans[2], 12);

	for(j=0; j<3; j++)
	{
		_cpara[j].x = FIXED_Float_To_Fixed_n(cpara[j][0], 12);
		_cpara[j].y = FIXED_Float_To_Fixed_n(cpara[j][1], 12);
		_cpara[j].z = FIXED_Float_To_Fixed_n(cpara[j][2], 12);
		_cpara3[j] = FIXED_Float_To_Fixed_n(cpara[j][3], 12);
	}

	for(j=0; j<num; j++)
	{
		_vertex[j].x = FIXED_Float_To_Fixed_n(vertex[j][0], BITS);
		_vertex[j].y = FIXED_Float_To_Fixed_n(vertex[j][1], BITS);
		_vertex[j].z = FIXED_Float_To_Fixed_n(vertex[j][2], BITS);

		_pos2d[j].x = FIXED_Float_To_Fixed_n(pos2d[j][0], BITS);
		_pos2d[j].y = FIXED_Float_To_Fixed_n(pos2d[j][1], BITS);
		_pos2d[j].z = 0;
	}


	arGetAngle( rot, &a, &b, &c );


	a2 = a;
	b2 = b;
	c2 = c;
	factor = (ARFloat)(10.0*MD_PI/180.0);
	for( j = 0; j < 10; j++ ) {
		_minerr = 0x40000000;		// value

		for(t1=-1;t1<=1;t1++) {
			for(t2=-1;t2<=1;t2++) {
				for(t3=-1;t3<=1;t3++) {
					a1 = a2 + factor*t1;
					b1 = b2 + factor*t2;
					c1 = c2 + factor*t3;

					_a1 = FIXED_Float_To_Fixed_n(a1, 12);
					_b1 = FIXED_Float_To_Fixed_n(b1, 12);
					_c1 = FIXED_Float_To_Fixed_n(c1, 12);

					arGetNewMatrix12(_a1, _b1, _c1, _trans, NULL, _cpara, _cpara3, _combo, _combo3, profiler);

					for(k=0; k<3; k++)
					{
						_combo[k].x >>= 4;
						_combo[k].y >>= 4;
						_combo[k].z >>= 4;
						_combo3[k] >>= 4;
					}

					_err = 0;
					for( i = 0; i < num; i++ ) {
						FIXED_VEC3_DOT(_combo+0, _vertex+i, &_hx, BITS);
						_hx += _combo3[0];

						FIXED_VEC3_DOT(_combo+1, _vertex+i, &_hy, BITS);
						_hy += _combo3[1];

						FIXED_VEC3_DOT(_combo+2, _vertex+i, &_h, BITS);
						_h += _combo3[2];

						FIXED_DIV2(_hx, _h, _vec1.x, BITS)
							FIXED_DIV2(_hy, _h, _vec1.y, BITS)

							FIXED_VEC3_SUB(_pos2d+i, &_vec1, &_vec2);
						FIXED_VEC3_LENGTH_SQ(&_vec2, &_ures, BITS);
						_err += _ures;
					}

					if( _err < _minerr ) {
						_minerr = _err;
						ma = a1;
						mb = b1;
						mc = c1;
						s1 = t1; s2 = t2; s3 = t3;
					}
				}
			}
		}

		if( s1 == 0 && s2 == 0 && s3 == 0 ) factor *= (ARFloat)0.5;
		a2 = ma;
		b2 = mb;
		c2 = mc;
	}

	free(_vertex);
	free(_pos2d);

	arGetRot( ma, mb, mc, rot );

	minerr = FIXED_Fixed_n_To_Float(_minerr, BITS);

	return minerr/num;
}

*/


//
//
//             FIXEDPOINT specific code ends here
//
//////////////////////////////////////////////////////////////


#ifdef BITS
#undef BITS
#endif


#endif //_FIXEDPOINT_MATH_ACTIVATED_


}  // namespace ARToolKitPlus
// End of ../../src/core/arGetTransMat2.cxx
// Included ../../src/core/arGetTransMat3.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id: arGetTransMat3.cxx 162 2006-04-19 21:28:10Z grabner $
 * @file
 * ======================================================================== */

#define CHECK_CALC 0

//#include <stdlib.h>
//#include <math.h>
//#include <ARToolKitPlus/Tracker.h>
//#include <ARToolKitPlus/matrix.h>


namespace ARToolKitPlus {


#define MD_PI         3.14159265358979323846

static int  check_rotation( ARFloat rot[2][3] );
static int  check_dir( ARFloat dir[3], ARFloat st[2], ARFloat ed[2],
                       ARFloat cpara[3][4] );

AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arGetAngle( ARFloat rot[3][3], ARFloat *wa, ARFloat *wb, ARFloat *wc )
{
	PROFILE_BEGINSEC(profiler, GETANGLE)
	
	ARFloat      a, b, c;
    ARFloat      sina, cosa, sinb, cosb, sinc, cosc;
#if CHECK_CALC
ARFloat   w[3];
int      i;
for(i=0;i<3;i++) w[i] = rot[i][0];
for(i=0;i<3;i++) rot[i][0] = rot[i][1];
for(i=0;i<3;i++) rot[i][1] = rot[i][2];
for(i=0;i<3;i++) rot[i][2] = w[i];
#endif

    if( rot[2][2] > 1.0 ) {
        /* printf("cos(beta) = %f\n", rot[2][2]); */
        rot[2][2] = 1.0;
    }
    else if( rot[2][2] < -1.0 ) {
        /* printf("cos(beta) = %f\n", rot[2][2]); */
        rot[2][2] = -1.0;
    }
    cosb = rot[2][2];
    b = (ARFloat)acos( cosb );
    sinb = (ARFloat)sin( b );
    if( b >= 0.000001 || b <= -0.000001) {
        cosa = rot[0][2] / sinb;
        sina = rot[1][2] / sinb;
        if( cosa > 1.0 ) {
            /* printf("cos(alph) = %f\n", cosa); */
            cosa = 1.0;
            sina = 0.0;
        }
        if( cosa < -1.0 ) {
            /* printf("cos(alph) = %f\n", cosa); */
            cosa = -1.0;
            sina =  0.0;
        }
        if( sina > 1.0 ) {
            /* printf("sin(alph) = %f\n", sina); */
            sina = 1.0;
            cosa = 0.0;
        }
        if( sina < -1.0 ) {
            /* printf("sin(alph) = %f\n", sina); */
            sina = -1.0;
            cosa =  0.0;
        }
        a = (ARFloat)acos( cosa );
        if( sina < 0 ) a = -a;

        sinc =  (rot[2][1]*rot[0][2]-rot[2][0]*rot[1][2])
              / (rot[0][2]*rot[0][2]+rot[1][2]*rot[1][2]);
        cosc =  -(rot[0][2]*rot[2][0]+rot[1][2]*rot[2][1])
               / (rot[0][2]*rot[0][2]+rot[1][2]*rot[1][2]);
        if( cosc > 1.0 ) {
            /* printf("cos(r) = %f\n", cosc); */
            cosc = 1.0;
            sinc = 0.0;
        }
        if( cosc < -1.0 ) {
            /* printf("cos(r) = %f\n", cosc); */
            cosc = -1.0;
            sinc =  0.0;
        }
        if( sinc > 1.0 ) {
            /* printf("sin(r) = %f\n", sinc); */
            sinc = 1.0;
            cosc = 0.0;
        }
        if( sinc < -1.0 ) {
            /* printf("sin(r) = %f\n", sinc); */
            sinc = -1.0;
            cosc =  0.0;
        }
        c = (ARFloat)acos( cosc );
        if( sinc < 0 ) c = -c;
    }
    else {
        a = b = 0.0;
        cosa = cosb = 1.0;
        sina = sinb = 0.0;
        cosc = rot[0][0];
        sinc = rot[1][0];
        if( cosc > 1.0 ) {
            /* printf("cos(r) = %f\n", cosc); */
            cosc = 1.0;
            sinc = 0.0;
        }
        if( cosc < -1.0 ) {
            /* printf("cos(r) = %f\n", cosc); */
            cosc = -1.0;
            sinc =  0.0;
        }
        if( sinc > 1.0 ) {
            /* printf("sin(r) = %f\n", sinc); */
            sinc = 1.0;
            cosc = 0.0;
        }
        if( sinc < -1.0 ) {
            /* printf("sin(r) = %f\n", sinc); */
            sinc = -1.0;
            cosc =  0.0;
        }
        c = (ARFloat)acos( cosc );
        if( sinc < 0 ) c = -c;
    }

    *wa = a;
    *wb = b;
    *wc = c;

	PROFILE_ENDSEC(profiler, GETANGLE)
    return 0;
}


#ifndef _FIXEDPOINT_MATH_ACTIVATED_


// Non-FixedPoint version of arGetRot
//
AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arGetRot( ARFloat a, ARFloat b, ARFloat c, ARFloat rot[3][3] )
{
	PROFILE_BEGINSEC(profiler, GETROT)

    ARFloat   sina, sinb, sinc;
    ARFloat   cosa, cosb, cosc;
#if CHECK_CALC
    ARFloat   w[3];
    int      i;
#endif

    sina = (ARFloat)sin(a); cosa = (ARFloat)cos(a);
    sinb = (ARFloat)sin(b); cosb = (ARFloat)cos(b);
    sinc = (ARFloat)sin(c); cosc = (ARFloat)cos(c);
    rot[0][0] = cosa*cosa*cosb*cosc+sina*sina*cosc+sina*cosa*cosb*sinc-sina*cosa*sinc;
    rot[0][1] = -cosa*cosa*cosb*sinc-sina*sina*sinc+sina*cosa*cosb*cosc-sina*cosa*cosc;
    rot[0][2] = cosa*sinb;
    rot[1][0] = sina*cosa*cosb*cosc-sina*cosa*cosc+sina*sina*cosb*sinc+cosa*cosa*sinc;
    rot[1][1] = -sina*cosa*cosb*sinc+sina*cosa*sinc+sina*sina*cosb*cosc+cosa*cosa*cosc;
    rot[1][2] = sina*sinb;
    rot[2][0] = -cosa*sinb*cosc-sina*sinb*sinc;
    rot[2][1] = cosa*sinb*sinc-sina*sinb*cosc;
    rot[2][2] = cosb;

#if CHECK_CALC
    for(i=0;i<3;i++) w[i] = rot[i][2];
    for(i=0;i<3;i++) rot[i][2] = rot[i][1];
    for(i=0;i<3;i++) rot[i][1] = rot[i][0];
    for(i=0;i<3;i++) rot[i][0] = w[i];
#endif

	PROFILE_ENDSEC(profiler, GETROT)
    return 0;
}

#endif //_FIXEDPOINT_MATH_ACTIVATED_


AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arGetNewMatrix(ARFloat a, ARFloat b, ARFloat c,
						ARFloat trans[3], ARFloat trans2[3][4],
						ARFloat cpara[3][4], ARFloat ret[3][4])
{
    ARFloat   cpara2[3][4];
    ARFloat   rot[3][3];
    int      i, j;

	PROFILE_BEGINSEC(profiler, GETNEWMATRIX)

    arGetRot( a, b, c, rot );

    if( trans2 != NULL ) {
        for( j = 0; j < 3; j++ ) {
            for( i = 0; i < 4; i++ ) {
                cpara2[j][i] = cpara[j][0] * trans2[0][i]
                             + cpara[j][1] * trans2[1][i]
                             + cpara[j][2] * trans2[2][i];
            }
        }
    }
    else {
        for( j = 0; j < 3; j++ ) {
            for( i = 0; i < 4; i++ ) {
                cpara2[j][i] = cpara[j][i];
            }
        }
    }

    for( j = 0; j < 3; j++ ) {
        for( i = 0; i < 3; i++ ) {
            ret[j][i] = cpara2[j][0] * rot[0][i]
                      + cpara2[j][1] * rot[1][i]
                      + cpara2[j][2] * rot[2][i];
        }
        ret[j][3] = cpara2[j][0] * trans[0]
                  + cpara2[j][1] * trans[1]
                  + cpara2[j][2] * trans[2]
                  + cpara2[j][3];
    }

	PROFILE_ENDSEC(profiler, GETNEWMATRIX)
    return(0);
}

AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arGetInitRot( ARMarkerInfo *marker_info, ARFloat cpara[3][4], ARFloat rot[3][3] )
{
    ARFloat  wdir[3][3];
    ARFloat  w, w1, w2, w3;
    int     dir;
    int     j;

	PROFILE_BEGINSEC(profiler, GETINITROT)

    dir = marker_info->dir;

    for( j = 0; j < 2; j++ ) {
        w1 = marker_info->line[(4-dir+j)%4][0] * marker_info->line[(6-dir+j)%4][1]
           - marker_info->line[(6-dir+j)%4][0] * marker_info->line[(4-dir+j)%4][1];
        w2 = marker_info->line[(4-dir+j)%4][1] * marker_info->line[(6-dir+j)%4][2]
           - marker_info->line[(6-dir+j)%4][1] * marker_info->line[(4-dir+j)%4][2];
        w3 = marker_info->line[(4-dir+j)%4][2] * marker_info->line[(6-dir+j)%4][0]
           - marker_info->line[(6-dir+j)%4][2] * marker_info->line[(4-dir+j)%4][0];

        wdir[j][0] =  w1*(cpara[0][1]*cpara[1][2]-cpara[0][2]*cpara[1][1])
                   +  w2*cpara[1][1]
                   -  w3*cpara[0][1];
        wdir[j][1] = -w1*cpara[0][0]*cpara[1][2]
                   +  w3*cpara[0][0];
        wdir[j][2] =  w1*cpara[0][0]*cpara[1][1];
        w = (ARFloat)sqrt( wdir[j][0]*wdir[j][0]
                         + wdir[j][1]*wdir[j][1]
                         + wdir[j][2]*wdir[j][2] );
        wdir[j][0] /= w;
        wdir[j][1] /= w;
        wdir[j][2] /= w;
    }

    if( check_dir(wdir[0], marker_info->vertex[(4-dir)%4],
                  marker_info->vertex[(5-dir)%4], cpara) < 0 )
	{
		PROFILE_ENDSEC(profiler, GETINITROT)
		return -1;
	}

	if( check_dir(wdir[1], marker_info->vertex[(7-dir)%4],
                  marker_info->vertex[(4-dir)%4], cpara) < 0 )
	{
		PROFILE_ENDSEC(profiler, GETINITROT)
		return -1;
	}

    if( check_rotation(wdir) < 0 )
	{
		PROFILE_ENDSEC(profiler, GETINITROT)
		return -1;
	}

    wdir[2][0] = wdir[0][1]*wdir[1][2] - wdir[0][2]*wdir[1][1];
    wdir[2][1] = wdir[0][2]*wdir[1][0] - wdir[0][0]*wdir[1][2];
    wdir[2][2] = wdir[0][0]*wdir[1][1] - wdir[0][1]*wdir[1][0];
    w = (ARFloat)sqrt( wdir[2][0]*wdir[2][0]
                     + wdir[2][1]*wdir[2][1]
                     + wdir[2][2]*wdir[2][2] );
    wdir[2][0] /= w;
    wdir[2][1] /= w;
    wdir[2][2] /= w;
/*
    if( wdir[2][2] < 0 ) {
        wdir[2][0] /= -w;
        wdir[2][1] /= -w;
        wdir[2][2] /= -w;
    }
    else {
        wdir[2][0] /= w;
        wdir[2][1] /= w;
        wdir[2][2] /= w;
    }
*/

    rot[0][0] = wdir[0][0];
    rot[1][0] = wdir[0][1];
    rot[2][0] = wdir[0][2];
    rot[0][1] = wdir[1][0];
    rot[1][1] = wdir[1][1];
    rot[2][1] = wdir[1][2];
    rot[0][2] = wdir[2][0];
    rot[1][2] = wdir[2][1];
    rot[2][2] = wdir[2][2];

	PROFILE_ENDSEC(profiler, GETINITROT)
    return 0;
}


//////////////////////////////////////////////////////////////
//
//             POCKETPC specific code starts here
//


#ifdef _FIXEDPOINT_MATH_ACTIVATED_

#define BITS 28

// FixedPoint version of arGetRot - n is 28
//
// TODO: we can further optimize this routine
//       by storing products
//

#pragma message(">>> Using arGetRot_28()")
int arGetRot_28(I32 _a, I32 _b, I32 _c, FIXED_VEC3D _rot[3])
{
	I32 _sina, _sinb, _sinc, _cosa, _cosb, _cosc;
	I32 _tmp1=0, _tmp2=0;
	I32 _res1,_res2,_res3,_res4,  _res5,_res6;

	FIXED_SINCOS(_a, &_sina, &_cosa, BITS);
	FIXED_SINCOS(_b, &_sinb, &_cosb, BITS);
	FIXED_SINCOS(_c, &_sinc, &_cosc, BITS);

	//FIXED_SIN(_a, &_sina, BITS);
	//FIXED_SIN(_b, &_sinb, BITS);
	//FIXED_SIN(_c, &_sinc, BITS);
	//FIXED_COS(_a, &_cosa, BITS);
	//FIXED_COS(_b, &_cosb, BITS);
	//FIXED_COS(_c, &_cosc, BITS);

	I32 _cosa2, _sina2, _sina_cosa, _cosb_cosc,
		_sina_sinb, _cosa_sinb, _cosb_sinc;

	// some intermediate results that can be used more than once later
	//
	FIXED_MUL2(_cosa,_cosa, _cosa2, BITS);
	FIXED_MUL2(_sina,_sina, _sina2, BITS);
	FIXED_MUL2(_sina,_cosa, _sina_cosa, BITS);
	FIXED_MUL2(_cosb,_cosc, _cosb_cosc, BITS);
	FIXED_MUL2(_sina,_sinb, _sina_sinb, BITS);
	FIXED_MUL2(_cosa,_sinb, _cosa_sinb, BITS);
	FIXED_MUL2(_cosb,_sinc, _cosb_sinc, BITS);



	// rot[0][0]
	FIXED_MUL2(_cosa2,	_cosb_cosc,	_res1,BITS);
	FIXED_MUL2(_sina2,	_cosc, _res2,BITS);
	FIXED_MUL2(_sina_cosa,	_cosb_sinc, _res3,BITS);
	FIXED_MUL2(_sina_cosa,	_sinc, _res5,BITS);					//store for later use
	_rot[0].x = _res1+_res2+_res3-_res5;


	// rot[0][1]
	FIXED_MUL2(_cosa2,	_cosb_sinc, _res1,BITS);
	FIXED_MUL2(_sina2,	_sinc, _res2,BITS);
	FIXED_MUL2(_sina_cosa,	_cosb_cosc, _res3,BITS);
	FIXED_MUL2(_sina_cosa,	_cosc, _res6,BITS);					//store for later use
	_rot[0].y = -_res1-_res2+_res3-_res6;


	// rot[0][2]
	_rot[0].z = _cosa_sinb;


	// rot[1][0]
	FIXED_MUL2(_sina_cosa, _cosb_cosc, _res1,BITS);
	FIXED_MUL2(_sina2, _cosb_sinc, _res3,BITS);
	FIXED_MUL2(_cosa2, _sinc, _res4,BITS);
	_rot[1].x = _res1-_res6+_res3+_res4;


	// rot[1][1]
	FIXED_MUL2(_sina_cosa,	_cosb_sinc, _res1,BITS);
	FIXED_MUL2(_sina2,	_cosb_cosc, _res3,BITS);
	FIXED_MUL2(_cosa2,	_cosc, _res4,BITS);
	_rot[1].y = -_res1+_res5+_res3+_res4;


	// rot[1][2]
	_rot[1].z = _sina_sinb;


	// rot[2][0]
	FIXED_MUL2(_cosa_sinb,	_cosc, _res1,BITS);
	FIXED_MUL2(_sina_sinb,	_sinc, _res2,BITS);
	_rot[2].x = -_res1-_res2;


	// rot[2][1]
	FIXED_MUL2(_cosa_sinb,	_sinc, _res1,BITS);
	FIXED_MUL2(_sina_sinb,	_cosc, _res2,BITS);
	_rot[2].y = _res1-_res2;


	// rot[2][2]
	_rot[2].z = _cosb;


    return 0;
}


int arGetRot_28_old(I32 _a, I32 _b, I32 _c, FIXED_VEC3D _rot[3])
{
	I32 _sina, _sinb, _sinc, _cosa, _cosb, _cosc;
	I32 _tmp1=0, _tmp2=0;
	I32 _res1,_res2,_res3,_res4;

	FIXED_SINCOS(_a, &_sina, &_cosa, BITS);
	FIXED_SINCOS(_b, &_sinb, &_cosb, BITS);
	FIXED_SINCOS(_c, &_sinc, &_cosc, BITS);

	//FIXED_SIN(_a, &_sina, BITS);
	//FIXED_SIN(_b, &_sinb, BITS);
	//FIXED_SIN(_c, &_sinc, BITS);
	//FIXED_COS(_a, &_cosa, BITS);
	//FIXED_COS(_b, &_cosb, BITS);
	//FIXED_COS(_c, &_cosc, BITS);

	// rot[0][0]
	FIXED_MUL4(_cosa,_cosa,_cosb,_cosc, _res1, BITS)			// cosa*cosa*cosb*cosc
	FIXED_MUL3(_sina,_sina,_cosc, _res2, BITS)					// sina*sina*cosc
	FIXED_MUL4(_sina,_cosa,_cosb,_sinc, _res3, BITS)			// sina*cosa*cosb*sinc
	FIXED_MUL3(_sina,_cosa,_sinc, _res4, BITS)					// sina*cosa*sinc
	_rot[0].x = _res1+_res2+_res3-_res4;

	// rot[0][1]
	FIXED_MUL4(_cosa,_cosa,_cosb,_sinc, _res1, BITS)			// cosa*cosa*cosb*sinc
	FIXED_MUL3(_sina,_sina,_sinc, _res2, BITS)					// sina*sina*sinc
	FIXED_MUL4(_sina,_cosa,_cosb,_cosc, _res3, BITS)			// sina*cosa*cosb*cosc
	FIXED_MUL3(_sina,_cosa,_cosc, _res4, BITS)					// sina*cosa*cosc
	_rot[0].y = -_res1-_res2+_res3-_res4;

	// rot[0][2]
	FIXED_MUL2(_cosa,_sinb, _res1, BITS);						// cosa*sinb
	_rot[0].z = _res1;

	// rot[1][0]
	FIXED_MUL4(_sina,_cosa,_cosb,_cosc, _res1, BITS)			// sina*cosa*cosb*cosc
	FIXED_MUL3(_sina,_cosa,_cosc, _res2, BITS)					// sina*cosa*cosc
	FIXED_MUL4(_sina,_sina,_cosb,_sinc, _res3, BITS)			// sina*sina*cosb*sinc
	FIXED_MUL3(_cosa,_cosa,_sinc, _res4, BITS)					// cosa*cosa*sinc
	_rot[1].x = _res1-_res2+_res3+_res4;

	// rot[1][1]
	FIXED_MUL4(_sina,_cosa,_cosb,_sinc, _res1, BITS)			// sina*cosa*cosb*sinc
	FIXED_MUL3(_sina,_cosa,_sinc, _res2, BITS)					// sina*cosa*sinc
	FIXED_MUL4(_sina,_sina,_cosb,_cosc, _res3, BITS)			// sina*sina*cosb*cosc
	FIXED_MUL3(_cosa,_cosa,_cosc, _res4, BITS)					// cosa*cosa*cosc
	_rot[1].y = -_res1+_res2+_res3+_res4;

	// rot[1][2]
	FIXED_MUL2(_sina,_sinb, _res1, BITS);						// sina*sinb
	_rot[1].z = _res1;

	// rot[2][0]
	FIXED_MUL3(_cosa,_sinb,_cosc, _res1, BITS)					// cosa*sinb*cosc
	FIXED_MUL3(_sina,_sinb,_sinc, _res2, BITS)					// sina*sinb*sinc
	_rot[2].x = -_res1-_res2;

	// rot[2][1]
	FIXED_MUL3(_cosa,_sinb,_sinc, _res1, BITS)					// cosa*sinb*sinc
	FIXED_MUL3(_sina,_sinb,_cosc, _res2, BITS)					// sina*sinb*cosc
	_rot[2].y = _res1-_res2;

	// rot[2][2]
	_rot[2].z = _cosb;

	return 0;
}


// FixedPoint version of arGetRot with std interface
//
AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arGetRot( ARFloat a, ARFloat b, ARFloat c, ARFloat rot[3][3] )
{
#if CHECK_CALC
    ARFloat   w[3];
    int      i;
#endif


	I32 _a, _b, _c;
	I32 _sina, _sinb, _sinc, _cosa, _cosb, _cosc;
	I32 _tmp1=0, _tmp2=0;
	I32 _res1,_res2,_res3,_res4;

	_a = FIXED_Float_To_Fixed_n(a, BITS);
	_b = FIXED_Float_To_Fixed_n(b, BITS);
	_c = FIXED_Float_To_Fixed_n(c, BITS);

	FIXED_SINCOS(_a, &_sina, &_cosa, BITS);
	FIXED_SINCOS(_b, &_sinb, &_cosb, BITS);
	FIXED_SINCOS(_c, &_sinc, &_cosc, BITS);

	//FIXED_SIN(_a, &_sina, BITS);
	//FIXED_SIN(_b, &_sinb, BITS);
	//FIXED_SIN(_c, &_sinc, BITS);
	//FIXED_COS(_a, &_cosa, BITS);
	//FIXED_COS(_b, &_cosb, BITS);
	//FIXED_COS(_c, &_cosc, BITS);

	// rot[0][0]
	FIXED_MUL4(_cosa,_cosa,_cosb,_cosc, _res1, BITS)			// cosa*cosa*cosb*cosc
	FIXED_MUL3(_sina,_sina,_cosc, _res2, BITS)					// sina*sina*cosc
	FIXED_MUL4(_sina,_cosa,_cosb,_sinc, _res3, BITS)			// sina*cosa*cosb*sinc
	FIXED_MUL3(_sina,_cosa,_sinc, _res4, BITS)					// sina*cosa*sinc
	rot[0][0] = FIXED_Fixed_n_To_Float((_res1+_res2+_res3-_res4), BITS);

	// rot[0][1]
	FIXED_MUL4(_cosa,_cosa,_cosb,_sinc, _res1, BITS)			// cosa*cosa*cosb*sinc
	FIXED_MUL3(_sina,_sina,_sinc, _res2, BITS)					// sina*sina*sinc
	FIXED_MUL4(_sina,_cosa,_cosb,_cosc, _res3, BITS)			// sina*cosa*cosb*cosc
	FIXED_MUL3(_sina,_cosa,_cosc, _res4, BITS)					// sina*cosa*cosc
	rot[0][1] = FIXED_Fixed_n_To_Float((-_res1-_res2+_res3-_res4), BITS);

	// rot[0][2]
	FIXED_MUL2(_cosa,_sinb, _res1, BITS);						// cosa*sinb
	rot[0][2] = FIXED_Fixed_n_To_Float(_res1, BITS);

	// rot[1][0]
	FIXED_MUL4(_sina,_cosa,_cosb,_cosc, _res1, BITS)			// sina*cosa*cosb*cosc
	FIXED_MUL3(_sina,_cosa,_cosc, _res2, BITS)					// sina*cosa*cosc
	FIXED_MUL4(_sina,_sina,_cosb,_sinc, _res3, BITS)			// sina*sina*cosb*sinc
	FIXED_MUL3(_cosa,_cosa,_sinc, _res4, BITS)					// cosa*cosa*sinc
	rot[1][0] = FIXED_Fixed_n_To_Float((_res1-_res2+_res3+_res4), BITS);

	// rot[1][1]
	FIXED_MUL4(_sina,_cosa,_cosb,_sinc, _res1, BITS)			// sina*cosa*cosb*sinc
	FIXED_MUL3(_sina,_cosa,_sinc, _res2, BITS)					// sina*cosa*sinc
	FIXED_MUL4(_sina,_sina,_cosb,_cosc, _res3, BITS)			// sina*sina*cosb*cosc
	FIXED_MUL3(_cosa,_cosa,_cosc, _res4, BITS)					// cosa*cosa*cosc
	rot[1][1] = FIXED_Fixed_n_To_Float((-_res1+_res2+_res3+_res4), BITS);

	// rot[1][2]
	FIXED_MUL2(_sina,_sinb, _res1, BITS)						// sina*sinb
	rot[1][2] = FIXED_Fixed_n_To_Float(_res1, BITS);

	// rot[2][0]
	FIXED_MUL3(_cosa,_sinb,_cosc, _res1, BITS)					// cosa*sinb*cosc
	FIXED_MUL3(_sina,_sinb,_sinc, _res2, BITS)					// sina*sinb*sinc
	rot[2][0] = FIXED_Fixed_n_To_Float((-_res1-_res2), BITS);

	// rot[2][1]
	FIXED_MUL3(_cosa,_sinb,_sinc, _res1, BITS)					// cosa*sinb*sinc
	FIXED_MUL3(_sina,_sinb,_cosc, _res2, BITS)					// sina*sinb*cosc
	rot[2][1] = FIXED_Fixed_n_To_Float((_res1-_res2), BITS);

	// rot[2][2]
	rot[2][2] = FIXED_Fixed_n_To_Float((_cosb), BITS);


#if CHECK_CALC
    for(i=0;i<3;i++) w[i] = rot[i][2];
    for(i=0;i<3;i++) rot[i][2] = rot[i][1];
    for(i=0;i<3;i++) rot[i][1] = rot[i][0];
    for(i=0;i<3;i++) rot[i][0] = w[i];
#endif

    return 0;
}


#ifdef BITS
#undef BITS
#endif
#define BITS 12


#pragma message(">>> Using arGetNewMatrix12()")

// optimized arGetNewMatrix() version which completely
// works with intel's fixed point library
// (n is 12)
int arGetNewMatrix12(I32 _a, I32 _b, I32 _c,
                    FIXED_VEC3D _trans, ARFloat trans2[3][4],
                    FIXED_VEC3D _cpara2[3], I32 _cpara32[3], FIXED_VEC3D ret[3], I32 _ret3[3], Profiler& nProfiler)
{
    int      j;

	const unsigned int dbits = 16;
	const unsigned int conv = 1<<dbits;

	// all this only works if trans2 is NULL (no stereo)
	//
	FIXED_VEC3D	_rot[3], _rot2[3];

	PROFILE_BEGINSEC(nProfiler, GETROT)
	arGetRot_28(_a*conv, _b*conv, _c*conv, _rot2);
	//arGetRot_28_old(_a*conv, _b*conv, _c*conv, _rot2);
	PROFILE_ENDSEC(nProfiler, GETROT)

	// fill _rot mirrored for better access (vector dot-methods)
	//
	_rot[0].x = _rot2[0].x >> dbits;
	_rot[0].y = _rot2[1].x >> dbits;
	_rot[0].z = _rot2[2].x >> dbits;
	_rot[1].x = _rot2[0].y >> dbits;
	_rot[1].y = _rot2[1].y >> dbits;
	_rot[1].z = _rot2[2].y >> dbits;
	_rot[2].x = _rot2[0].z >> dbits;
	_rot[2].y = _rot2[1].z >> dbits;
	_rot[2].z = _rot2[2].z >> dbits;

	for(j=0; j<3; j++)
	{
		FIXED_VEC3_DOT(_cpara2+j, _rot+0, &ret[j].x, BITS);
		FIXED_VEC3_DOT(_cpara2+j, _rot+1, &ret[j].y, BITS);
		FIXED_VEC3_DOT(_cpara2+j, _rot+2, &ret[j].z, BITS);

		FIXED_VEC3_DOT(_cpara2+j, &_trans, _ret3+j, BITS);
		_ret3[j] += _cpara32[j];
	}

    return(0);
}


#ifdef BITS
#undef BITS
#endif


#endif //_FIXEDPOINT_MATH_ACTIVATED_


//
//             POCKETPC specific code ends here
//
//////////////////////////////////////////////////////////////



static int
check_dir( ARFloat dir[3], ARFloat st[2], ARFloat ed[2],
                      ARFloat cpara[3][4] )
{
    ARMat     *mat_a;
    ARFloat    world[2][3];
    ARFloat    camera[2][2];
    ARFloat    v[2][2];
    ARFloat    h;
    int       i, j;

    mat_a = Matrix::alloc( 3, 3 );
    for(j=0;j<3;j++) for(i=0;i<3;i++) mat_a->m[j*3+i] = cpara[j][i];
    Matrix::selfInv( mat_a );
    world[0][0] = mat_a->m[0]*st[0]*(ARFloat)10.0
                + mat_a->m[1]*st[1]*(ARFloat)10.0
                + mat_a->m[2]*(ARFloat)10.0;
    world[0][1] = mat_a->m[3]*st[0]*(ARFloat)10.0
                + mat_a->m[4]*st[1]*(ARFloat)10.0
                + mat_a->m[5]*(ARFloat)10.0;
    world[0][2] = mat_a->m[6]*st[0]*(ARFloat)10.0
                + mat_a->m[7]*st[1]*(ARFloat)10.0
                + mat_a->m[8]*(ARFloat)10.0;
    Matrix::free( mat_a );
    world[1][0] = world[0][0] + dir[0];
    world[1][1] = world[0][1] + dir[1];
    world[1][2] = world[0][2] + dir[2];

    for( i = 0; i < 2; i++ ) {
        h = cpara[2][0] * world[i][0]
          + cpara[2][1] * world[i][1]
          + cpara[2][2] * world[i][2];
        if( h == 0.0 ) return -1;
        camera[i][0] = (cpara[0][0] * world[i][0]
                      + cpara[0][1] * world[i][1]
                      + cpara[0][2] * world[i][2]) / h;
        camera[i][1] = (cpara[1][0] * world[i][0]
                      + cpara[1][1] * world[i][1]
                      + cpara[1][2] * world[i][2]) / h;
    }

    v[0][0] = ed[0] - st[0];
    v[0][1] = ed[1] - st[1];
    v[1][0] = camera[1][0] - camera[0][0];
    v[1][1] = camera[1][1] - camera[0][1];

    if( v[0][0]*v[1][0] + v[0][1]*v[1][1] < 0 ) {
        dir[0] = -dir[0];
        dir[1] = -dir[1];
        dir[2] = -dir[2];
    }

    return 0;
}

static int
check_rotation( ARFloat rot[2][3] )
{
    ARFloat  v1[3], v2[3], v3[3];
    ARFloat  ca, cb, k1, k2, k3, k4;
    ARFloat  a, b, c, d;
    ARFloat  p1, q1, r1;
    ARFloat  p2, q2, r2;
    ARFloat  p3, q3, r3;
    ARFloat  p4, q4, r4;
    ARFloat  w;
    ARFloat  e1, e2, e3, e4;
    int     f;

    v1[0] = rot[0][0];
    v1[1] = rot[0][1];
    v1[2] = rot[0][2];
    v2[0] = rot[1][0];
    v2[1] = rot[1][1];
    v2[2] = rot[1][2];
    v3[0] = v1[1]*v2[2] - v1[2]*v2[1];
    v3[1] = v1[2]*v2[0] - v1[0]*v2[2];
    v3[2] = v1[0]*v2[1] - v1[1]*v2[0];
    w = (ARFloat)sqrt( v3[0]*v3[0]+v3[1]*v3[1]+v3[2]*v3[2] );
    if( w == 0.0 ) return -1;
    v3[0] /= w;
    v3[1] /= w;
    v3[2] /= w;

    cb = v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2];
    if( cb < 0 ) cb *= -1.0;
    ca = ((ARFloat)sqrt(cb+1.0) + (ARFloat)sqrt(1.0-cb)) * (ARFloat)0.5;

    if( v3[1]*v1[0] - v1[1]*v3[0] != 0.0 ) {
        f = 0;
    }
    else {
        if( v3[2]*v1[0] - v1[2]*v3[0] != 0.0 ) {
            w = v1[1]; v1[1] = v1[2]; v1[2] = w;
            w = v3[1]; v3[1] = v3[2]; v3[2] = w;
            f = 1;
        }
        else {
            w = v1[0]; v1[0] = v1[2]; v1[2] = w;
            w = v3[0]; v3[0] = v3[2]; v3[2] = w;
            f = 2;
        }
    }
    if( v3[1]*v1[0] - v1[1]*v3[0] == 0.0 ) return -1;
    k1 = (v1[1]*v3[2] - v3[1]*v1[2]) / (v3[1]*v1[0] - v1[1]*v3[0]);
    k2 = (v3[1] * ca) / (v3[1]*v1[0] - v1[1]*v3[0]);
    k3 = (v1[0]*v3[2] - v3[0]*v1[2]) / (v3[0]*v1[1] - v1[0]*v3[1]);
    k4 = (v3[0] * ca) / (v3[0]*v1[1] - v1[0]*v3[1]);

    a = k1*k1 + k3*k3 + 1;
    b = k1*k2 + k3*k4;
    c = k2*k2 + k4*k4 - 1;

    d = b*b - a*c;
    if( d < 0 ) return -1;
    r1 = (-b + (ARFloat)sqrt(d))/a;
    p1 = k1*r1 + k2;
    q1 = k3*r1 + k4;
    r2 = (-b - (ARFloat)sqrt(d))/a;
    p2 = k1*r2 + k2;
    q2 = k3*r2 + k4;
    if( f == 1 ) {
        w = q1; q1 = r1; r1 = w;
        w = q2; q2 = r2; r2 = w;
        w = v1[1]; v1[1] = v1[2]; v1[2] = w;
        w = v3[1]; v3[1] = v3[2]; v3[2] = w;
        f = 0;
    }
    if( f == 2 ) {
        w = p1; p1 = r1; r1 = w;
        w = p2; p2 = r2; r2 = w;
        w = v1[0]; v1[0] = v1[2]; v1[2] = w;
        w = v3[0]; v3[0] = v3[2]; v3[2] = w;
        f = 0;
    }

    if( v3[1]*v2[0] - v2[1]*v3[0] != 0.0 ) {
        f = 0;
    }
    else {
        if( v3[2]*v2[0] - v2[2]*v3[0] != 0.0 ) {
            w = v2[1]; v2[1] = v2[2]; v2[2] = w;
            w = v3[1]; v3[1] = v3[2]; v3[2] = w;
            f = 1;
        }
        else {
            w = v2[0]; v2[0] = v2[2]; v2[2] = w;
            w = v3[0]; v3[0] = v3[2]; v3[2] = w;
            f = 2;
        }
    }
    if( v3[1]*v2[0] - v2[1]*v3[0] == 0.0 ) return -1;
    k1 = (v2[1]*v3[2] - v3[1]*v2[2]) / (v3[1]*v2[0] - v2[1]*v3[0]);
    k2 = (v3[1] * ca) / (v3[1]*v2[0] - v2[1]*v3[0]);
    k3 = (v2[0]*v3[2] - v3[0]*v2[2]) / (v3[0]*v2[1] - v2[0]*v3[1]);
    k4 = (v3[0] * ca) / (v3[0]*v2[1] - v2[0]*v3[1]);

    a = k1*k1 + k3*k3 + 1;
    b = k1*k2 + k3*k4;
    c = k2*k2 + k4*k4 - 1;

    d = b*b - a*c;
    if( d < 0 ) return -1;
    r3 = (-b + (ARFloat)sqrt(d))/a;
    p3 = k1*r3 + k2;
    q3 = k3*r3 + k4;
    r4 = (-b - (ARFloat)sqrt(d))/a;
    p4 = k1*r4 + k2;
    q4 = k3*r4 + k4;
    if( f == 1 ) {
        w = q3; q3 = r3; r3 = w;
        w = q4; q4 = r4; r4 = w;
        w = v2[1]; v2[1] = v2[2]; v2[2] = w;
        w = v3[1]; v3[1] = v3[2]; v3[2] = w;
        f = 0;
    }
    if( f == 2 ) {
        w = p3; p3 = r3; r3 = w;
        w = p4; p4 = r4; r4 = w;
        w = v2[0]; v2[0] = v2[2]; v2[2] = w;
        w = v3[0]; v3[0] = v3[2]; v3[2] = w;
        f = 0;
    }

    e1 = p1*p3+q1*q3+r1*r3; if( e1 < 0 ) e1 = -e1;
    e2 = p1*p4+q1*q4+r1*r4; if( e2 < 0 ) e2 = -e2;
    e3 = p2*p3+q2*q3+r2*r3; if( e3 < 0 ) e3 = -e3;
    e4 = p2*p4+q2*q4+r2*r4; if( e4 < 0 ) e4 = -e4;
    if( e1 < e2 ) {
        if( e1 < e3 ) {
            if( e1 < e4 ) {
                rot[0][0] = p1;
                rot[0][1] = q1;
                rot[0][2] = r1;
                rot[1][0] = p3;
                rot[1][1] = q3;
                rot[1][2] = r3;
            }
            else {
                rot[0][0] = p2;
                rot[0][1] = q2;
                rot[0][2] = r2;
                rot[1][0] = p4;
                rot[1][1] = q4;
                rot[1][2] = r4;
            }
        }
        else {
            if( e3 < e4 ) {
                rot[0][0] = p2;
                rot[0][1] = q2;
                rot[0][2] = r2;
                rot[1][0] = p3;
                rot[1][1] = q3;
                rot[1][2] = r3;
            }
            else {
                rot[0][0] = p2;
                rot[0][1] = q2;
                rot[0][2] = r2;
                rot[1][0] = p4;
                rot[1][1] = q4;
                rot[1][2] = r4;
            }
        }
    }
    else {
        if( e2 < e3 ) {
            if( e2 < e4 ) {
                rot[0][0] = p1;
                rot[0][1] = q1;
                rot[0][2] = r1;
                rot[1][0] = p4;
                rot[1][1] = q4;
                rot[1][2] = r4;
            }
            else {
                rot[0][0] = p2;
                rot[0][1] = q2;
                rot[0][2] = r2;
                rot[1][0] = p4;
                rot[1][1] = q4;
                rot[1][2] = r4;
            }
        }
        else {
            if( e3 < e4 ) {
                rot[0][0] = p2;
                rot[0][1] = q2;
                rot[0][2] = r2;
                rot[1][0] = p3;
                rot[1][1] = q3;
                rot[1][2] = r3;
            }
            else {
                rot[0][0] = p2;
                rot[0][1] = q2;
                rot[0][2] = r2;
                rot[1][0] = p4;
                rot[1][1] = q4;
                rot[1][2] = r4;
            }
        }
    }

    return 0;
}


}  // namespace ARToolKitPlus
// End of ../../src/core/arGetTransMat3.cxx
// Included ../../src/core/rppGetTransMat.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Thomas Pintaric
 *
 * $Id: rppGetTransMat.cxx 162 2006-04-19 21:28:10Z grabner $
 * @file
 * ======================================================================== */


#include <stdlib.h>
#include <math.h>

#include <ARToolKitPlus/Tracker.h>
#include <ARToolKitPlus/matrix.h>
#include <ARToolKitPlus/extra/rpp.h>


namespace ARToolKitPlus {

AR_TEMPL_FUNC ARFloat
AR_TEMPL_TRACKER::rppGetTransMat(ARMarkerInfo *marker_info, ARFloat center[2], ARFloat width, ARFloat conv[3][4])
{
	const bool initial_estimate_with_arGetInitRot = false; // only for testing
	
	rpp_float err = 1e+20;
	rpp_mat R, R_init;
	rpp_vec t;

	if(initial_estimate_with_arGetInitRot )
	{
		ARFloat  rot[3][3];
		if( arGetInitRot( marker_info, arCamera->mat, rot ) < 0 ) return -1;
		for(int i=0; i<3; i++)
			for(int j=0; j<3; j++)
				R_init[i][j] = (rpp_float)rot[i][j];
	}

	int dir = marker_info->dir;
	rpp_vec ppos2d[4];
	rpp_vec ppos3d[4];
	const unsigned int n_pts = 4;
	const rpp_float model_z =  0;
	const rpp_float iprts_z =  1;

	ppos2d[0][0] = marker_info->vertex[(4-dir)%4][0];
    ppos2d[0][1] = marker_info->vertex[(4-dir)%4][1];
	ppos2d[0][2] = iprts_z;
    ppos2d[1][0] = marker_info->vertex[(5-dir)%4][0];
    ppos2d[1][1] = marker_info->vertex[(5-dir)%4][1];
	ppos2d[1][2] = iprts_z;
    ppos2d[2][0] = marker_info->vertex[(6-dir)%4][0];
    ppos2d[2][1] = marker_info->vertex[(6-dir)%4][1];
	ppos2d[2][2] = iprts_z;
    ppos2d[3][0] = marker_info->vertex[(7-dir)%4][0];
    ppos2d[3][1] = marker_info->vertex[(7-dir)%4][1];
	ppos2d[3][2] = iprts_z;

	ppos3d[0][0] = center[0] - width*(ARFloat)0.5;
    ppos3d[0][1] = center[1] + width*(ARFloat)0.5;
	ppos3d[0][2] = model_z;
    ppos3d[1][0] = center[0] + width*(ARFloat)0.5;
    ppos3d[1][1] = center[1] + width*(ARFloat)0.5;
	ppos3d[1][2] = model_z;
    ppos3d[2][0] = center[0] + width*(ARFloat)0.5;
    ppos3d[2][1] = center[1] - width*(ARFloat)0.5;
	ppos3d[2][2] = model_z;
    ppos3d[3][0] = center[0] - width*(ARFloat)0.5;
    ppos3d[3][1] = center[1] - width*(ARFloat)0.5;
	ppos3d[3][2] = model_z;

	const rpp_float cc[2] = {arCamera->mat[0][2],arCamera->mat[1][2]};
	const rpp_float fc[2] = {arCamera->mat[0][0],arCamera->mat[1][1]};

	robustPlanarPose(err,R,t,cc,fc,ppos3d,ppos2d,n_pts,R_init, !initial_estimate_with_arGetInitRot,0,0,0);

	for(int i=0; i<3; i++)
	{
		conv[i][3] = (ARFloat)t[i];
		for(int j=0; j<3; j++)
			conv[i][j] = (ARFloat)R[i][j];
	}

	if(err > 1e+10) return(-1); // an actual error has occurred in robustPlanarPose()
	return(ARFloat(err)); // NOTE: err is a real number from the interval [0,1e+10]
}

}  // namespace ARToolKitPlus
// End of ../../src/core/rppGetTransMat.cxx
// Included ../../src/core/arGetTransMatCont.cxx
/* ========================================================================
* PROJECT: ARToolKitPlus
* ========================================================================
* This work is based on the original ARToolKit developed by
*   Hirokazu Kato
*   Mark Billinghurst
*   HITLab, University of Washington, Seattle
* http://www.hitl.washington.edu/artoolkit/
*
* Copyright of the derived and new portions of this work
*     (C) 2006 Graz University of Technology
*
* This framework is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This framework is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this framework; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* For further information please contact 
*   Dieter Schmalstieg
*   <schmalstieg@icg.tu-graz.ac.at>
*   Graz University of Technology, 
*   Institut for Computer Graphics and Vision,
*   Inffeldgasse 16a, 8010 Graz, Austria.
* ========================================================================
** @author   Daniel Wagner
*
* $Id: arGetTransMatCont.cxx 162 2006-04-19 21:28:10Z grabner $
* @file
 * ======================================================================== */

//static double arGetTransMatContSub( ARMarkerInfo *marker_info, double prev_conv[3][4],
//                                    double center[2], double width, double conv[3][4] );


namespace ARToolKitPlus {


AR_TEMPL_FUNC ARFloat
AR_TEMPL_TRACKER::arGetTransMatCont2(ARMarkerInfo *marker_info, ARFloat center[2], ARFloat width, ARFloat conv[3][4])
{
	return arGetTransMatCont(marker_info, conv, center, width, conv);
}


AR_TEMPL_FUNC ARFloat
AR_TEMPL_TRACKER::arGetTransMatCont(ARMarkerInfo *marker_info, ARFloat prev_conv[3][4], ARFloat center[2], ARFloat width, ARFloat conv[3][4])
{
    ARFloat  err1, err2;
    ARFloat wtrans[3][4];
    int     i, j;

    err1 = arGetTransMatContSub(marker_info, prev_conv, center, width, conv);
    if( err1 > AR_GET_TRANS_CONT_MAT_MAX_FIT_ERROR ) {
        err2 = arGetTransMat(marker_info, center, width, wtrans);
        if( err2 < err1 ) {
            for( j = 0; j < 3; j++ ) {
                for( i = 0; i < 4; i++ ) conv[j][i] = wtrans[j][i];
            }
            err1 = err2;
        }
    }

    return err1;
}


AR_TEMPL_FUNC ARFloat
AR_TEMPL_TRACKER::arGetTransMatContSub(ARMarkerInfo *marker_info, ARFloat prev_conv[3][4], ARFloat center[2], ARFloat width, ARFloat conv[3][4])
{
    ARFloat  rot[3][3];
    ARFloat  ppos2d[4][2];
    ARFloat  ppos3d[4][2];
    int     dir;
    ARFloat  err;
    int     i, j;

    for( i = 0; i < 3; i++ ) {
        for( j = 0; j < 3; j++ ) {
            rot[i][j] = prev_conv[i][j];
        }
    }

    dir = marker_info->dir;
    ppos2d[0][0] = marker_info->vertex[(4-dir)%4][0];
    ppos2d[0][1] = marker_info->vertex[(4-dir)%4][1];
    ppos2d[1][0] = marker_info->vertex[(5-dir)%4][0];
    ppos2d[1][1] = marker_info->vertex[(5-dir)%4][1];
    ppos2d[2][0] = marker_info->vertex[(6-dir)%4][0];
    ppos2d[2][1] = marker_info->vertex[(6-dir)%4][1];
    ppos2d[3][0] = marker_info->vertex[(7-dir)%4][0];
    ppos2d[3][1] = marker_info->vertex[(7-dir)%4][1];
	ppos3d[0][0] = center[0] - width*(ARFloat)0.5;
	ppos3d[0][1] = center[1] + width*(ARFloat)0.5;
	ppos3d[1][0] = center[0] + width*(ARFloat)0.5;
	ppos3d[1][1] = center[1] + width*(ARFloat)0.5;
	ppos3d[2][0] = center[0] + width*(ARFloat)0.5;
	ppos3d[2][1] = center[1] - width*(ARFloat)0.5;
	ppos3d[3][0] = center[0] - width*(ARFloat)0.5;
	ppos3d[3][1] = center[1] - width*(ARFloat)0.5;

    for( i = 0; i < AR_GET_TRANS_MAT_MAX_LOOP_COUNT; i++ ) {
        err = arGetTransMat3( rot, ppos2d, ppos3d, 4, conv, arCamera);
        if( err < AR_GET_TRANS_MAT_MAX_FIT_ERROR ) break;
    }
    return err;
}


}  // namespace ARToolKitPlus
// End of ../../src/core/arGetTransMatCont.cxx
// Included ../../src/core/arLabeling.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id: arLabeling.cxx 162 2006-04-19 21:28:10Z grabner $
 * @file
 * ======================================================================== */


#include <stdlib.h>
#include <stdio.h>
#include <ARToolKitPlus/Tracker.h>


namespace ARToolKitPlus {


static void
put_zero( ARUint8 *p, int size )
{
    while( (size--) > 0 ) *(p++) = 0;
}


// in order to get no speed loss we use the preprocessor
// to create 5 different versions of labeling2, for the
// five different major pixel formats
//
#define _DEF_PIXEL_FORMAT_ABGR
#define LABEL_FUNC_NAME arLabeling_ABGR
// Included arLabelingImpl.hxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id$
 * @file
 * ======================================================================== */


AR_TEMPL_FUNC ARInt16*
AR_TEMPL_TRACKER::LABEL_FUNC_NAME(ARUint8 *image, int thresh, int *label_num, int **area,
								  ARFloat **pos, int **clip, int **label_ref)
{
    ARUint8   *pnt;                     /*  image pointer       */
    ARInt16   *pnt1, *pnt2;             /*  image pointer       */
    int       *wk;                      /*  pointer for work    */
    int       wk_max;                   /*  work                */
    int       m,n;                      /*  work                */
    int       i,j,k;                    /*  for loop            */
    int       lxsize, lysize;
    int       poff;
    ARInt16   *l_image;
    int       *work, *work2;
    int       *wlabel_num;
    int       *warea;
    int       *wclip;
    ARFloat    *wpos;
#ifndef _DISABLE_TP_OPTIMIZATIONS_
	int		  pnt2_index, wmax_idx;   // [t.pintaric]
#else
	#pragma message(">> Performance Warning: arlabeling() optimizations disabled.")
#endif //_!DISABLE_TP_OPTIMIZATIONS_

	if(pixelFormat==PIXEL_FORMAT_RGB565)
		checkRGB565LUT();


	assert(l_imageL && "checkImageBuffer() must be called before labeling2(). this should happen automatically in arDetectMarker() & arDetectMarkerLite()");

    l_image = &l_imageL[0];
    work    = &workL[0];
    work2   = &work2L[0];
    wlabel_num = &wlabel_numL;
    warea   = &wareaL[0];
    wclip   = &wclipL[0];
    wpos    = &wposL[0];


	if(pixelFormat!=PIXEL_FORMAT_RGB565 && pixelFormat!=PIXEL_FORMAT_LUM)
		thresh *= 3;

    if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) {
        lxsize = arImXsize / 2;
        lysize = arImYsize / 2;
    }
    else {
        lxsize = arImXsize;
        lysize = arImYsize;
    }

    pnt1 = &l_image[0];
    pnt2 = &l_image[(lysize-1)*lxsize];
    for(i = 0; i < lxsize; i++) {
        *(pnt1++) = *(pnt2++) = 0;
    }

    pnt1 = &l_image[0];
    pnt2 = &l_image[lxsize-1];
    for(i = 0; i < lysize; i++) {
        *pnt1 = *pnt2 = 0;
        pnt1 += lxsize;
        pnt2 += lxsize;
    }

    wk_max = 0;
    pnt2 = &(l_image[lxsize+1]);
    if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) {
        pnt = &(image[(arImXsize*2+2)*pixelSize]);
        poff = pixelSize*2;
    }
    else {
        pnt = &(image[(arImXsize+1)*pixelSize]);
        poff = pixelSize;
    }


//	int diffCorners = -60,
//		diffLeftRight = -40,
//		difftopBottom = -20;

	const int shiftBits = 10;
	int iHalf=lxsize/2, jHalf=lysize/2;

	int threshFact = (pixelFormat!=PIXEL_FORMAT_RGB565 && pixelFormat!=PIXEL_FORMAT_LUM) ? 3 : 1;

	int corrLeftY = (vignetting.corners*threshFact)<<shiftBits,
		dCorrLeftY = ((vignetting.leftright-vignetting.corners*threshFact)<<shiftBits)/jHalf,
		corrCenterY = (vignetting.bottomtop*threshFact)<<shiftBits,
		dCorrCenterY = -corrCenterY/jHalf,
		corrX = 0, dCorrX = 0,
		corrThresh;


	for(j = 1; j < lysize-1; j++, pnt+=poff*2, pnt2+=2)
	{
		if(vignetting.enabled)
		{
			corrX = corrLeftY;
			dCorrX = (corrCenterY-corrLeftY)/iHalf;

			if(j==jHalf)
			{
				dCorrLeftY = -dCorrLeftY;
				dCorrCenterY = -dCorrCenterY;
			
			}

			corrLeftY += dCorrLeftY;
			corrCenterY += dCorrCenterY;
		}

		for(i = 1; i < lxsize-1; i++, pnt+=poff, pnt2++)
		{
			if(vignetting.enabled)
			{
				if(i==iHalf)
					dCorrX = -dCorrX;
				corrX += dCorrX;

				corrThresh = thresh + (corrX>>shiftBits);
			}
			else
				corrThresh = thresh;

			bool isBlack = false;


#ifdef _DEF_PIXEL_FORMAT_ABGR
	            isBlack = ( *(pnt+1) + *(pnt+2) + *(pnt+3) <= corrThresh );
#endif
#ifdef _DEF_PIXEL_FORMAT_BGR
				isBlack = ( *(pnt+0) + *(pnt+1) + *(pnt+2) <= corrThresh );
#endif
#ifdef _DEF_PIXEL_FORMAT_RGB
				isBlack = ( *(pnt+0) + *(pnt+1) + *(pnt+2) <= corrThresh );
#endif
#ifdef _DEF_PIXEL_FORMAT_RGB565
				isBlack = (getLUM8_from_RGB565(pnt) <= corrThresh );
#endif
#ifdef _DEF_PIXEL_FORMAT_LUM
				isBlack = ( *pnt <= corrThresh );
#endif

			if(isBlack) {
				pnt1 = &(pnt2[-lxsize]);
                if( *pnt1 > 0 ) {
                    *pnt2 = *pnt1;

#ifdef _DISABLE_TP_OPTIMIZATIONS_
					// ORIGINAL CODE
					work2[((*pnt2)-1)*7+0] ++;
                    work2[((*pnt2)-1)*7+1] += i;
                    work2[((*pnt2)-1)*7+2] += j;
                    work2[((*pnt2)-1)*7+6] = j;
#else
					// OPTIMIZED CODE [tp]
					// ((*pnt2)-1)*7 should be treated as constant, since
					//  work2[n] (n=0..xsize*ysize) cannot overwrite (*pnt2)
					pnt2_index = ((*pnt2)-1) * 7;
                    work2[pnt2_index+0]++;
                    work2[pnt2_index+1]+= i;
                    work2[pnt2_index+2]+= j;
                    work2[pnt2_index+6] = j;
					// --------------------------------
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                }
                else if( *(pnt1+1) > 0 ) {
                    if( *(pnt1-1) > 0 ) {
                        m = work[*(pnt1+1)-1];
                        n = work[*(pnt1-1)-1];
                        if( m > n ) {
                            *pnt2 = n;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == m ) *wk = n;
                                wk++;
                            }
                        }
                        else if( m < n ) {
                            *pnt2 = m;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == n ) *wk = m;
                                wk++;
                            }
                        }
                        else *pnt2 = m;

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
						work2[((*pnt2)-1)*7+0] ++;
                        work2[((*pnt2)-1)*7+1] += i;
                        work2[((*pnt2)-1)*7+2] += j;
                        work2[((*pnt2)-1)*7+6] = j;
#else
						// PERFORMANCE OPTIMIZATION:
						pnt2_index = ((*pnt2)-1) * 7;
						work2[pnt2_index+0]++;
						work2[pnt2_index+1]+= i;
						work2[pnt2_index+2]+= j;
						work2[pnt2_index+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                    }
                    else if( *(pnt2-1) > 0 ) {
                        m = work[*(pnt1+1)-1];
                        n = work[*(pnt2-1)-1];
                        if( m > n ) {
                            *pnt2 = n;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == m ) *wk = n;
                                wk++;
                            }
                        }
                        else if( m < n ) {
                            *pnt2 = m;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == n ) *wk = m;
                                wk++;
                            }
                        }
                        else *pnt2 = m;

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                        work2[((*pnt2)-1)*7+0] ++;
                        work2[((*pnt2)-1)*7+1] += i;
                        work2[((*pnt2)-1)*7+2] += j;
#else
						// PERFORMANCE OPTIMIZATION:
						pnt2_index = ((*pnt2)-1) * 7;
						work2[pnt2_index+0]++;
						work2[pnt2_index+1]+= i;
						work2[pnt2_index+2]+= j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                    }
                    else {
                        *pnt2 = *(pnt1+1);

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                        work2[((*pnt2)-1)*7+0] ++;
                        work2[((*pnt2)-1)*7+1] += i;
                        work2[((*pnt2)-1)*7+2] += j;
                        if( work2[((*pnt2)-1)*7+3] > i ) work2[((*pnt2)-1)*7+3] = i;
                        work2[((*pnt2)-1)*7+6] = j;
#else
						// PERFORMANCE OPTIMIZATION:
						pnt2_index = ((*pnt2)-1) * 7;
						work2[pnt2_index+0]++;
						work2[pnt2_index+1]+= i;
						work2[pnt2_index+2]+= j;
                        if( work2[pnt2_index+3] > i ) work2[pnt2_index+3] = i;
						work2[pnt2_index+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                    }
                }
                else if( *(pnt1-1) > 0 ) {
                    *pnt2 = *(pnt1-1);

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                    work2[((*pnt2)-1)*7+0] ++;
                    work2[((*pnt2)-1)*7+1] += i;
                    work2[((*pnt2)-1)*7+2] += j;
                    if( work2[((*pnt2)-1)*7+4] < i ) work2[((*pnt2)-1)*7+4] = i;
                    work2[((*pnt2)-1)*7+6] = j;
#else
					// PERFORMANCE OPTIMIZATION:
					pnt2_index = ((*pnt2)-1) * 7;
					work2[pnt2_index+0]++;
					work2[pnt2_index+1]+= i;
					work2[pnt2_index+2]+= j;
                    if( work2[pnt2_index+4] < i ) work2[pnt2_index+4] = i;
					work2[pnt2_index+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                }
                else if( *(pnt2-1) > 0) {
                    *pnt2 = *(pnt2-1);

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                    work2[((*pnt2)-1)*7+0] ++;
                    work2[((*pnt2)-1)*7+1] += i;
                    work2[((*pnt2)-1)*7+2] += j;
                    if( work2[((*pnt2)-1)*7+4] < i ) work2[((*pnt2)-1)*7+4] = i;
#else
					// PERFORMANCE OPTIMIZATION:
					pnt2_index = ((*pnt2)-1) * 7;
					work2[pnt2_index+0]++;
					work2[pnt2_index+1]+= i;
					work2[pnt2_index+2]+= j;
                    if( work2[pnt2_index+4] < i ) work2[pnt2_index+4] = i;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                }
                else {
                    wk_max++;
                    if( wk_max > WORK_SIZE ) {
                        return(0);
                    }
                    work[wk_max-1] = *pnt2 = wk_max;
#ifdef _DISABLE_TP_OPTIMIZATIONS_
                    work2[(wk_max-1)*7+0] = 1;
                    work2[(wk_max-1)*7+1] = i;
                    work2[(wk_max-1)*7+2] = j;
                    work2[(wk_max-1)*7+3] = i;
                    work2[(wk_max-1)*7+4] = i;
                    work2[(wk_max-1)*7+5] = j;
                    work2[(wk_max-1)*7+6] = j;
#else
					wmax_idx = (wk_max-1)*7;
                    work2[wmax_idx+0] = 1;
                    work2[wmax_idx+1] = i;
                    work2[wmax_idx+2] = j;
                    work2[wmax_idx+3] = i;
                    work2[wmax_idx+4] = i;
                    work2[wmax_idx+5] = j;
                    work2[wmax_idx+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_
                }
            }
            else {
                *pnt2 = 0;
            }

		}	// end for x
		
		if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) pnt += arImXsize*pixelSize;

	}	// end for y


    j = 1;
    wk = &(work[0]);
    for(i = 1; i <= wk_max; i++, wk++) {
        *wk = (*wk==i)? j++: work[(*wk)-1];
    }
    *label_num = *wlabel_num = j - 1;
    if( *label_num == 0 ) {
        return( l_image );
    }

    put_zero( (ARUint8 *)warea, *label_num *     sizeof(int) );
    put_zero( (ARUint8 *)wpos,  *label_num * 2 * sizeof(ARFloat) );

#ifdef _DISABLE_TP_OPTIMIZATIONS_
    for(i = 0; i < *label_num; i++) {
        wclip[i*4+0] = lxsize;
        wclip[i*4+1] = 0;
        wclip[i*4+2] = lysize;
        wclip[i*4+3] = 0;
    }
    for(i = 0; i < wk_max; i++) {
        j = work[i] - 1;
        warea[j]    += work2[i*7+0];
        wpos[j*2+0] += work2[i*7+1];
        wpos[j*2+1] += work2[i*7+2];
        if( wclip[j*4+0] > work2[i*7+3] ) wclip[j*4+0] = work2[i*7+3];
        if( wclip[j*4+1] < work2[i*7+4] ) wclip[j*4+1] = work2[i*7+4];
        if( wclip[j*4+2] > work2[i*7+5] ) wclip[j*4+2] = work2[i*7+5];
        if( wclip[j*4+3] < work2[i*7+6] ) wclip[j*4+3] = work2[i*7+6];
    }

    for( i = 0; i < *label_num; i++ ) {
        wpos[i*2+0] /= warea[i];
        wpos[i*2+1] /= warea[i];
    }
#else
	int *wclipRun, *work2Run;
	int iDown;

	wclipRun = wclip;
	iDown = *label_num+1;
	while(--iDown) {
		*wclipRun++ = lxsize;
		*wclipRun++ = 0;
		*wclipRun++ = lysize;
		*wclipRun++ = 0;
    }

	work2Run = work2;
    for(i = 0; i < wk_max; i++) {
        j = work[i] - 1;

        warea[j]    += *work2Run++;
        wpos[j*2+0] += *work2Run++;
        wpos[j*2+1] += *work2Run++;

		wclipRun = wclip+j*4;

		if(*wclipRun > *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;

		if(*wclipRun < *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;

		if(*wclipRun > *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;

		if(*wclipRun < *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;
    }


    for( i = 0; i < *label_num; i++ ) {
        wpos[i*2+0] /= warea[i];
        wpos[i*2+1] /= warea[i];
    }
#endif //!_DISABLE_TP_OPTIMIZATIONS_

    *label_ref = work;
    *area      = warea;
    *pos       = wpos;
    *clip      = wclip;
    return( l_image );
}


#undef LABEL_FUNC_NAME
// End of arLabelingImpl.hxx
#undef _DEF_PIXEL_FORMAT_ABGR

#define _DEF_PIXEL_FORMAT_BGR
#define LABEL_FUNC_NAME arLabeling_BGR
// Included arLabelingImpl.hxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id$
 * @file
 * ======================================================================== */


AR_TEMPL_FUNC ARInt16*
AR_TEMPL_TRACKER::LABEL_FUNC_NAME(ARUint8 *image, int thresh, int *label_num, int **area,
								  ARFloat **pos, int **clip, int **label_ref)
{
    ARUint8   *pnt;                     /*  image pointer       */
    ARInt16   *pnt1, *pnt2;             /*  image pointer       */
    int       *wk;                      /*  pointer for work    */
    int       wk_max;                   /*  work                */
    int       m,n;                      /*  work                */
    int       i,j,k;                    /*  for loop            */
    int       lxsize, lysize;
    int       poff;
    ARInt16   *l_image;
    int       *work, *work2;
    int       *wlabel_num;
    int       *warea;
    int       *wclip;
    ARFloat    *wpos;
#ifndef _DISABLE_TP_OPTIMIZATIONS_
	int		  pnt2_index, wmax_idx;   // [t.pintaric]
#else
	#pragma message(">> Performance Warning: arlabeling() optimizations disabled.")
#endif //_!DISABLE_TP_OPTIMIZATIONS_

	if(pixelFormat==PIXEL_FORMAT_RGB565)
		checkRGB565LUT();


	assert(l_imageL && "checkImageBuffer() must be called before labeling2(). this should happen automatically in arDetectMarker() & arDetectMarkerLite()");

    l_image = &l_imageL[0];
    work    = &workL[0];
    work2   = &work2L[0];
    wlabel_num = &wlabel_numL;
    warea   = &wareaL[0];
    wclip   = &wclipL[0];
    wpos    = &wposL[0];


	if(pixelFormat!=PIXEL_FORMAT_RGB565 && pixelFormat!=PIXEL_FORMAT_LUM)
		thresh *= 3;

    if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) {
        lxsize = arImXsize / 2;
        lysize = arImYsize / 2;
    }
    else {
        lxsize = arImXsize;
        lysize = arImYsize;
    }

    pnt1 = &l_image[0];
    pnt2 = &l_image[(lysize-1)*lxsize];
    for(i = 0; i < lxsize; i++) {
        *(pnt1++) = *(pnt2++) = 0;
    }

    pnt1 = &l_image[0];
    pnt2 = &l_image[lxsize-1];
    for(i = 0; i < lysize; i++) {
        *pnt1 = *pnt2 = 0;
        pnt1 += lxsize;
        pnt2 += lxsize;
    }

    wk_max = 0;
    pnt2 = &(l_image[lxsize+1]);
    if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) {
        pnt = &(image[(arImXsize*2+2)*pixelSize]);
        poff = pixelSize*2;
    }
    else {
        pnt = &(image[(arImXsize+1)*pixelSize]);
        poff = pixelSize;
    }


//	int diffCorners = -60,
//		diffLeftRight = -40,
//		difftopBottom = -20;

	const int shiftBits = 10;
	int iHalf=lxsize/2, jHalf=lysize/2;

	int threshFact = (pixelFormat!=PIXEL_FORMAT_RGB565 && pixelFormat!=PIXEL_FORMAT_LUM) ? 3 : 1;

	int corrLeftY = (vignetting.corners*threshFact)<<shiftBits,
		dCorrLeftY = ((vignetting.leftright-vignetting.corners*threshFact)<<shiftBits)/jHalf,
		corrCenterY = (vignetting.bottomtop*threshFact)<<shiftBits,
		dCorrCenterY = -corrCenterY/jHalf,
		corrX = 0, dCorrX = 0,
		corrThresh;


	for(j = 1; j < lysize-1; j++, pnt+=poff*2, pnt2+=2)
	{
		if(vignetting.enabled)
		{
			corrX = corrLeftY;
			dCorrX = (corrCenterY-corrLeftY)/iHalf;

			if(j==jHalf)
			{
				dCorrLeftY = -dCorrLeftY;
				dCorrCenterY = -dCorrCenterY;
			
			}

			corrLeftY += dCorrLeftY;
			corrCenterY += dCorrCenterY;
		}

		for(i = 1; i < lxsize-1; i++, pnt+=poff, pnt2++)
		{
			if(vignetting.enabled)
			{
				if(i==iHalf)
					dCorrX = -dCorrX;
				corrX += dCorrX;

				corrThresh = thresh + (corrX>>shiftBits);
			}
			else
				corrThresh = thresh;

			bool isBlack = false;


#ifdef _DEF_PIXEL_FORMAT_ABGR
	            isBlack = ( *(pnt+1) + *(pnt+2) + *(pnt+3) <= corrThresh );
#endif
#ifdef _DEF_PIXEL_FORMAT_BGR
				isBlack = ( *(pnt+0) + *(pnt+1) + *(pnt+2) <= corrThresh );
#endif
#ifdef _DEF_PIXEL_FORMAT_RGB
				isBlack = ( *(pnt+0) + *(pnt+1) + *(pnt+2) <= corrThresh );
#endif
#ifdef _DEF_PIXEL_FORMAT_RGB565
				isBlack = (getLUM8_from_RGB565(pnt) <= corrThresh );
#endif
#ifdef _DEF_PIXEL_FORMAT_LUM
				isBlack = ( *pnt <= corrThresh );
#endif

			if(isBlack) {
				pnt1 = &(pnt2[-lxsize]);
                if( *pnt1 > 0 ) {
                    *pnt2 = *pnt1;

#ifdef _DISABLE_TP_OPTIMIZATIONS_
					// ORIGINAL CODE
					work2[((*pnt2)-1)*7+0] ++;
                    work2[((*pnt2)-1)*7+1] += i;
                    work2[((*pnt2)-1)*7+2] += j;
                    work2[((*pnt2)-1)*7+6] = j;
#else
					// OPTIMIZED CODE [tp]
					// ((*pnt2)-1)*7 should be treated as constant, since
					//  work2[n] (n=0..xsize*ysize) cannot overwrite (*pnt2)
					pnt2_index = ((*pnt2)-1) * 7;
                    work2[pnt2_index+0]++;
                    work2[pnt2_index+1]+= i;
                    work2[pnt2_index+2]+= j;
                    work2[pnt2_index+6] = j;
					// --------------------------------
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                }
                else if( *(pnt1+1) > 0 ) {
                    if( *(pnt1-1) > 0 ) {
                        m = work[*(pnt1+1)-1];
                        n = work[*(pnt1-1)-1];
                        if( m > n ) {
                            *pnt2 = n;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == m ) *wk = n;
                                wk++;
                            }
                        }
                        else if( m < n ) {
                            *pnt2 = m;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == n ) *wk = m;
                                wk++;
                            }
                        }
                        else *pnt2 = m;

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
						work2[((*pnt2)-1)*7+0] ++;
                        work2[((*pnt2)-1)*7+1] += i;
                        work2[((*pnt2)-1)*7+2] += j;
                        work2[((*pnt2)-1)*7+6] = j;
#else
						// PERFORMANCE OPTIMIZATION:
						pnt2_index = ((*pnt2)-1) * 7;
						work2[pnt2_index+0]++;
						work2[pnt2_index+1]+= i;
						work2[pnt2_index+2]+= j;
						work2[pnt2_index+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                    }
                    else if( *(pnt2-1) > 0 ) {
                        m = work[*(pnt1+1)-1];
                        n = work[*(pnt2-1)-1];
                        if( m > n ) {
                            *pnt2 = n;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == m ) *wk = n;
                                wk++;
                            }
                        }
                        else if( m < n ) {
                            *pnt2 = m;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == n ) *wk = m;
                                wk++;
                            }
                        }
                        else *pnt2 = m;

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                        work2[((*pnt2)-1)*7+0] ++;
                        work2[((*pnt2)-1)*7+1] += i;
                        work2[((*pnt2)-1)*7+2] += j;
#else
						// PERFORMANCE OPTIMIZATION:
						pnt2_index = ((*pnt2)-1) * 7;
						work2[pnt2_index+0]++;
						work2[pnt2_index+1]+= i;
						work2[pnt2_index+2]+= j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                    }
                    else {
                        *pnt2 = *(pnt1+1);

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                        work2[((*pnt2)-1)*7+0] ++;
                        work2[((*pnt2)-1)*7+1] += i;
                        work2[((*pnt2)-1)*7+2] += j;
                        if( work2[((*pnt2)-1)*7+3] > i ) work2[((*pnt2)-1)*7+3] = i;
                        work2[((*pnt2)-1)*7+6] = j;
#else
						// PERFORMANCE OPTIMIZATION:
						pnt2_index = ((*pnt2)-1) * 7;
						work2[pnt2_index+0]++;
						work2[pnt2_index+1]+= i;
						work2[pnt2_index+2]+= j;
                        if( work2[pnt2_index+3] > i ) work2[pnt2_index+3] = i;
						work2[pnt2_index+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                    }
                }
                else if( *(pnt1-1) > 0 ) {
                    *pnt2 = *(pnt1-1);

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                    work2[((*pnt2)-1)*7+0] ++;
                    work2[((*pnt2)-1)*7+1] += i;
                    work2[((*pnt2)-1)*7+2] += j;
                    if( work2[((*pnt2)-1)*7+4] < i ) work2[((*pnt2)-1)*7+4] = i;
                    work2[((*pnt2)-1)*7+6] = j;
#else
					// PERFORMANCE OPTIMIZATION:
					pnt2_index = ((*pnt2)-1) * 7;
					work2[pnt2_index+0]++;
					work2[pnt2_index+1]+= i;
					work2[pnt2_index+2]+= j;
                    if( work2[pnt2_index+4] < i ) work2[pnt2_index+4] = i;
					work2[pnt2_index+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                }
                else if( *(pnt2-1) > 0) {
                    *pnt2 = *(pnt2-1);

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                    work2[((*pnt2)-1)*7+0] ++;
                    work2[((*pnt2)-1)*7+1] += i;
                    work2[((*pnt2)-1)*7+2] += j;
                    if( work2[((*pnt2)-1)*7+4] < i ) work2[((*pnt2)-1)*7+4] = i;
#else
					// PERFORMANCE OPTIMIZATION:
					pnt2_index = ((*pnt2)-1) * 7;
					work2[pnt2_index+0]++;
					work2[pnt2_index+1]+= i;
					work2[pnt2_index+2]+= j;
                    if( work2[pnt2_index+4] < i ) work2[pnt2_index+4] = i;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                }
                else {
                    wk_max++;
                    if( wk_max > WORK_SIZE ) {
                        return(0);
                    }
                    work[wk_max-1] = *pnt2 = wk_max;
#ifdef _DISABLE_TP_OPTIMIZATIONS_
                    work2[(wk_max-1)*7+0] = 1;
                    work2[(wk_max-1)*7+1] = i;
                    work2[(wk_max-1)*7+2] = j;
                    work2[(wk_max-1)*7+3] = i;
                    work2[(wk_max-1)*7+4] = i;
                    work2[(wk_max-1)*7+5] = j;
                    work2[(wk_max-1)*7+6] = j;
#else
					wmax_idx = (wk_max-1)*7;
                    work2[wmax_idx+0] = 1;
                    work2[wmax_idx+1] = i;
                    work2[wmax_idx+2] = j;
                    work2[wmax_idx+3] = i;
                    work2[wmax_idx+4] = i;
                    work2[wmax_idx+5] = j;
                    work2[wmax_idx+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_
                }
            }
            else {
                *pnt2 = 0;
            }

		}	// end for x
		
		if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) pnt += arImXsize*pixelSize;

	}	// end for y


    j = 1;
    wk = &(work[0]);
    for(i = 1; i <= wk_max; i++, wk++) {
        *wk = (*wk==i)? j++: work[(*wk)-1];
    }
    *label_num = *wlabel_num = j - 1;
    if( *label_num == 0 ) {
        return( l_image );
    }

    put_zero( (ARUint8 *)warea, *label_num *     sizeof(int) );
    put_zero( (ARUint8 *)wpos,  *label_num * 2 * sizeof(ARFloat) );

#ifdef _DISABLE_TP_OPTIMIZATIONS_
    for(i = 0; i < *label_num; i++) {
        wclip[i*4+0] = lxsize;
        wclip[i*4+1] = 0;
        wclip[i*4+2] = lysize;
        wclip[i*4+3] = 0;
    }
    for(i = 0; i < wk_max; i++) {
        j = work[i] - 1;
        warea[j]    += work2[i*7+0];
        wpos[j*2+0] += work2[i*7+1];
        wpos[j*2+1] += work2[i*7+2];
        if( wclip[j*4+0] > work2[i*7+3] ) wclip[j*4+0] = work2[i*7+3];
        if( wclip[j*4+1] < work2[i*7+4] ) wclip[j*4+1] = work2[i*7+4];
        if( wclip[j*4+2] > work2[i*7+5] ) wclip[j*4+2] = work2[i*7+5];
        if( wclip[j*4+3] < work2[i*7+6] ) wclip[j*4+3] = work2[i*7+6];
    }

    for( i = 0; i < *label_num; i++ ) {
        wpos[i*2+0] /= warea[i];
        wpos[i*2+1] /= warea[i];
    }
#else
	int *wclipRun, *work2Run;
	int iDown;

	wclipRun = wclip;
	iDown = *label_num+1;
	while(--iDown) {
		*wclipRun++ = lxsize;
		*wclipRun++ = 0;
		*wclipRun++ = lysize;
		*wclipRun++ = 0;
    }

	work2Run = work2;
    for(i = 0; i < wk_max; i++) {
        j = work[i] - 1;

        warea[j]    += *work2Run++;
        wpos[j*2+0] += *work2Run++;
        wpos[j*2+1] += *work2Run++;

		wclipRun = wclip+j*4;

		if(*wclipRun > *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;

		if(*wclipRun < *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;

		if(*wclipRun > *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;

		if(*wclipRun < *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;
    }


    for( i = 0; i < *label_num; i++ ) {
        wpos[i*2+0] /= warea[i];
        wpos[i*2+1] /= warea[i];
    }
#endif //!_DISABLE_TP_OPTIMIZATIONS_

    *label_ref = work;
    *area      = warea;
    *pos       = wpos;
    *clip      = wclip;
    return( l_image );
}


#undef LABEL_FUNC_NAME
// End of arLabelingImpl.hxx
#undef _DEF_PIXEL_FORMAT_BGR

#define _DEF_PIXEL_FORMAT_RGB
#define LABEL_FUNC_NAME arLabeling_RGB
// Included arLabelingImpl.hxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id$
 * @file
 * ======================================================================== */


AR_TEMPL_FUNC ARInt16*
AR_TEMPL_TRACKER::LABEL_FUNC_NAME(ARUint8 *image, int thresh, int *label_num, int **area,
								  ARFloat **pos, int **clip, int **label_ref)
{
    ARUint8   *pnt;                     /*  image pointer       */
    ARInt16   *pnt1, *pnt2;             /*  image pointer       */
    int       *wk;                      /*  pointer for work    */
    int       wk_max;                   /*  work                */
    int       m,n;                      /*  work                */
    int       i,j,k;                    /*  for loop            */
    int       lxsize, lysize;
    int       poff;
    ARInt16   *l_image;
    int       *work, *work2;
    int       *wlabel_num;
    int       *warea;
    int       *wclip;
    ARFloat    *wpos;
#ifndef _DISABLE_TP_OPTIMIZATIONS_
	int		  pnt2_index, wmax_idx;   // [t.pintaric]
#else
	#pragma message(">> Performance Warning: arlabeling() optimizations disabled.")
#endif //_!DISABLE_TP_OPTIMIZATIONS_

	if(pixelFormat==PIXEL_FORMAT_RGB565)
		checkRGB565LUT();


	assert(l_imageL && "checkImageBuffer() must be called before labeling2(). this should happen automatically in arDetectMarker() & arDetectMarkerLite()");

    l_image = &l_imageL[0];
    work    = &workL[0];
    work2   = &work2L[0];
    wlabel_num = &wlabel_numL;
    warea   = &wareaL[0];
    wclip   = &wclipL[0];
    wpos    = &wposL[0];


	if(pixelFormat!=PIXEL_FORMAT_RGB565 && pixelFormat!=PIXEL_FORMAT_LUM)
		thresh *= 3;

    if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) {
        lxsize = arImXsize / 2;
        lysize = arImYsize / 2;
    }
    else {
        lxsize = arImXsize;
        lysize = arImYsize;
    }

    pnt1 = &l_image[0];
    pnt2 = &l_image[(lysize-1)*lxsize];
    for(i = 0; i < lxsize; i++) {
        *(pnt1++) = *(pnt2++) = 0;
    }

    pnt1 = &l_image[0];
    pnt2 = &l_image[lxsize-1];
    for(i = 0; i < lysize; i++) {
        *pnt1 = *pnt2 = 0;
        pnt1 += lxsize;
        pnt2 += lxsize;
    }

    wk_max = 0;
    pnt2 = &(l_image[lxsize+1]);
    if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) {
        pnt = &(image[(arImXsize*2+2)*pixelSize]);
        poff = pixelSize*2;
    }
    else {
        pnt = &(image[(arImXsize+1)*pixelSize]);
        poff = pixelSize;
    }


//	int diffCorners = -60,
//		diffLeftRight = -40,
//		difftopBottom = -20;

	const int shiftBits = 10;
	int iHalf=lxsize/2, jHalf=lysize/2;

	int threshFact = (pixelFormat!=PIXEL_FORMAT_RGB565 && pixelFormat!=PIXEL_FORMAT_LUM) ? 3 : 1;

	int corrLeftY = (vignetting.corners*threshFact)<<shiftBits,
		dCorrLeftY = ((vignetting.leftright-vignetting.corners*threshFact)<<shiftBits)/jHalf,
		corrCenterY = (vignetting.bottomtop*threshFact)<<shiftBits,
		dCorrCenterY = -corrCenterY/jHalf,
		corrX = 0, dCorrX = 0,
		corrThresh;


	for(j = 1; j < lysize-1; j++, pnt+=poff*2, pnt2+=2)
	{
		if(vignetting.enabled)
		{
			corrX = corrLeftY;
			dCorrX = (corrCenterY-corrLeftY)/iHalf;

			if(j==jHalf)
			{
				dCorrLeftY = -dCorrLeftY;
				dCorrCenterY = -dCorrCenterY;
			
			}

			corrLeftY += dCorrLeftY;
			corrCenterY += dCorrCenterY;
		}

		for(i = 1; i < lxsize-1; i++, pnt+=poff, pnt2++)
		{
			if(vignetting.enabled)
			{
				if(i==iHalf)
					dCorrX = -dCorrX;
				corrX += dCorrX;

				corrThresh = thresh + (corrX>>shiftBits);
			}
			else
				corrThresh = thresh;

			bool isBlack = false;


#ifdef _DEF_PIXEL_FORMAT_ABGR
	            isBlack = ( *(pnt+1) + *(pnt+2) + *(pnt+3) <= corrThresh );
#endif
#ifdef _DEF_PIXEL_FORMAT_BGR
				isBlack = ( *(pnt+0) + *(pnt+1) + *(pnt+2) <= corrThresh );
#endif
#ifdef _DEF_PIXEL_FORMAT_RGB
				isBlack = ( *(pnt+0) + *(pnt+1) + *(pnt+2) <= corrThresh );
#endif
#ifdef _DEF_PIXEL_FORMAT_RGB565
				isBlack = (getLUM8_from_RGB565(pnt) <= corrThresh );
#endif
#ifdef _DEF_PIXEL_FORMAT_LUM
				isBlack = ( *pnt <= corrThresh );
#endif

			if(isBlack) {
				pnt1 = &(pnt2[-lxsize]);
                if( *pnt1 > 0 ) {
                    *pnt2 = *pnt1;

#ifdef _DISABLE_TP_OPTIMIZATIONS_
					// ORIGINAL CODE
					work2[((*pnt2)-1)*7+0] ++;
                    work2[((*pnt2)-1)*7+1] += i;
                    work2[((*pnt2)-1)*7+2] += j;
                    work2[((*pnt2)-1)*7+6] = j;
#else
					// OPTIMIZED CODE [tp]
					// ((*pnt2)-1)*7 should be treated as constant, since
					//  work2[n] (n=0..xsize*ysize) cannot overwrite (*pnt2)
					pnt2_index = ((*pnt2)-1) * 7;
                    work2[pnt2_index+0]++;
                    work2[pnt2_index+1]+= i;
                    work2[pnt2_index+2]+= j;
                    work2[pnt2_index+6] = j;
					// --------------------------------
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                }
                else if( *(pnt1+1) > 0 ) {
                    if( *(pnt1-1) > 0 ) {
                        m = work[*(pnt1+1)-1];
                        n = work[*(pnt1-1)-1];
                        if( m > n ) {
                            *pnt2 = n;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == m ) *wk = n;
                                wk++;
                            }
                        }
                        else if( m < n ) {
                            *pnt2 = m;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == n ) *wk = m;
                                wk++;
                            }
                        }
                        else *pnt2 = m;

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
						work2[((*pnt2)-1)*7+0] ++;
                        work2[((*pnt2)-1)*7+1] += i;
                        work2[((*pnt2)-1)*7+2] += j;
                        work2[((*pnt2)-1)*7+6] = j;
#else
						// PERFORMANCE OPTIMIZATION:
						pnt2_index = ((*pnt2)-1) * 7;
						work2[pnt2_index+0]++;
						work2[pnt2_index+1]+= i;
						work2[pnt2_index+2]+= j;
						work2[pnt2_index+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                    }
                    else if( *(pnt2-1) > 0 ) {
                        m = work[*(pnt1+1)-1];
                        n = work[*(pnt2-1)-1];
                        if( m > n ) {
                            *pnt2 = n;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == m ) *wk = n;
                                wk++;
                            }
                        }
                        else if( m < n ) {
                            *pnt2 = m;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == n ) *wk = m;
                                wk++;
                            }
                        }
                        else *pnt2 = m;

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                        work2[((*pnt2)-1)*7+0] ++;
                        work2[((*pnt2)-1)*7+1] += i;
                        work2[((*pnt2)-1)*7+2] += j;
#else
						// PERFORMANCE OPTIMIZATION:
						pnt2_index = ((*pnt2)-1) * 7;
						work2[pnt2_index+0]++;
						work2[pnt2_index+1]+= i;
						work2[pnt2_index+2]+= j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                    }
                    else {
                        *pnt2 = *(pnt1+1);

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                        work2[((*pnt2)-1)*7+0] ++;
                        work2[((*pnt2)-1)*7+1] += i;
                        work2[((*pnt2)-1)*7+2] += j;
                        if( work2[((*pnt2)-1)*7+3] > i ) work2[((*pnt2)-1)*7+3] = i;
                        work2[((*pnt2)-1)*7+6] = j;
#else
						// PERFORMANCE OPTIMIZATION:
						pnt2_index = ((*pnt2)-1) * 7;
						work2[pnt2_index+0]++;
						work2[pnt2_index+1]+= i;
						work2[pnt2_index+2]+= j;
                        if( work2[pnt2_index+3] > i ) work2[pnt2_index+3] = i;
						work2[pnt2_index+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                    }
                }
                else if( *(pnt1-1) > 0 ) {
                    *pnt2 = *(pnt1-1);

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                    work2[((*pnt2)-1)*7+0] ++;
                    work2[((*pnt2)-1)*7+1] += i;
                    work2[((*pnt2)-1)*7+2] += j;
                    if( work2[((*pnt2)-1)*7+4] < i ) work2[((*pnt2)-1)*7+4] = i;
                    work2[((*pnt2)-1)*7+6] = j;
#else
					// PERFORMANCE OPTIMIZATION:
					pnt2_index = ((*pnt2)-1) * 7;
					work2[pnt2_index+0]++;
					work2[pnt2_index+1]+= i;
					work2[pnt2_index+2]+= j;
                    if( work2[pnt2_index+4] < i ) work2[pnt2_index+4] = i;
					work2[pnt2_index+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                }
                else if( *(pnt2-1) > 0) {
                    *pnt2 = *(pnt2-1);

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                    work2[((*pnt2)-1)*7+0] ++;
                    work2[((*pnt2)-1)*7+1] += i;
                    work2[((*pnt2)-1)*7+2] += j;
                    if( work2[((*pnt2)-1)*7+4] < i ) work2[((*pnt2)-1)*7+4] = i;
#else
					// PERFORMANCE OPTIMIZATION:
					pnt2_index = ((*pnt2)-1) * 7;
					work2[pnt2_index+0]++;
					work2[pnt2_index+1]+= i;
					work2[pnt2_index+2]+= j;
                    if( work2[pnt2_index+4] < i ) work2[pnt2_index+4] = i;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                }
                else {
                    wk_max++;
                    if( wk_max > WORK_SIZE ) {
                        return(0);
                    }
                    work[wk_max-1] = *pnt2 = wk_max;
#ifdef _DISABLE_TP_OPTIMIZATIONS_
                    work2[(wk_max-1)*7+0] = 1;
                    work2[(wk_max-1)*7+1] = i;
                    work2[(wk_max-1)*7+2] = j;
                    work2[(wk_max-1)*7+3] = i;
                    work2[(wk_max-1)*7+4] = i;
                    work2[(wk_max-1)*7+5] = j;
                    work2[(wk_max-1)*7+6] = j;
#else
					wmax_idx = (wk_max-1)*7;
                    work2[wmax_idx+0] = 1;
                    work2[wmax_idx+1] = i;
                    work2[wmax_idx+2] = j;
                    work2[wmax_idx+3] = i;
                    work2[wmax_idx+4] = i;
                    work2[wmax_idx+5] = j;
                    work2[wmax_idx+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_
                }
            }
            else {
                *pnt2 = 0;
            }

		}	// end for x
		
		if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) pnt += arImXsize*pixelSize;

	}	// end for y


    j = 1;
    wk = &(work[0]);
    for(i = 1; i <= wk_max; i++, wk++) {
        *wk = (*wk==i)? j++: work[(*wk)-1];
    }
    *label_num = *wlabel_num = j - 1;
    if( *label_num == 0 ) {
        return( l_image );
    }

    put_zero( (ARUint8 *)warea, *label_num *     sizeof(int) );
    put_zero( (ARUint8 *)wpos,  *label_num * 2 * sizeof(ARFloat) );

#ifdef _DISABLE_TP_OPTIMIZATIONS_
    for(i = 0; i < *label_num; i++) {
        wclip[i*4+0] = lxsize;
        wclip[i*4+1] = 0;
        wclip[i*4+2] = lysize;
        wclip[i*4+3] = 0;
    }
    for(i = 0; i < wk_max; i++) {
        j = work[i] - 1;
        warea[j]    += work2[i*7+0];
        wpos[j*2+0] += work2[i*7+1];
        wpos[j*2+1] += work2[i*7+2];
        if( wclip[j*4+0] > work2[i*7+3] ) wclip[j*4+0] = work2[i*7+3];
        if( wclip[j*4+1] < work2[i*7+4] ) wclip[j*4+1] = work2[i*7+4];
        if( wclip[j*4+2] > work2[i*7+5] ) wclip[j*4+2] = work2[i*7+5];
        if( wclip[j*4+3] < work2[i*7+6] ) wclip[j*4+3] = work2[i*7+6];
    }

    for( i = 0; i < *label_num; i++ ) {
        wpos[i*2+0] /= warea[i];
        wpos[i*2+1] /= warea[i];
    }
#else
	int *wclipRun, *work2Run;
	int iDown;

	wclipRun = wclip;
	iDown = *label_num+1;
	while(--iDown) {
		*wclipRun++ = lxsize;
		*wclipRun++ = 0;
		*wclipRun++ = lysize;
		*wclipRun++ = 0;
    }

	work2Run = work2;
    for(i = 0; i < wk_max; i++) {
        j = work[i] - 1;

        warea[j]    += *work2Run++;
        wpos[j*2+0] += *work2Run++;
        wpos[j*2+1] += *work2Run++;

		wclipRun = wclip+j*4;

		if(*wclipRun > *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;

		if(*wclipRun < *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;

		if(*wclipRun > *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;

		if(*wclipRun < *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;
    }


    for( i = 0; i < *label_num; i++ ) {
        wpos[i*2+0] /= warea[i];
        wpos[i*2+1] /= warea[i];
    }
#endif //!_DISABLE_TP_OPTIMIZATIONS_

    *label_ref = work;
    *area      = warea;
    *pos       = wpos;
    *clip      = wclip;
    return( l_image );
}


#undef LABEL_FUNC_NAME
// End of arLabelingImpl.hxx
#undef _DEF_PIXEL_FORMAT_RGB

#define _DEF_PIXEL_FORMAT_RGB565
#define LABEL_FUNC_NAME arLabeling_RGB565
// Included arLabelingImpl.hxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id$
 * @file
 * ======================================================================== */


AR_TEMPL_FUNC ARInt16*
AR_TEMPL_TRACKER::LABEL_FUNC_NAME(ARUint8 *image, int thresh, int *label_num, int **area,
								  ARFloat **pos, int **clip, int **label_ref)
{
    ARUint8   *pnt;                     /*  image pointer       */
    ARInt16   *pnt1, *pnt2;             /*  image pointer       */
    int       *wk;                      /*  pointer for work    */
    int       wk_max;                   /*  work                */
    int       m,n;                      /*  work                */
    int       i,j,k;                    /*  for loop            */
    int       lxsize, lysize;
    int       poff;
    ARInt16   *l_image;
    int       *work, *work2;
    int       *wlabel_num;
    int       *warea;
    int       *wclip;
    ARFloat    *wpos;
#ifndef _DISABLE_TP_OPTIMIZATIONS_
	int		  pnt2_index, wmax_idx;   // [t.pintaric]
#else
	#pragma message(">> Performance Warning: arlabeling() optimizations disabled.")
#endif //_!DISABLE_TP_OPTIMIZATIONS_

	if(pixelFormat==PIXEL_FORMAT_RGB565)
		checkRGB565LUT();


	assert(l_imageL && "checkImageBuffer() must be called before labeling2(). this should happen automatically in arDetectMarker() & arDetectMarkerLite()");

    l_image = &l_imageL[0];
    work    = &workL[0];
    work2   = &work2L[0];
    wlabel_num = &wlabel_numL;
    warea   = &wareaL[0];
    wclip   = &wclipL[0];
    wpos    = &wposL[0];


	if(pixelFormat!=PIXEL_FORMAT_RGB565 && pixelFormat!=PIXEL_FORMAT_LUM)
		thresh *= 3;

    if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) {
        lxsize = arImXsize / 2;
        lysize = arImYsize / 2;
    }
    else {
        lxsize = arImXsize;
        lysize = arImYsize;
    }

    pnt1 = &l_image[0];
    pnt2 = &l_image[(lysize-1)*lxsize];
    for(i = 0; i < lxsize; i++) {
        *(pnt1++) = *(pnt2++) = 0;
    }

    pnt1 = &l_image[0];
    pnt2 = &l_image[lxsize-1];
    for(i = 0; i < lysize; i++) {
        *pnt1 = *pnt2 = 0;
        pnt1 += lxsize;
        pnt2 += lxsize;
    }

    wk_max = 0;
    pnt2 = &(l_image[lxsize+1]);
    if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) {
        pnt = &(image[(arImXsize*2+2)*pixelSize]);
        poff = pixelSize*2;
    }
    else {
        pnt = &(image[(arImXsize+1)*pixelSize]);
        poff = pixelSize;
    }


//	int diffCorners = -60,
//		diffLeftRight = -40,
//		difftopBottom = -20;

	const int shiftBits = 10;
	int iHalf=lxsize/2, jHalf=lysize/2;

	int threshFact = (pixelFormat!=PIXEL_FORMAT_RGB565 && pixelFormat!=PIXEL_FORMAT_LUM) ? 3 : 1;

	int corrLeftY = (vignetting.corners*threshFact)<<shiftBits,
		dCorrLeftY = ((vignetting.leftright-vignetting.corners*threshFact)<<shiftBits)/jHalf,
		corrCenterY = (vignetting.bottomtop*threshFact)<<shiftBits,
		dCorrCenterY = -corrCenterY/jHalf,
		corrX = 0, dCorrX = 0,
		corrThresh;


	for(j = 1; j < lysize-1; j++, pnt+=poff*2, pnt2+=2)
	{
		if(vignetting.enabled)
		{
			corrX = corrLeftY;
			dCorrX = (corrCenterY-corrLeftY)/iHalf;

			if(j==jHalf)
			{
				dCorrLeftY = -dCorrLeftY;
				dCorrCenterY = -dCorrCenterY;
			
			}

			corrLeftY += dCorrLeftY;
			corrCenterY += dCorrCenterY;
		}

		for(i = 1; i < lxsize-1; i++, pnt+=poff, pnt2++)
		{
			if(vignetting.enabled)
			{
				if(i==iHalf)
					dCorrX = -dCorrX;
				corrX += dCorrX;

				corrThresh = thresh + (corrX>>shiftBits);
			}
			else
				corrThresh = thresh;

			bool isBlack = false;


#ifdef _DEF_PIXEL_FORMAT_ABGR
	            isBlack = ( *(pnt+1) + *(pnt+2) + *(pnt+3) <= corrThresh );
#endif
#ifdef _DEF_PIXEL_FORMAT_BGR
				isBlack = ( *(pnt+0) + *(pnt+1) + *(pnt+2) <= corrThresh );
#endif
#ifdef _DEF_PIXEL_FORMAT_RGB
				isBlack = ( *(pnt+0) + *(pnt+1) + *(pnt+2) <= corrThresh );
#endif
#ifdef _DEF_PIXEL_FORMAT_RGB565
				isBlack = (getLUM8_from_RGB565(pnt) <= corrThresh );
#endif
#ifdef _DEF_PIXEL_FORMAT_LUM
				isBlack = ( *pnt <= corrThresh );
#endif

			if(isBlack) {
				pnt1 = &(pnt2[-lxsize]);
                if( *pnt1 > 0 ) {
                    *pnt2 = *pnt1;

#ifdef _DISABLE_TP_OPTIMIZATIONS_
					// ORIGINAL CODE
					work2[((*pnt2)-1)*7+0] ++;
                    work2[((*pnt2)-1)*7+1] += i;
                    work2[((*pnt2)-1)*7+2] += j;
                    work2[((*pnt2)-1)*7+6] = j;
#else
					// OPTIMIZED CODE [tp]
					// ((*pnt2)-1)*7 should be treated as constant, since
					//  work2[n] (n=0..xsize*ysize) cannot overwrite (*pnt2)
					pnt2_index = ((*pnt2)-1) * 7;
                    work2[pnt2_index+0]++;
                    work2[pnt2_index+1]+= i;
                    work2[pnt2_index+2]+= j;
                    work2[pnt2_index+6] = j;
					// --------------------------------
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                }
                else if( *(pnt1+1) > 0 ) {
                    if( *(pnt1-1) > 0 ) {
                        m = work[*(pnt1+1)-1];
                        n = work[*(pnt1-1)-1];
                        if( m > n ) {
                            *pnt2 = n;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == m ) *wk = n;
                                wk++;
                            }
                        }
                        else if( m < n ) {
                            *pnt2 = m;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == n ) *wk = m;
                                wk++;
                            }
                        }
                        else *pnt2 = m;

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
						work2[((*pnt2)-1)*7+0] ++;
                        work2[((*pnt2)-1)*7+1] += i;
                        work2[((*pnt2)-1)*7+2] += j;
                        work2[((*pnt2)-1)*7+6] = j;
#else
						// PERFORMANCE OPTIMIZATION:
						pnt2_index = ((*pnt2)-1) * 7;
						work2[pnt2_index+0]++;
						work2[pnt2_index+1]+= i;
						work2[pnt2_index+2]+= j;
						work2[pnt2_index+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                    }
                    else if( *(pnt2-1) > 0 ) {
                        m = work[*(pnt1+1)-1];
                        n = work[*(pnt2-1)-1];
                        if( m > n ) {
                            *pnt2 = n;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == m ) *wk = n;
                                wk++;
                            }
                        }
                        else if( m < n ) {
                            *pnt2 = m;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == n ) *wk = m;
                                wk++;
                            }
                        }
                        else *pnt2 = m;

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                        work2[((*pnt2)-1)*7+0] ++;
                        work2[((*pnt2)-1)*7+1] += i;
                        work2[((*pnt2)-1)*7+2] += j;
#else
						// PERFORMANCE OPTIMIZATION:
						pnt2_index = ((*pnt2)-1) * 7;
						work2[pnt2_index+0]++;
						work2[pnt2_index+1]+= i;
						work2[pnt2_index+2]+= j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                    }
                    else {
                        *pnt2 = *(pnt1+1);

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                        work2[((*pnt2)-1)*7+0] ++;
                        work2[((*pnt2)-1)*7+1] += i;
                        work2[((*pnt2)-1)*7+2] += j;
                        if( work2[((*pnt2)-1)*7+3] > i ) work2[((*pnt2)-1)*7+3] = i;
                        work2[((*pnt2)-1)*7+6] = j;
#else
						// PERFORMANCE OPTIMIZATION:
						pnt2_index = ((*pnt2)-1) * 7;
						work2[pnt2_index+0]++;
						work2[pnt2_index+1]+= i;
						work2[pnt2_index+2]+= j;
                        if( work2[pnt2_index+3] > i ) work2[pnt2_index+3] = i;
						work2[pnt2_index+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                    }
                }
                else if( *(pnt1-1) > 0 ) {
                    *pnt2 = *(pnt1-1);

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                    work2[((*pnt2)-1)*7+0] ++;
                    work2[((*pnt2)-1)*7+1] += i;
                    work2[((*pnt2)-1)*7+2] += j;
                    if( work2[((*pnt2)-1)*7+4] < i ) work2[((*pnt2)-1)*7+4] = i;
                    work2[((*pnt2)-1)*7+6] = j;
#else
					// PERFORMANCE OPTIMIZATION:
					pnt2_index = ((*pnt2)-1) * 7;
					work2[pnt2_index+0]++;
					work2[pnt2_index+1]+= i;
					work2[pnt2_index+2]+= j;
                    if( work2[pnt2_index+4] < i ) work2[pnt2_index+4] = i;
					work2[pnt2_index+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                }
                else if( *(pnt2-1) > 0) {
                    *pnt2 = *(pnt2-1);

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                    work2[((*pnt2)-1)*7+0] ++;
                    work2[((*pnt2)-1)*7+1] += i;
                    work2[((*pnt2)-1)*7+2] += j;
                    if( work2[((*pnt2)-1)*7+4] < i ) work2[((*pnt2)-1)*7+4] = i;
#else
					// PERFORMANCE OPTIMIZATION:
					pnt2_index = ((*pnt2)-1) * 7;
					work2[pnt2_index+0]++;
					work2[pnt2_index+1]+= i;
					work2[pnt2_index+2]+= j;
                    if( work2[pnt2_index+4] < i ) work2[pnt2_index+4] = i;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                }
                else {
                    wk_max++;
                    if( wk_max > WORK_SIZE ) {
                        return(0);
                    }
                    work[wk_max-1] = *pnt2 = wk_max;
#ifdef _DISABLE_TP_OPTIMIZATIONS_
                    work2[(wk_max-1)*7+0] = 1;
                    work2[(wk_max-1)*7+1] = i;
                    work2[(wk_max-1)*7+2] = j;
                    work2[(wk_max-1)*7+3] = i;
                    work2[(wk_max-1)*7+4] = i;
                    work2[(wk_max-1)*7+5] = j;
                    work2[(wk_max-1)*7+6] = j;
#else
					wmax_idx = (wk_max-1)*7;
                    work2[wmax_idx+0] = 1;
                    work2[wmax_idx+1] = i;
                    work2[wmax_idx+2] = j;
                    work2[wmax_idx+3] = i;
                    work2[wmax_idx+4] = i;
                    work2[wmax_idx+5] = j;
                    work2[wmax_idx+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_
                }
            }
            else {
                *pnt2 = 0;
            }

		}	// end for x
		
		if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) pnt += arImXsize*pixelSize;

	}	// end for y


    j = 1;
    wk = &(work[0]);
    for(i = 1; i <= wk_max; i++, wk++) {
        *wk = (*wk==i)? j++: work[(*wk)-1];
    }
    *label_num = *wlabel_num = j - 1;
    if( *label_num == 0 ) {
        return( l_image );
    }

    put_zero( (ARUint8 *)warea, *label_num *     sizeof(int) );
    put_zero( (ARUint8 *)wpos,  *label_num * 2 * sizeof(ARFloat) );

#ifdef _DISABLE_TP_OPTIMIZATIONS_
    for(i = 0; i < *label_num; i++) {
        wclip[i*4+0] = lxsize;
        wclip[i*4+1] = 0;
        wclip[i*4+2] = lysize;
        wclip[i*4+3] = 0;
    }
    for(i = 0; i < wk_max; i++) {
        j = work[i] - 1;
        warea[j]    += work2[i*7+0];
        wpos[j*2+0] += work2[i*7+1];
        wpos[j*2+1] += work2[i*7+2];
        if( wclip[j*4+0] > work2[i*7+3] ) wclip[j*4+0] = work2[i*7+3];
        if( wclip[j*4+1] < work2[i*7+4] ) wclip[j*4+1] = work2[i*7+4];
        if( wclip[j*4+2] > work2[i*7+5] ) wclip[j*4+2] = work2[i*7+5];
        if( wclip[j*4+3] < work2[i*7+6] ) wclip[j*4+3] = work2[i*7+6];
    }

    for( i = 0; i < *label_num; i++ ) {
        wpos[i*2+0] /= warea[i];
        wpos[i*2+1] /= warea[i];
    }
#else
	int *wclipRun, *work2Run;
	int iDown;

	wclipRun = wclip;
	iDown = *label_num+1;
	while(--iDown) {
		*wclipRun++ = lxsize;
		*wclipRun++ = 0;
		*wclipRun++ = lysize;
		*wclipRun++ = 0;
    }

	work2Run = work2;
    for(i = 0; i < wk_max; i++) {
        j = work[i] - 1;

        warea[j]    += *work2Run++;
        wpos[j*2+0] += *work2Run++;
        wpos[j*2+1] += *work2Run++;

		wclipRun = wclip+j*4;

		if(*wclipRun > *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;

		if(*wclipRun < *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;

		if(*wclipRun > *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;

		if(*wclipRun < *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;
    }


    for( i = 0; i < *label_num; i++ ) {
        wpos[i*2+0] /= warea[i];
        wpos[i*2+1] /= warea[i];
    }
#endif //!_DISABLE_TP_OPTIMIZATIONS_

    *label_ref = work;
    *area      = warea;
    *pos       = wpos;
    *clip      = wclip;
    return( l_image );
}


#undef LABEL_FUNC_NAME
// End of arLabelingImpl.hxx
#undef _DEF_PIXEL_FORMAT_RGB565

#define _DEF_PIXEL_FORMAT_LUM
#define LABEL_FUNC_NAME arLabeling_LUM
// Included arLabelingImpl.hxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id$
 * @file
 * ======================================================================== */


AR_TEMPL_FUNC ARInt16*
AR_TEMPL_TRACKER::LABEL_FUNC_NAME(ARUint8 *image, int thresh, int *label_num, int **area,
								  ARFloat **pos, int **clip, int **label_ref)
{
    ARUint8   *pnt;                     /*  image pointer       */
    ARInt16   *pnt1, *pnt2;             /*  image pointer       */
    int       *wk;                      /*  pointer for work    */
    int       wk_max;                   /*  work                */
    int       m,n;                      /*  work                */
    int       i,j,k;                    /*  for loop            */
    int       lxsize, lysize;
    int       poff;
    ARInt16   *l_image;
    int       *work, *work2;
    int       *wlabel_num;
    int       *warea;
    int       *wclip;
    ARFloat    *wpos;
#ifndef _DISABLE_TP_OPTIMIZATIONS_
	int		  pnt2_index, wmax_idx;   // [t.pintaric]
#else
	#pragma message(">> Performance Warning: arlabeling() optimizations disabled.")
#endif //_!DISABLE_TP_OPTIMIZATIONS_

	if(pixelFormat==PIXEL_FORMAT_RGB565)
		checkRGB565LUT();


	assert(l_imageL && "checkImageBuffer() must be called before labeling2(). this should happen automatically in arDetectMarker() & arDetectMarkerLite()");

    l_image = &l_imageL[0];
    work    = &workL[0];
    work2   = &work2L[0];
    wlabel_num = &wlabel_numL;
    warea   = &wareaL[0];
    wclip   = &wclipL[0];
    wpos    = &wposL[0];


	if(pixelFormat!=PIXEL_FORMAT_RGB565 && pixelFormat!=PIXEL_FORMAT_LUM)
		thresh *= 3;

    if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) {
        lxsize = arImXsize / 2;
        lysize = arImYsize / 2;
    }
    else {
        lxsize = arImXsize;
        lysize = arImYsize;
    }

    pnt1 = &l_image[0];
    pnt2 = &l_image[(lysize-1)*lxsize];
    for(i = 0; i < lxsize; i++) {
        *(pnt1++) = *(pnt2++) = 0;
    }

    pnt1 = &l_image[0];
    pnt2 = &l_image[lxsize-1];
    for(i = 0; i < lysize; i++) {
        *pnt1 = *pnt2 = 0;
        pnt1 += lxsize;
        pnt2 += lxsize;
    }

    wk_max = 0;
    pnt2 = &(l_image[lxsize+1]);
    if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) {
        pnt = &(image[(arImXsize*2+2)*pixelSize]);
        poff = pixelSize*2;
    }
    else {
        pnt = &(image[(arImXsize+1)*pixelSize]);
        poff = pixelSize;
    }


//	int diffCorners = -60,
//		diffLeftRight = -40,
//		difftopBottom = -20;

	const int shiftBits = 10;
	int iHalf=lxsize/2, jHalf=lysize/2;

	int threshFact = (pixelFormat!=PIXEL_FORMAT_RGB565 && pixelFormat!=PIXEL_FORMAT_LUM) ? 3 : 1;

	int corrLeftY = (vignetting.corners*threshFact)<<shiftBits,
		dCorrLeftY = ((vignetting.leftright-vignetting.corners*threshFact)<<shiftBits)/jHalf,
		corrCenterY = (vignetting.bottomtop*threshFact)<<shiftBits,
		dCorrCenterY = -corrCenterY/jHalf,
		corrX = 0, dCorrX = 0,
		corrThresh;


	for(j = 1; j < lysize-1; j++, pnt+=poff*2, pnt2+=2)
	{
		if(vignetting.enabled)
		{
			corrX = corrLeftY;
			dCorrX = (corrCenterY-corrLeftY)/iHalf;

			if(j==jHalf)
			{
				dCorrLeftY = -dCorrLeftY;
				dCorrCenterY = -dCorrCenterY;
			
			}

			corrLeftY += dCorrLeftY;
			corrCenterY += dCorrCenterY;
		}

		for(i = 1; i < lxsize-1; i++, pnt+=poff, pnt2++)
		{
			if(vignetting.enabled)
			{
				if(i==iHalf)
					dCorrX = -dCorrX;
				corrX += dCorrX;

				corrThresh = thresh + (corrX>>shiftBits);
			}
			else
				corrThresh = thresh;

			bool isBlack = false;


#ifdef _DEF_PIXEL_FORMAT_ABGR
	            isBlack = ( *(pnt+1) + *(pnt+2) + *(pnt+3) <= corrThresh );
#endif
#ifdef _DEF_PIXEL_FORMAT_BGR
				isBlack = ( *(pnt+0) + *(pnt+1) + *(pnt+2) <= corrThresh );
#endif
#ifdef _DEF_PIXEL_FORMAT_RGB
				isBlack = ( *(pnt+0) + *(pnt+1) + *(pnt+2) <= corrThresh );
#endif
#ifdef _DEF_PIXEL_FORMAT_RGB565
				isBlack = (getLUM8_from_RGB565(pnt) <= corrThresh );
#endif
#ifdef _DEF_PIXEL_FORMAT_LUM
				isBlack = ( *pnt <= corrThresh );
#endif

			if(isBlack) {
				pnt1 = &(pnt2[-lxsize]);
                if( *pnt1 > 0 ) {
                    *pnt2 = *pnt1;

#ifdef _DISABLE_TP_OPTIMIZATIONS_
					// ORIGINAL CODE
					work2[((*pnt2)-1)*7+0] ++;
                    work2[((*pnt2)-1)*7+1] += i;
                    work2[((*pnt2)-1)*7+2] += j;
                    work2[((*pnt2)-1)*7+6] = j;
#else
					// OPTIMIZED CODE [tp]
					// ((*pnt2)-1)*7 should be treated as constant, since
					//  work2[n] (n=0..xsize*ysize) cannot overwrite (*pnt2)
					pnt2_index = ((*pnt2)-1) * 7;
                    work2[pnt2_index+0]++;
                    work2[pnt2_index+1]+= i;
                    work2[pnt2_index+2]+= j;
                    work2[pnt2_index+6] = j;
					// --------------------------------
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                }
                else if( *(pnt1+1) > 0 ) {
                    if( *(pnt1-1) > 0 ) {
                        m = work[*(pnt1+1)-1];
                        n = work[*(pnt1-1)-1];
                        if( m > n ) {
                            *pnt2 = n;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == m ) *wk = n;
                                wk++;
                            }
                        }
                        else if( m < n ) {
                            *pnt2 = m;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == n ) *wk = m;
                                wk++;
                            }
                        }
                        else *pnt2 = m;

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
						work2[((*pnt2)-1)*7+0] ++;
                        work2[((*pnt2)-1)*7+1] += i;
                        work2[((*pnt2)-1)*7+2] += j;
                        work2[((*pnt2)-1)*7+6] = j;
#else
						// PERFORMANCE OPTIMIZATION:
						pnt2_index = ((*pnt2)-1) * 7;
						work2[pnt2_index+0]++;
						work2[pnt2_index+1]+= i;
						work2[pnt2_index+2]+= j;
						work2[pnt2_index+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                    }
                    else if( *(pnt2-1) > 0 ) {
                        m = work[*(pnt1+1)-1];
                        n = work[*(pnt2-1)-1];
                        if( m > n ) {
                            *pnt2 = n;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == m ) *wk = n;
                                wk++;
                            }
                        }
                        else if( m < n ) {
                            *pnt2 = m;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == n ) *wk = m;
                                wk++;
                            }
                        }
                        else *pnt2 = m;

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                        work2[((*pnt2)-1)*7+0] ++;
                        work2[((*pnt2)-1)*7+1] += i;
                        work2[((*pnt2)-1)*7+2] += j;
#else
						// PERFORMANCE OPTIMIZATION:
						pnt2_index = ((*pnt2)-1) * 7;
						work2[pnt2_index+0]++;
						work2[pnt2_index+1]+= i;
						work2[pnt2_index+2]+= j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                    }
                    else {
                        *pnt2 = *(pnt1+1);

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                        work2[((*pnt2)-1)*7+0] ++;
                        work2[((*pnt2)-1)*7+1] += i;
                        work2[((*pnt2)-1)*7+2] += j;
                        if( work2[((*pnt2)-1)*7+3] > i ) work2[((*pnt2)-1)*7+3] = i;
                        work2[((*pnt2)-1)*7+6] = j;
#else
						// PERFORMANCE OPTIMIZATION:
						pnt2_index = ((*pnt2)-1) * 7;
						work2[pnt2_index+0]++;
						work2[pnt2_index+1]+= i;
						work2[pnt2_index+2]+= j;
                        if( work2[pnt2_index+3] > i ) work2[pnt2_index+3] = i;
						work2[pnt2_index+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                    }
                }
                else if( *(pnt1-1) > 0 ) {
                    *pnt2 = *(pnt1-1);

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                    work2[((*pnt2)-1)*7+0] ++;
                    work2[((*pnt2)-1)*7+1] += i;
                    work2[((*pnt2)-1)*7+2] += j;
                    if( work2[((*pnt2)-1)*7+4] < i ) work2[((*pnt2)-1)*7+4] = i;
                    work2[((*pnt2)-1)*7+6] = j;
#else
					// PERFORMANCE OPTIMIZATION:
					pnt2_index = ((*pnt2)-1) * 7;
					work2[pnt2_index+0]++;
					work2[pnt2_index+1]+= i;
					work2[pnt2_index+2]+= j;
                    if( work2[pnt2_index+4] < i ) work2[pnt2_index+4] = i;
					work2[pnt2_index+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                }
                else if( *(pnt2-1) > 0) {
                    *pnt2 = *(pnt2-1);

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                    work2[((*pnt2)-1)*7+0] ++;
                    work2[((*pnt2)-1)*7+1] += i;
                    work2[((*pnt2)-1)*7+2] += j;
                    if( work2[((*pnt2)-1)*7+4] < i ) work2[((*pnt2)-1)*7+4] = i;
#else
					// PERFORMANCE OPTIMIZATION:
					pnt2_index = ((*pnt2)-1) * 7;
					work2[pnt2_index+0]++;
					work2[pnt2_index+1]+= i;
					work2[pnt2_index+2]+= j;
                    if( work2[pnt2_index+4] < i ) work2[pnt2_index+4] = i;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                }
                else {
                    wk_max++;
                    if( wk_max > WORK_SIZE ) {
                        return(0);
                    }
                    work[wk_max-1] = *pnt2 = wk_max;
#ifdef _DISABLE_TP_OPTIMIZATIONS_
                    work2[(wk_max-1)*7+0] = 1;
                    work2[(wk_max-1)*7+1] = i;
                    work2[(wk_max-1)*7+2] = j;
                    work2[(wk_max-1)*7+3] = i;
                    work2[(wk_max-1)*7+4] = i;
                    work2[(wk_max-1)*7+5] = j;
                    work2[(wk_max-1)*7+6] = j;
#else
					wmax_idx = (wk_max-1)*7;
                    work2[wmax_idx+0] = 1;
                    work2[wmax_idx+1] = i;
                    work2[wmax_idx+2] = j;
                    work2[wmax_idx+3] = i;
                    work2[wmax_idx+4] = i;
                    work2[wmax_idx+5] = j;
                    work2[wmax_idx+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_
                }
            }
            else {
                *pnt2 = 0;
            }

		}	// end for x
		
		if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) pnt += arImXsize*pixelSize;

	}	// end for y


    j = 1;
    wk = &(work[0]);
    for(i = 1; i <= wk_max; i++, wk++) {
        *wk = (*wk==i)? j++: work[(*wk)-1];
    }
    *label_num = *wlabel_num = j - 1;
    if( *label_num == 0 ) {
        return( l_image );
    }

    put_zero( (ARUint8 *)warea, *label_num *     sizeof(int) );
    put_zero( (ARUint8 *)wpos,  *label_num * 2 * sizeof(ARFloat) );

#ifdef _DISABLE_TP_OPTIMIZATIONS_
    for(i = 0; i < *label_num; i++) {
        wclip[i*4+0] = lxsize;
        wclip[i*4+1] = 0;
        wclip[i*4+2] = lysize;
        wclip[i*4+3] = 0;
    }
    for(i = 0; i < wk_max; i++) {
        j = work[i] - 1;
        warea[j]    += work2[i*7+0];
        wpos[j*2+0] += work2[i*7+1];
        wpos[j*2+1] += work2[i*7+2];
        if( wclip[j*4+0] > work2[i*7+3] ) wclip[j*4+0] = work2[i*7+3];
        if( wclip[j*4+1] < work2[i*7+4] ) wclip[j*4+1] = work2[i*7+4];
        if( wclip[j*4+2] > work2[i*7+5] ) wclip[j*4+2] = work2[i*7+5];
        if( wclip[j*4+3] < work2[i*7+6] ) wclip[j*4+3] = work2[i*7+6];
    }

    for( i = 0; i < *label_num; i++ ) {
        wpos[i*2+0] /= warea[i];
        wpos[i*2+1] /= warea[i];
    }
#else
	int *wclipRun, *work2Run;
	int iDown;

	wclipRun = wclip;
	iDown = *label_num+1;
	while(--iDown) {
		*wclipRun++ = lxsize;
		*wclipRun++ = 0;
		*wclipRun++ = lysize;
		*wclipRun++ = 0;
    }

	work2Run = work2;
    for(i = 0; i < wk_max; i++) {
        j = work[i] - 1;

        warea[j]    += *work2Run++;
        wpos[j*2+0] += *work2Run++;
        wpos[j*2+1] += *work2Run++;

		wclipRun = wclip+j*4;

		if(*wclipRun > *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;

		if(*wclipRun < *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;

		if(*wclipRun > *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;

		if(*wclipRun < *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;
    }


    for( i = 0; i < *label_num; i++ ) {
        wpos[i*2+0] /= warea[i];
        wpos[i*2+1] /= warea[i];
    }
#endif //!_DISABLE_TP_OPTIMIZATIONS_

    *label_ref = work;
    *area      = warea;
    *pos       = wpos;
    *clip      = wclip;
    return( l_image );
}


#undef LABEL_FUNC_NAME
// End of arLabelingImpl.hxx
#undef _DEF_PIXEL_FORMAT_LUM


AR_TEMPL_FUNC ARInt16*
AR_TEMPL_TRACKER::arLabeling(ARUint8 *image, int thresh, int *label_num, int **area,
					ARFloat **pos, int **clip, int **label_ref )
{
	ARInt16* ret = NULL;

	PROFILE_BEGINSEC(profiler, LABELING)
	//ret = labeling2(image, thresh, label_num, area, pos, clip, label_ref, 1);

	switch(pixelFormat)
	{
	case PIXEL_FORMAT_ABGR:
		ret = arLabeling_ABGR(image, thresh, label_num, area, pos, clip, label_ref);
		break;

	case PIXEL_FORMAT_BGRA:
	case PIXEL_FORMAT_BGR:
		ret = arLabeling_BGR(image, thresh, label_num, area, pos, clip, label_ref);
		break;

	case PIXEL_FORMAT_RGBA:
	case PIXEL_FORMAT_RGB:
		ret = arLabeling_RGB(image, thresh, label_num, area, pos, clip, label_ref);
		break;

	case PIXEL_FORMAT_RGB565:
		ret = arLabeling_RGB565(image, thresh, label_num, area, pos, clip, label_ref);
		break;

	case PIXEL_FORMAT_LUM:
		ret = arLabeling_LUM(image, thresh, label_num, area, pos, clip, label_ref);
		break;
	}

    PROFILE_ENDSEC(profiler, LABELING)

	return ret;
}



#if 0
AR_TEMPL_FUNC ARInt16*
AR_TEMPL_TRACKER::labeling2(ARUint8 *image, int thresh, int *label_num, int **area,
				   ARFloat **pos, int **clip, int **label_ref, int LorR)
{
    ARUint8   *pnt;                     /*  image pointer       */
    ARInt16   *pnt1, *pnt2;             /*  image pointer       */
    int       *wk;                      /*  pointer for work    */
    int       wk_max;                   /*  work                */
    int       m,n;                      /*  work                */
    int       i,j,k;                    /*  for loop            */
    int       lxsize, lysize;
    int       poff;
    ARInt16   *l_image;
    int       *work, *work2;
    int       *wlabel_num;
    int       *warea;
    int       *wclip;
    ARFloat    *wpos;
#ifndef _DISABLE_TP_OPTIMIZATIONS_
	int		  pnt2_index, wmax_idx;   // [t.pintaric]
#else
	#pragma message(">> Performance Warning: arlabeling() optimizations disabled.")
#endif //_!DISABLE_TP_OPTIMIZATIONS_

//#ifdef AR_PIX_FORMAT_RGB565
	if(PIX_FORMAT==PIXEL_FORMAT_RGB565)
		checkRGB565LUT();
//#endif

	THESHOLD_FUNC threshFunc = thresholdLUM8;


	assert(l_imageL && "checkImageBuffer() must be called before labeling2(). this should happen automatically in arDetectMarker() & arDetectMarkerLite()");

    if( LorR ) {
        l_image = &l_imageL[0];
        work    = &workL[0];
        work2   = &work2L[0];
        wlabel_num = &wlabel_numL;
        warea   = &wareaL[0];
        wclip   = &wclipL[0];
        wpos    = &wposL[0];
    }
    else {
        l_image = &l_imageR[0];
        work    = &workR[0];
        work2   = &work2R[0];
        wlabel_num = &wlabel_numR;
        warea   = &wareaR[0];
        wclip   = &wclipR[0];
        wpos    = &wposR[0];
    }


	if(PIX_FORMAT!=PIXEL_FORMAT_RGB565 && PIX_FORMAT!=PIXEL_FORMAT_LUM)
		thresh *= 3;

    if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) {
        lxsize = arImXsize / 2;
        lysize = arImYsize / 2;
    }
    else {
        lxsize = arImXsize;
        lysize = arImYsize;
    }

    pnt1 = &l_image[0];
    pnt2 = &l_image[(lysize-1)*lxsize];
    for(i = 0; i < lxsize; i++) {
        *(pnt1++) = *(pnt2++) = 0;
    }

    pnt1 = &l_image[0];
    pnt2 = &l_image[lxsize-1];
    for(i = 0; i < lysize; i++) {
        *pnt1 = *pnt2 = 0;
        pnt1 += lxsize;
        pnt2 += lxsize;
    }

    wk_max = 0;
    pnt2 = &(l_image[lxsize+1]);
    if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) {
        pnt = &(image[(arImXsize*2+2)*PIX_SIZE]);
        poff = PIX_SIZE*2;
    }
    else {
        pnt = &(image[(arImXsize+1)*PIX_SIZE]);
        poff = PIX_SIZE;
    }


//	int diffCorners = -60,
//		diffLeftRight = -40,
//		difftopBottom = -20;

	const int shiftBits = 10;
	int iHalf=lxsize/2, jHalf=lysize/2;

	int threshFact = (PIX_FORMAT!=PIXEL_FORMAT_RGB565 && PIX_FORMAT!=PIXEL_FORMAT_LUM) ? 3 : 1;

	int corrLeftY = (vignetting.corners*threshFact)<<shiftBits,
		dCorrLeftY = ((vignetting.leftright-vignetting.corners*threshFact)<<shiftBits)/jHalf,
		corrCenterY = (vignetting.bottomtop*threshFact)<<shiftBits,
		dCorrCenterY = -corrCenterY/jHalf,
		corrX, dCorrX,
		corrThresh;


	for(j = 1; j < lysize-1; j++, pnt+=poff*2, pnt2+=2)
	{
		if(vignetting.enabled)
		{
			corrX = corrLeftY;
			dCorrX = (corrCenterY-corrLeftY)/iHalf;

			if(j==jHalf)
			{
				dCorrLeftY = -dCorrLeftY;
				dCorrCenterY = -dCorrCenterY;
			
			}

			corrLeftY += dCorrLeftY;
			corrCenterY += dCorrCenterY;
		}

		for(i = 1; i < lxsize-1; i++, pnt+=poff, pnt2++)
		{
			if(vignetting.enabled)
			{
				if(i==iHalf)
					dCorrX = -dCorrX;
				corrX += dCorrX;

				corrThresh = thresh + (corrX>>shiftBits);
			}
			else
				corrThresh = thresh;

			bool isBlack = false;

			if(PIX_FORMAT==PIXEL_FORMAT_ABGR)
	            isBlack = ( *(pnt+1) + *(pnt+2) + *(pnt+3) <= corrThresh );
			if(PIX_FORMAT==PIXEL_FORMAT_BGRA)
				isBlack = ( *(pnt+0) + *(pnt+1) + *(pnt+2) <= corrThresh );
			if(PIX_FORMAT==PIXEL_FORMAT_BGR)
				isBlack = ( *(pnt+0) + *(pnt+1) + *(pnt+2) <= corrThresh );
			if(PIX_FORMAT==PIXEL_FORMAT_RGBA)
				isBlack = ( *(pnt+0) + *(pnt+1) + *(pnt+2) <= corrThresh );
			if(PIX_FORMAT==PIXEL_FORMAT_RGB)
				isBlack = ( *(pnt+0) + *(pnt+1) + *(pnt+2) <= corrThresh );
			if(PIX_FORMAT==PIXEL_FORMAT_RGB565)
				isBlack = (getLUM8_from_RGB565(pnt) <= corrThresh );
			if(PIX_FORMAT==PIXEL_FORMAT_LUM)
				isBlack = ( *pnt <= corrThresh );

			if(isBlack) {
				pnt1 = &(pnt2[-lxsize]);
                if( *pnt1 > 0 ) {
                    *pnt2 = *pnt1;

#ifdef _DISABLE_TP_OPTIMIZATIONS_
					// ORIGINAL CODE
					work2[((*pnt2)-1)*7+0] ++;
                    work2[((*pnt2)-1)*7+1] += i;
                    work2[((*pnt2)-1)*7+2] += j;
                    work2[((*pnt2)-1)*7+6] = j;
#else
					// OPTIMIZED CODE [tp]
					// ((*pnt2)-1)*7 should be treated as constant, since
					//  work2[n] (n=0..xsize*ysize) cannot overwrite (*pnt2)
					pnt2_index = ((*pnt2)-1) * 7;
                    work2[pnt2_index+0]++;
                    work2[pnt2_index+1]+= i;
                    work2[pnt2_index+2]+= j;
                    work2[pnt2_index+6] = j;
					// --------------------------------
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                }
                else if( *(pnt1+1) > 0 ) {
                    if( *(pnt1-1) > 0 ) {
                        m = work[*(pnt1+1)-1];
                        n = work[*(pnt1-1)-1];
                        if( m > n ) {
                            *pnt2 = n;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == m ) *wk = n;
                                wk++;
                            }
                        }
                        else if( m < n ) {
                            *pnt2 = m;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == n ) *wk = m;
                                wk++;
                            }
                        }
                        else *pnt2 = m;

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
						work2[((*pnt2)-1)*7+0] ++;
                        work2[((*pnt2)-1)*7+1] += i;
                        work2[((*pnt2)-1)*7+2] += j;
                        work2[((*pnt2)-1)*7+6] = j;
#else
						// PERFORMANCE OPTIMIZATION:
						pnt2_index = ((*pnt2)-1) * 7;
						work2[pnt2_index+0]++;
						work2[pnt2_index+1]+= i;
						work2[pnt2_index+2]+= j;
						work2[pnt2_index+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                    }
                    else if( *(pnt2-1) > 0 ) {
                        m = work[*(pnt1+1)-1];
                        n = work[*(pnt2-1)-1];
                        if( m > n ) {
                            *pnt2 = n;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == m ) *wk = n;
                                wk++;
                            }
                        }
                        else if( m < n ) {
                            *pnt2 = m;
                            wk = &(work[0]);
                            for(k = 0; k < wk_max; k++) {
                                if( *wk == n ) *wk = m;
                                wk++;
                            }
                        }
                        else *pnt2 = m;

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                        work2[((*pnt2)-1)*7+0] ++;
                        work2[((*pnt2)-1)*7+1] += i;
                        work2[((*pnt2)-1)*7+2] += j;
#else
						// PERFORMANCE OPTIMIZATION:
						pnt2_index = ((*pnt2)-1) * 7;
						work2[pnt2_index+0]++;
						work2[pnt2_index+1]+= i;
						work2[pnt2_index+2]+= j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                    }
                    else {
                        *pnt2 = *(pnt1+1);

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                        work2[((*pnt2)-1)*7+0] ++;
                        work2[((*pnt2)-1)*7+1] += i;
                        work2[((*pnt2)-1)*7+2] += j;
                        if( work2[((*pnt2)-1)*7+3] > i ) work2[((*pnt2)-1)*7+3] = i;
                        work2[((*pnt2)-1)*7+6] = j;
#else
						// PERFORMANCE OPTIMIZATION:
						pnt2_index = ((*pnt2)-1) * 7;
						work2[pnt2_index+0]++;
						work2[pnt2_index+1]+= i;
						work2[pnt2_index+2]+= j;
                        if( work2[pnt2_index+3] > i ) work2[pnt2_index+3] = i;
						work2[pnt2_index+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                    }
                }
                else if( *(pnt1-1) > 0 ) {
                    *pnt2 = *(pnt1-1);

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                    work2[((*pnt2)-1)*7+0] ++;
                    work2[((*pnt2)-1)*7+1] += i;
                    work2[((*pnt2)-1)*7+2] += j;
                    if( work2[((*pnt2)-1)*7+4] < i ) work2[((*pnt2)-1)*7+4] = i;
                    work2[((*pnt2)-1)*7+6] = j;
#else
					// PERFORMANCE OPTIMIZATION:
					pnt2_index = ((*pnt2)-1) * 7;
					work2[pnt2_index+0]++;
					work2[pnt2_index+1]+= i;
					work2[pnt2_index+2]+= j;
                    if( work2[pnt2_index+4] < i ) work2[pnt2_index+4] = i;
					work2[pnt2_index+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                }
                else if( *(pnt2-1) > 0) {
                    *pnt2 = *(pnt2-1);

#ifdef _DISABLE_TP_OPTIMIZATIONS_
						// ORIGINAL CODE
                    work2[((*pnt2)-1)*7+0] ++;
                    work2[((*pnt2)-1)*7+1] += i;
                    work2[((*pnt2)-1)*7+2] += j;
                    if( work2[((*pnt2)-1)*7+4] < i ) work2[((*pnt2)-1)*7+4] = i;
#else
					// PERFORMANCE OPTIMIZATION:
					pnt2_index = ((*pnt2)-1) * 7;
					work2[pnt2_index+0]++;
					work2[pnt2_index+1]+= i;
					work2[pnt2_index+2]+= j;
                    if( work2[pnt2_index+4] < i ) work2[pnt2_index+4] = i;
#endif //!_DISABLE_TP_OPTIMIZATIONS_

                }
                else {
                    wk_max++;
                    if( wk_max > WORK_SIZE ) {
                        return(0);
                    }
                    work[wk_max-1] = *pnt2 = wk_max;
#ifdef _DISABLE_TP_OPTIMIZATIONS_
                    work2[(wk_max-1)*7+0] = 1;
                    work2[(wk_max-1)*7+1] = i;
                    work2[(wk_max-1)*7+2] = j;
                    work2[(wk_max-1)*7+3] = i;
                    work2[(wk_max-1)*7+4] = i;
                    work2[(wk_max-1)*7+5] = j;
                    work2[(wk_max-1)*7+6] = j;
#else
					wmax_idx = (wk_max-1)*7;
                    work2[wmax_idx+0] = 1;
                    work2[wmax_idx+1] = i;
                    work2[wmax_idx+2] = j;
                    work2[wmax_idx+3] = i;
                    work2[wmax_idx+4] = i;
                    work2[wmax_idx+5] = j;
                    work2[wmax_idx+6] = j;
#endif //!_DISABLE_TP_OPTIMIZATIONS_
                }
            }
            else {
                *pnt2 = 0;
            }

		}	// end for x
		
		if( arImageProcMode == AR_IMAGE_PROC_IN_HALF ) pnt += arImXsize*PIX_SIZE;

	}	// end for y


    j = 1;
    wk = &(work[0]);
    for(i = 1; i <= wk_max; i++, wk++) {
        *wk = (*wk==i)? j++: work[(*wk)-1];
    }
    *label_num = *wlabel_num = j - 1;
    if( *label_num == 0 ) {
        return( l_image );
    }

    put_zero( (ARUint8 *)warea, *label_num *     sizeof(int) );
    put_zero( (ARUint8 *)wpos,  *label_num * 2 * sizeof(ARFloat) );

#ifdef _DISABLE_TP_OPTIMIZATIONS_
    for(i = 0; i < *label_num; i++) {
        wclip[i*4+0] = lxsize;
        wclip[i*4+1] = 0;
        wclip[i*4+2] = lysize;
        wclip[i*4+3] = 0;
    }
    for(i = 0; i < wk_max; i++) {
        j = work[i] - 1;
        warea[j]    += work2[i*7+0];
        wpos[j*2+0] += work2[i*7+1];
        wpos[j*2+1] += work2[i*7+2];
        if( wclip[j*4+0] > work2[i*7+3] ) wclip[j*4+0] = work2[i*7+3];
        if( wclip[j*4+1] < work2[i*7+4] ) wclip[j*4+1] = work2[i*7+4];
        if( wclip[j*4+2] > work2[i*7+5] ) wclip[j*4+2] = work2[i*7+5];
        if( wclip[j*4+3] < work2[i*7+6] ) wclip[j*4+3] = work2[i*7+6];
    }

    for( i = 0; i < *label_num; i++ ) {
        wpos[i*2+0] /= warea[i];
        wpos[i*2+1] /= warea[i];
    }
#else
	int *wclipRun, *work2Run;
	int iDown;

	wclipRun = wclip;
	iDown = *label_num+1;
	while(--iDown) {
		*wclipRun++ = lxsize;
		*wclipRun++ = 0;
		*wclipRun++ = lysize;
		*wclipRun++ = 0;
    }

	work2Run = work2;
    for(i = 0; i < wk_max; i++) {
        j = work[i] - 1;

        warea[j]    += *work2Run++;
        wpos[j*2+0] += *work2Run++;
        wpos[j*2+1] += *work2Run++;

		wclipRun = wclip+j*4;

		if(*wclipRun > *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;

		if(*wclipRun < *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;

		if(*wclipRun > *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;

		if(*wclipRun < *work2Run)
			*wclipRun = *work2Run;
		wclipRun++;
		work2Run++;
    }


    for( i = 0; i < *label_num; i++ ) {
        wpos[i*2+0] /= warea[i];
        wpos[i*2+1] /= warea[i];
    }
#endif //!_DISABLE_TP_OPTIMIZATIONS_

    *label_ref = work;
    *area      = warea;
    *pos       = wpos;
    *clip      = wclip;
    return( l_image );
}
#endif // #if 0


}  // namespace ARToolKitPlus
// End of ../../src/core/arLabeling.cxx
// Included ../../src/core/arMultiActivate.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id: arMultiActivate.cxx 162 2006-04-19 21:28:10Z grabner $
 * @file
 * ======================================================================== */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ARToolKitPlus/Tracker.h>
#include <ARToolKitPlus/matrix.h>
#include <ARToolKitPlus/arMulti.h>


namespace ARToolKitPlus {


AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arMultiActivate( ARMultiMarkerInfoT *config )
{
    int    i;

    config->prevF = 0;

    for( i = 0; i < config->marker_num; i++ ) {
        arActivatePatt( config->marker[i].patt_id );
    }

    return 0;
}

AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arMultiDeactivate( ARMultiMarkerInfoT *config )
{
    int    i;

    config->prevF = 0;

    for( i = 0; i < config->marker_num; i++ ) {
        arDeactivatePatt( config->marker[i].patt_id );
    }

    return 0;
}

AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arMultiFreeConfig( ARMultiMarkerInfoT *config )
{
    int    i;

    for( i = 0; i < config->marker_num; i++ ) {
        arFreePatt( config->marker[i].patt_id );
    }
    free( config->marker );
    free( config );
    config = NULL;

    return 0;
}


}  // namespace ARToolKitPlus
// End of ../../src/core/arMultiActivate.cxx
// Included ../../src/core/arMultiGetTransMat.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id: arMultiGetTransMat.cxx 162 2006-04-19 21:28:10Z grabner $
 * @file
 * ======================================================================== */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ARToolKitPlus/Tracker.h>
#include <ARToolKitPlus/matrix.h>


namespace ARToolKitPlus {


//#define  THRESH_1            2.0
// chan
#define  THRESH_1            3.0
#define  THRESH_2           20.0
#define  THRESH_3           10.0
#define  AR_MULTI_GET_TRANS_MAT_MAX_LOOP_COUNT   2
#define  AR_MULTI_GET_TRANS_MAT_MAX_FIT_ERROR    10.0

typedef struct {
    ARFloat   pos[4][2];
    ARFloat   thresh;
    ARFloat   err;
    int      marker;
    int      dir;
} arMultiEachMarkerInternalInfoT;


AR_TEMPL_FUNC ARFloat
AR_TEMPL_TRACKER::arMultiGetTransMat(ARMarkerInfo *marker_info, int marker_num, ARMultiMarkerInfoT *config)
{
    ARFloat                *pos2d, *pos3d;
    ARFloat                rot[3][3], trans1[3][4], trans2[3][4];
    ARFloat                err = 0, err2;
    int                   max, max_area = 0, max_marker, vnum;
    int                   dir;
    int                   i, j, k;

    if( config->prevF ) {
        verify_markers( marker_info, marker_num, config );
    }

    max = -1;
    vnum = 0;
    for( i = 0; i < config->marker_num; i++ ) {
        k = -1;
        for( j = 0; j < marker_num; j++ ) {
            if( marker_info[j].id != config->marker[i].patt_id ) continue;
            if( marker_info[j].cf < 0.70 ) continue;

            if( k == -1 ) k = j;
            else if( marker_info[k].cf < marker_info[j].cf ) k = j;
        }
        if( (config->marker[i].visible=k) == -1) continue;

		// Changed by Daniel: use the selected pose estimator for this now. i'm though not sure if
		//                    it is wise to use arGetTransMatCont for multi-marker tracking...
		//
        //err = arGetTransMat(&marker_info[k], config->marker[i].center, config->marker[i].width, trans1);
		//err = (this->*poseEstimator_func)(&marker_info[k], config->marker[i].center, config->marker[i].width, trans1);
		err = executeSingleMarkerPoseEstimator(&marker_info[k], config->marker[i].center, config->marker[i].width, trans1);


#ifdef ARTK_DEBUG
	printf("##err = %10.5f %d %10.5f %10.5f\n", err, marker_info[k].dir, marker_info[k].pos[0], marker_info[k].pos[1]);
#endif
        if( err > THRESH_1 ) {
            config->marker[i].visible = -1;
            continue;
        }

        vnum++;
        if( max == -1 
         || marker_info[k].area > max_area ) {
            max = i;
            max_marker = k;
            max_area   = marker_info[k].area;
            for( j = 0; j < 3; j++ ) {
                for( k = 0; k < 4; k++ ) {
                    trans2[j][k] = trans1[j][k];
                }
            }
        }
    }
    if( max == -1 ) {
        config->prevF = 0;
        return -1;
    }

    arMalloc(pos2d, ARFloat, vnum*4*2);
    arMalloc(pos3d, ARFloat, vnum*4*3);

    j = 0;
    for( i = 0; i < config->marker_num; i++ ) {
        if( (k=config->marker[i].visible) < 0 ) continue;

        dir = marker_info[k].dir;
        pos2d[j*8+0] = marker_info[k].vertex[(4-dir)%4][0];
        pos2d[j*8+1] = marker_info[k].vertex[(4-dir)%4][1];
        pos2d[j*8+2] = marker_info[k].vertex[(5-dir)%4][0];
        pos2d[j*8+3] = marker_info[k].vertex[(5-dir)%4][1];
        pos2d[j*8+4] = marker_info[k].vertex[(6-dir)%4][0];
        pos2d[j*8+5] = marker_info[k].vertex[(6-dir)%4][1];
        pos2d[j*8+6] = marker_info[k].vertex[(7-dir)%4][0];
        pos2d[j*8+7] = marker_info[k].vertex[(7-dir)%4][1];
        pos3d[j*12+0] = config->marker[i].pos3d[0][0];
        pos3d[j*12+1] = config->marker[i].pos3d[0][1];
        pos3d[j*12+2] = config->marker[i].pos3d[0][2];
        pos3d[j*12+3] = config->marker[i].pos3d[1][0];
        pos3d[j*12+4] = config->marker[i].pos3d[1][1];
        pos3d[j*12+5] = config->marker[i].pos3d[1][2];
        pos3d[j*12+6] = config->marker[i].pos3d[2][0];
        pos3d[j*12+7] = config->marker[i].pos3d[2][1];
        pos3d[j*12+8] = config->marker[i].pos3d[2][2];
        pos3d[j*12+9] = config->marker[i].pos3d[3][0];
        pos3d[j*12+10] = config->marker[i].pos3d[3][1];
        pos3d[j*12+11] = config->marker[i].pos3d[3][2];
        j++;
    }

    if( config->prevF ) {
        for( j = 0; j < 3; j++ ) {
            for( i = 0; i < 3; i++ ) {
                rot[j][i] = config->trans[j][i];
            }
        }
        for( i = 0; i < AR_MULTI_GET_TRANS_MAT_MAX_LOOP_COUNT; i++ ) {
            err = arGetTransMat4( rot, (ARFloat (*)[2])pos2d,
                                       (ARFloat (*)[3])pos3d,
                                        vnum*4, config->trans );
            if( err < AR_MULTI_GET_TRANS_MAT_MAX_FIT_ERROR ) break;
        }

        if( err < THRESH_2 ) {
            config->prevF = 1;
            free(pos3d);
            free(pos2d);
            return err;
        }
    }

    arUtilMatMul( trans2, config->marker[max].itrans, trans1 );
    for( j = 0; j < 3; j++ ) {
        for( i = 0; i < 3; i++ ) {
            rot[j][i] = trans1[j][i];
        }
    }

    for( i = 0; i < AR_MULTI_GET_TRANS_MAT_MAX_LOOP_COUNT; i++ ) {
        err2 = arGetTransMat4( rot, (ARFloat (*)[2])pos2d, (ARFloat (*)[3])pos3d,
                              vnum*4, trans2 );
        if( err2 < AR_MULTI_GET_TRANS_MAT_MAX_FIT_ERROR ) break;
    }

    if( config->prevF == 0 || err2 < err ) {
        for( j = 0; j < 3; j++ ) {
            for( i = 0; i < 4; i++ ) {
                config->trans[j][i] = trans2[j][i];
            }
        }
        err = err2;
    }

    if( err < THRESH_3 ) {
        config->prevF = 1;
    }
    else {
        config->prevF = 0;
    }

    free(pos3d);
    free(pos2d);
    return err;
}


AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::verify_markers(ARMarkerInfo *marker_info, int marker_num, ARMultiMarkerInfoT *config)
{
    arMultiEachMarkerInternalInfoT *winfo;
    ARFloat                         wtrans[3][4];
    ARFloat                         pos3d[4][2];
    ARFloat                         wx, wy, wz, hx, hy, h;
    int                            dir1, dir2 = 0, marker2;
    ARFloat                         err, err1 = 0, err2;
    ARFloat                         x1 = 0, x2 = 0, y1 = 0, y2 = 0;
    int                            w1, w2;
    int                            i, j, k;

    arMalloc(winfo,arMultiEachMarkerInternalInfoT,config->marker_num);

    for( i = 0; i < config->marker_num; i++ ) {
		arUtilMatMul(config->trans, config->marker[i].trans, wtrans);
        pos3d[0][0] = config->marker[i].center[0] - config->marker[i].width*0.5f;
        pos3d[0][1] = config->marker[i].center[1] + config->marker[i].width*0.5f;
        pos3d[1][0] = config->marker[i].center[0] + config->marker[i].width*0.5f;
        pos3d[1][1] = config->marker[i].center[1] + config->marker[i].width*0.5f;
        pos3d[2][0] = config->marker[i].center[0] + config->marker[i].width*0.5f;
        pos3d[2][1] = config->marker[i].center[1] - config->marker[i].width*0.5f;
        pos3d[3][0] = config->marker[i].center[0] - config->marker[i].width*0.5f;
        pos3d[3][1] = config->marker[i].center[1] - config->marker[i].width*0.5f;
        for( j = 0; j < 4; j++ ) {
            wx = wtrans[0][0] * pos3d[j][0]
               + wtrans[0][1] * pos3d[j][1]
               + wtrans[0][3];
            wy = wtrans[1][0] * pos3d[j][0]
               + wtrans[1][1] * pos3d[j][1]
               + wtrans[1][3];
            wz = wtrans[2][0] * pos3d[j][0]
               + wtrans[2][1] * pos3d[j][1]
               + wtrans[2][3];
            hx = arCamera->mat[0][0] * wx
               + arCamera->mat[0][1] * wy
               + arCamera->mat[0][2] * wz
               + arCamera->mat[0][3];
            hy = arCamera->mat[1][0] * wx
               + arCamera->mat[1][1] * wy
               + arCamera->mat[1][2] * wz
               + arCamera->mat[1][3];
            h  = arCamera->mat[2][0] * wx
               + arCamera->mat[2][1] * wy
               + arCamera->mat[2][2] * wz
               + arCamera->mat[2][3];
            winfo[i].pos[j][0] = hx / h;
            winfo[i].pos[j][1] = hy / h;

            if(j ==0) {x1=x2=winfo[i].pos[j][0]; y1=y2=winfo[i].pos[j][1];}
            else {
                if( winfo[i].pos[j][0] < x1 ) x1 = winfo[i].pos[j][0];
                if( winfo[i].pos[j][0] > x2 ) x2 = winfo[i].pos[j][0];
                if( winfo[i].pos[j][1] < y1 ) y1 = winfo[i].pos[j][1];
                if( winfo[i].pos[j][1] > y2 ) y2 = winfo[i].pos[j][1];
            }
        }
        winfo[i].thresh = (x2 - x1 + 1)*(y2 - y1 + 1) / 2;
    }

#ifdef ARTK_DEBUG
printf("\n");
printf("================================================================\n");
for( i = 0; i < config->marker_num; i++) {
printf("%3d: ", i+1);
for( j = 0; j < 4; j++ ) {
    printf("(%5.1f %5.1f) ", winfo[i].pos[j][0], winfo[i].pos[j][1]);
}
printf("\n");
}
printf("--------\n");
for( i = 0; i < marker_num; i++) {
printf("%3d: ", i+1);
for( j = 0; j < 4; j++ ) {
    printf("(%5.1f %5.1f) ", marker_info[i].vertex[j][0], marker_info[i].vertex[j][1]);
}
printf("\n");
}
#endif

    w1 = w2 = 0;
    for( i = 0; i < config->marker_num; i++ ) {
        marker2 = -1;
        err2 = winfo[i].thresh;
        for( j = 0; j < marker_num; j++ ) {
            if( marker_info[j].id != -1
             && marker_info[j].id != config->marker[i].patt_id
             && marker_info[j].cf > 0.7 ) continue;

            dir1 = -1;
            for( k = 0; k < 4; k++ ) {
                err = (winfo[i].pos[0][0] - marker_info[j].vertex[(k+0)%4][0])
                    * (winfo[i].pos[0][0] - marker_info[j].vertex[(k+0)%4][0])
                    + (winfo[i].pos[0][1] - marker_info[j].vertex[(k+0)%4][1])
                    * (winfo[i].pos[0][1] - marker_info[j].vertex[(k+0)%4][1])
                    + (winfo[i].pos[1][0] - marker_info[j].vertex[(k+1)%4][0])
                    * (winfo[i].pos[1][0] - marker_info[j].vertex[(k+1)%4][0])
                    + (winfo[i].pos[1][1] - marker_info[j].vertex[(k+1)%4][1])
                    * (winfo[i].pos[1][1] - marker_info[j].vertex[(k+1)%4][1])
                    + (winfo[i].pos[2][0] - marker_info[j].vertex[(k+2)%4][0])
                    * (winfo[i].pos[2][0] - marker_info[j].vertex[(k+2)%4][0])
                    + (winfo[i].pos[2][1] - marker_info[j].vertex[(k+2)%4][1])
                    * (winfo[i].pos[2][1] - marker_info[j].vertex[(k+2)%4][1])
                    + (winfo[i].pos[3][0] - marker_info[j].vertex[(k+3)%4][0])
                    * (winfo[i].pos[3][0] - marker_info[j].vertex[(k+3)%4][0])
                    + (winfo[i].pos[3][1] - marker_info[j].vertex[(k+3)%4][1])
                    * (winfo[i].pos[3][1] - marker_info[j].vertex[(k+3)%4][1]);
                if( dir1 == -1 || err < err1 ) {
                    err1 = err;
                    dir1 = k;
                }
            }
#ifdef ARTK_DEBUG
	printf("%f\n", err1);
#endif
            if( err1 < err2 ) {
                err2 = err1;
                dir2 = dir1;
                marker2 = j;
            }
        }

#ifdef ARTK_DEBUG
	printf("%3d<=>%3d, err = %f(%f)\n", i+1, marker2+1, err2, winfo[i].thresh);
#endif
        if( marker2 != -1 ) {
            winfo[i].marker = marker2;
            winfo[i].dir    = dir2;
            winfo[i].err    = err2;

            if( marker_info[marker2].id == config->marker[i].patt_id ) w1++;
            else if( marker_info[marker2].id != -1 )                   w2++;
        }
        else {
            winfo[i].marker = -1;
        }
    }
#ifdef ARTK_DEBUG
	printf("w1,w2 = %d,%d\n", w1, w2);
#endif
    if( w2 >= w1 ) {
        free(winfo);
        return -1;
    }

    for( i = 0; i < config->marker_num; i++ ) {
        for( j = 0; j < marker_num; j++ ) {
                if( marker_info[j].id == config->marker[i].patt_id ) marker_info[j].id = -1;
        }
        if( winfo[i].marker != -1 ) {
            marker_info[winfo[i].marker].id  = config->marker[i].patt_id;
            marker_info[winfo[i].marker].dir = (4-winfo[i].dir)%4;
            marker_info[winfo[i].marker].cf  = 1.0;
        }
    }

    free(winfo);

    return 0;
}


}  // namespace ARToolKitPlus
// End of ../../src/core/arMultiGetTransMat.cxx
// Included ../../src/core/rppMultiGetTransMat.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Thomas Pintaric
 *
 * $Id: rppMultiGetTransMat.cxx 162 2006-04-19 21:28:10Z grabner $
 * @file
 * ======================================================================== */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <map>
#include <deque>
#include <ARToolKitPlus/Tracker.h>
#include <ARToolKitPlus/matrix.h>
#include <ARToolKitPlus/extra/rpp.h>


namespace ARToolKitPlus {

AR_TEMPL_FUNC ARFloat
AR_TEMPL_TRACKER::rppMultiGetTransMat(ARMarkerInfo *marker_info, int marker_num, ARMultiMarkerInfoT *config)
{
	rpp_float err = 1e+20;
	rpp_mat R, R_init;
	rpp_vec t;

	std::map<int, int> marker_id_freq;
	for(int i=0; i<marker_num; i++)
	{
		const int m_patt_id = marker_info[i].id;
		if(m_patt_id >= 0)
		{
			std::map<int, int>::iterator iter = marker_id_freq.find(m_patt_id);
			if(iter == marker_id_freq.end()) marker_id_freq.insert(std::make_pair(m_patt_id,1));
			else ((*iter).second)++;
		}
	}

	std::deque<std::pair<int,int> > config_patt_id;
	for(int j=0; j<config->marker_num; j++)
		config_patt_id.push_back(std::make_pair(j, config->marker[j].patt_id));

	std::map<int, int> m2c_idx;
	for(int m=0; m<marker_num; m++)
	{
		const int m_patt_id = marker_info[m].id;
		bool ignore_marker = (m_patt_id < 0);
		std::map<int, int>::iterator m_iter = marker_id_freq.find(m_patt_id);
		if(m_iter != marker_id_freq.end()) ignore_marker |= ((*m_iter).second > 1);
		if(!ignore_marker)
		{
			std::deque<std::pair<int,int> >::iterator c_iter = config_patt_id.begin();
			if(c_iter != config_patt_id.end()) do
			{
				const int patt_id = (*c_iter).second;
				if(marker_info[m].id == patt_id)
				{
					m2c_idx.insert(std::make_pair(m,(*c_iter).first));
					config_patt_id.erase(c_iter);
					c_iter = config_patt_id.end();
					continue;
				}
				else
				{
					c_iter++;
				}
			}
			while(c_iter != config_patt_id.end());
		}
	}

	// ----------------------------------------------------------------------
	const unsigned int n_markers = (unsigned int) m2c_idx.size();
	const unsigned int n_pts = 4*n_markers;

	if(n_markers == 0) return(-1);

	rpp_vec *ppos2d = NULL, *ppos3d = NULL;
	arMalloc( ppos2d, rpp_vec, n_pts);
	arMalloc( ppos3d, rpp_vec, n_pts);
	memset(ppos2d,0,sizeof(rpp_vec)*n_pts);
	memset(ppos3d,0,sizeof(rpp_vec)*n_pts);

	const rpp_float iprts_z =  1;

	int p=0;
	for(std::map<int, int>::iterator iter = m2c_idx.begin();
		iter != m2c_idx.end(); iter++)
	{
		const int m = (*iter).first;
		const int c = (*iter).second;

		const int dir = marker_info[m].dir;
		const int v_idx[4] = {(4-dir)%4, (5-dir)%4, (6-dir)%4, (7-dir)%4};

		for(int i=0; i<4; i++)
			for(int j=0; j<3; j++)
			{
				ppos2d[p+i][j] = (rpp_float) (j==2 ? iprts_z : marker_info[m].vertex[v_idx[i]][j]);
				ppos3d[p+i][j] = (rpp_float) config->marker[c].pos3d[i][j];
			}

		p += 4;
	}

	const rpp_float cc[2] = {arCamera->mat[0][2],arCamera->mat[1][2]};
	const rpp_float fc[2] = {arCamera->mat[0][0],arCamera->mat[1][1]};

	robustPlanarPose(err,R,t,cc,fc,ppos3d,ppos2d,n_pts,R_init,true,0,0,0);

	for(int k=0; k<3; k++)
	{
		config->trans[k][3] = (ARFloat)t[k];
		for(int j=0; j<3; j++)
			config->trans[k][j] = (ARFloat)R[k][j];
	}

	if(ppos2d != NULL) free(ppos2d);
	if(ppos3d != NULL) free(ppos3d);

	if(err > 1e+10) return(-1); // an actual error has occurred in robustPlanarPose()
	return(ARFloat(err)); // NOTE: err is a real number from the interval [0,1e+10]
}

}  // namespace ARToolKitPlus
// End of ../../src/core/rppMultiGetTransMat.cxx
// Included ../../src/core/arMultiReadConfigFile.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id: arMultiReadConfigFile.cxx 162 2006-04-19 21:28:10Z grabner $
 * @file
 * ======================================================================== */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ARToolKitPlus/Tracker.h>
#include <ARToolKitPlus/matrix.h>


namespace ARToolKitPlus {


static char*
get_buff( char *buf, int n, FILE *fp )
{
    char *ret;

    for(;;) {
        ret = fgets( buf, n, fp );
        if(ret==NULL)
			return(NULL);
        if(buf[0] != '\n' && buf[0] != '#')
			return(ret);
    }
}


static bool
isNumber(const char* nString)
{
	int i,len = (int)strlen(nString);

	for(i=0; i<len; i++)
		if(nString[i]<'0' || nString[i]>'9')
			return false;

	return true;
}


AR_TEMPL_FUNC ARMultiMarkerInfoT*
AR_TEMPL_TRACKER::arMultiReadConfigFile(const char *filename)
{
    FILE                   *fp;
    ARMultiEachMarkerInfoT *marker;
    ARMultiMarkerInfoT     *marker_info;
    ARFloat                 wpos3d[4][2];
    char                   buf[256], buf1[256];
    int                    num;
    int                    i, j;

    if( (fp=fopen(filename,"r")) == NULL ) return NULL;

    get_buff(buf, 256, fp);
    if( sscanf(buf, "%d", &num) != 1 ) {fclose(fp); return NULL;}

    arMalloc(marker,ARMultiEachMarkerInfoT,num);

    for( i = 0; i < num; i++ ) {
        get_buff(buf, 256, fp);
        if( sscanf(buf, "%s", buf1) != 1 ) {
            fclose(fp); free(marker); return NULL;
        }
		// Added by Daniel: if the markername is an integer number
		// we directly interprete this as the marker id (used for
		// id-based markers)
		if(isNumber(buf1))
			marker[i].patt_id = atoi(buf1);
		else
        if( (marker[i].patt_id = arLoadPatt(buf1)) < 0 ) {
            fclose(fp); free(marker); return NULL;
        }

        get_buff(buf, 256, fp);
#ifdef _USE_DOUBLE_
        if( sscanf(buf, "%lf", &marker[i].width) != 1 ) {
#else
        if( sscanf(buf, "%f", &marker[i].width) != 1 ) {
#endif
            fclose(fp); free(marker); return NULL;
        }

        get_buff(buf, 256, fp);
#ifdef _USE_DOUBLE_
		if( sscanf(buf, "%lf %lf", &marker[i].center[0], &marker[i].center[1]) != 2 ) {
#else
		if( sscanf(buf, "%f %f", &marker[i].center[0], &marker[i].center[1]) != 2 ) {
#endif
            fclose(fp); free(marker); return NULL;
        }

        for( j = 0; j < 3; j++ ) {
            get_buff(buf, 256, fp);
            if( sscanf(buf,
#ifdef _USE_DOUBLE_
                       "%lf %lf %lf %lf",
#else
                       "%f %f %f %f",
#endif
                       &marker[i].trans[j][0], &marker[i].trans[j][1], &marker[i].trans[j][2], &marker[i].trans[j][3]) != 4 )
			{
                fclose(fp); free(marker); return NULL;
            }
        }
        arUtilMatInv( marker[i].trans, marker[i].itrans );

        wpos3d[0][0] = marker[i].center[0] - marker[i].width*0.5f;
        wpos3d[0][1] = marker[i].center[1] + marker[i].width*0.5f;
        wpos3d[1][0] = marker[i].center[0] + marker[i].width*0.5f;
        wpos3d[1][1] = marker[i].center[1] + marker[i].width*0.5f;
        wpos3d[2][0] = marker[i].center[0] + marker[i].width*0.5f;
        wpos3d[2][1] = marker[i].center[1] - marker[i].width*0.5f;
        wpos3d[3][0] = marker[i].center[0] - marker[i].width*0.5f;
        wpos3d[3][1] = marker[i].center[1] - marker[i].width*0.5f;
        for( j = 0; j < 4; j++ ) {
            marker[i].pos3d[j][0] = marker[i].trans[0][0] * wpos3d[j][0]
                                  + marker[i].trans[0][1] * wpos3d[j][1]
                                  + marker[i].trans[0][3];
            marker[i].pos3d[j][1] = marker[i].trans[1][0] * wpos3d[j][0]
                                  + marker[i].trans[1][1] * wpos3d[j][1]
                                  + marker[i].trans[1][3];
            marker[i].pos3d[j][2] = marker[i].trans[2][0] * wpos3d[j][0]
                                  + marker[i].trans[2][1] * wpos3d[j][1]
                                  + marker[i].trans[2][3];
        }
    }

    fclose(fp);

    marker_info = (ARMultiMarkerInfoT *)malloc( sizeof(ARMultiMarkerInfoT) );
    if( marker_info == NULL ) {free(marker); return NULL;}
    marker_info->marker     = marker;
    marker_info->marker_num = num;
    marker_info->prevF      = 0;

    return marker_info;
}


}  // namespace ARToolKitPlus
// End of ../../src/core/arMultiReadConfigFile.cxx
// Included ../../src/core/arUtil.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id: arUtil.cxx 162 2006-04-19 21:28:10Z grabner $
 * @file
 * ======================================================================== */


#include <stdio.h>
#include <math.h>

#ifdef _WIN32

#ifndef _WIN32_WCE
#include <sys/timeb.h>
#endif

#include <windows.h>
#else
#include <sys/time.h>
#endif

#include <ARToolKitPlus/Tracker.h>
#include <ARToolKitPlus/param.h>
#include <ARToolKitPlus/matrix.h>


namespace ARToolKitPlus {


AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arInitCparam(Camera *pCam)
{
	// if the camera parameters change, the undistortion LUT has to be rebuilt.
	// (this is done automatically in arParamObserv2Ideal_LUT or arParamIdeal2Observ_LUT)
	//
	if(undistO2ITable && (arImXsize!=pCam->xsize || arImYsize!=pCam->ysize))
	{
		artkp_Free(undistO2ITable);
		undistO2ITable = NULL;
	}

	arImXsize = pCam->xsize;
	arImYsize = pCam->ysize;

    return(0);
}


AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arGetLine(int x_coord[], int y_coord[], int coord_num, int vertex[], ARFloat line[4][3], ARFloat v[4][2])
{
    //return arGetLine2( x_coord, y_coord, coord_num, vertex, line, v, arParam.dist_factor );
	return arGetLine2( x_coord, y_coord, coord_num, vertex, line, v, arCamera );
}


AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arGetLine2(int x_coord[], int y_coord[], int coord_num,
                    int vertex[], ARFloat line[4][3], ARFloat v[4][2], Camera *pCam) 
{
    ARMat    *input, *evec;
    ARVec    *ev, *mean;
    ARFloat   w1;
    int      st, ed, n;
    int      i, j;

    ev     = Vector::alloc( 2 );
    mean   = Vector::alloc( 2 );
    evec   = Matrix::alloc( 2, 2 );
    for( i = 0; i < 4; i++ ) {
        w1 = (ARFloat)(vertex[i+1]-vertex[i]+1) * (ARFloat)0.05 + (ARFloat)0.5;
        st = (int)(vertex[i]   + w1);
        ed = (int)(vertex[i+1] - w1);
        n = ed - st + 1;
        input  = Matrix::alloc( n, 2 );
        for( j = 0; j < n; j++ ) {
			// does not work
            (this->*arParamObserv2Ideal_func)( pCam, (ARFloat)x_coord[st+j], (ARFloat)y_coord[st+j], &(input->m[j*2+0]), &(input->m[j*2+1]) );
			//
        }
        if( arMatrixPCA(input, evec, ev, mean) < 0 ) {
            Matrix::free( input );
            Matrix::free( evec );
            Vector::free( mean );
            Vector::free( ev );
            return(-1);
        }
        line[i][0] =  evec->m[1];
        line[i][1] = -evec->m[0];
        line[i][2] = -(line[i][0]*mean->v[0] + line[i][1]*mean->v[1]);
        Matrix::free( input );
    }
    Matrix::free( evec );
    Vector::free( mean );
    Vector::free( ev );

    for( i = 0; i < 4; i++ ) {
        w1 = line[(i+3)%4][0] * line[i][1] - line[i][0] * line[(i+3)%4][1];
        if( w1 == 0.0 ) return(-1);
        v[i][0] = (  line[(i+3)%4][1] * line[i][2]
                   - line[i][1] * line[(i+3)%4][2] ) / w1;
        v[i][1] = (  line[i][0] * line[(i+3)%4][2]
                   - line[(i+3)%4][0] * line[i][2] ) / w1;
    }

    return(0);
}

AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arUtilMatMul( ARFloat s1[3][4], ARFloat s2[3][4], ARFloat d[3][4] )
{
    int     i, j;

    for( j = 0; j < 3; j++ ) {
        for( i = 0; i < 4; i++) {
            d[j][i] = s1[j][0] * s2[0][i]
                    + s1[j][1] * s2[1][i]
                    + s1[j][2] * s2[2][i];
        }
        d[j][3] += s1[j][3];
    }

    return 0;
}

AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arUtilMatInv(ARFloat s[3][4], ARFloat d[3][4])
{
    ARMat       *mat;
    int         i, j;

    mat = Matrix::alloc( 4, 4 );
    for( j = 0; j < 3; j++ ) {
        for( i = 0; i < 4; i++ ) {
            mat->m[j*4+i] = s[j][i];
        }
    }
    mat->m[3*4+0] = 0; mat->m[3*4+1] = 0;
    mat->m[3*4+2] = 0; mat->m[3*4+3] = 1;
    Matrix::selfInv( mat );
    for( j = 0; j < 3; j++ ) {
        for( i = 0; i < 4; i++ ) {
            d[j][i] = mat->m[j*4+i];
        }
    }
    Matrix::free( mat );

    return 0;
}


}  // namespace ARToolKitPlus
// End of ../../src/core/arUtil.cxx
// Included ../../src/core/matrix.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id: matrix.cxx 162 2006-04-19 21:28:10Z grabner $
 * @file
 * ======================================================================== */


#include <stdio.h>
#include <math.h>
#ifndef __APPLE__
#include <malloc.h>
#else
#include <stdlib.h>
#endif
#include <ARToolKitPlus/matrix.h>


namespace ARToolKitPlus {


namespace Matrix {


#define MATRIX(name,x,y,width)  ( *(name + (width) * (x) + (y)) )


// declaration of internal helper functions (see end of file)
// static ARFloat mdet(ARFloat *ap, int dimen, int rowa);
static ARFloat *minv( ARFloat *ap, int dimen, int rowa );


// from mAlloc.c
static ARMat*
alloc(int row, int clm)
{
	ARMat *m;

	m = (ARMat *)malloc(sizeof(ARMat));
	if( m == NULL ) return NULL;

	m->m = (ARFloat *)malloc(sizeof(ARFloat) * row * clm);
	if(m->m == NULL) {
		free(m);
		return NULL;
	}
	else {
		m->row = row;
		m->clm = clm;
	}

	return m;
}


// from mFree.c
static int
free(ARMat *m)
{
	::free(m->m);
	::free(m);

	return 0;
}


// from mAllocDup.c
static ARMat*
allocDup(ARMat *source)
{
	ARMat *dest;

	dest = alloc(source->row, source->clm);
        if( dest == NULL ) return NULL;

	if( dup(dest, source) < 0 ) {
		free(dest);
		return NULL;
	}

	return dest;
}


// from mAllocInv.c
/*
static ARMat*
allocInv(ARMat *source)
{
	ARMat *dest;

	dest = alloc(source->row, source->row);
	if( dest == NULL ) return NULL;

	if( inv(dest, source) < 0 ) {
		free( dest );
		return NULL;
	}

	return dest;
}
*/

// from mAllocMul.c
/*
static ARMat*
allocMul(ARMat *a, ARMat *b)
{
	ARMat *dest;

	dest = alloc(a->row, b->clm);
	if( dest == NULL ) return NULL;

	if( mul(dest, a, b) < 0 ) {
		free(dest);
		return NULL;
	}

	return dest;
}
*/

// from mAllocTrans.c
/*
static ARMat*
allocTrans(ARMat *source)
{
	ARMat *dest;

	dest = alloc(source->clm, source->row);
	if( dest == NULL ) return NULL;

	if( trans(dest, source) < 0 ) {
		free(dest);
		return NULL;
	}

	return dest;
}
*/

// from mAllocUnit.c
/*
static ARMat*
allocUnit(int dim)
{
	ARMat *m;

	m = alloc(dim, dim);
	if( m == NULL ) return NULL;

	if( unit(m) < 0 ) {
		free(m);
		return NULL;
	}

	return m;
}
*/

// from mDet.c
/*
static ARFloat
det(ARMat *m)
{

	if(m->row != m->clm) return 0.0;

	return mdet(m->m, m->row, m->row);
}
*/

// from mDisp.c
/*
static int
disp(ARMat *m)
{
	int r, c;

	printf(" === matrix (%d,%d) ===\n", m->row, m->clm);
	for(r = 0; r < m->row; r++) {
		printf(" |");
		for(c = 0; c < m->clm; c++) {
			printf(" %10g", ARELEM0(m, r, c));
		}
		printf(" |\n");
	}
	printf(" ======================\n");

	return 0;
}
*/

// from mDup.c
static int
dup(ARMat *dest, ARMat *source)
{
	int r,c;

	if(dest->row != source->row || dest->clm != source->clm) {
		return -1;
	}
	for(r = 0; r < source->row; r++) {
		for(c = 0; c < source->clm; c++) {
			ARELEM0(dest, r, c) = ARELEM0(source, r, c);
		}
	}
	return 0;
}


// from mInv.c
/*
static int
inv(ARMat *dest, ARMat *source)
{
	if(dup(dest, source) < 0) return -1;

	return selfInv(dest);
}
*/

// from mMul.c
static int
mul(ARMat *dest, ARMat *a, ARMat *b)
{
	int r, c, i;

	if(a->clm != b->row || dest->row != a->row || dest->clm != b->clm) return -1;

	for(r = 0; r < dest->row; r++) {
		for(c = 0; c < dest->clm; c++) {
			ARELEM0(dest, r, c) = 0.0;
			for(i = 0; i < a->clm; i++) {
				ARELEM0(dest, r, c) += ARELEM0(a, r, i) * ARELEM0(b, i, c);
			}
		}
	}

	return 0;
}


// from mSelfInv.c
static int
selfInv(ARMat *m)
{
	if(minv(m->m, m->row, m->row) == NULL) return -1;

	return 0;
}


// from mTrans.c
/*
static int
trans(ARMat *dest, ARMat *source)
{
	int r, c;

	if(dest->row != source->clm || dest->clm != source->row) return -1;

	for(r = 0; r < dest->row; r++) {
		for(c = 0; c < dest->clm; c++) {
			ARELEM0(dest, r, c) = ARELEM0(source, c, r);
		}
	}

	return 0;
}
*/

// from mUnit.c
/*
static int
unit(ARMat *unit)
{
	int r, c;

	if(unit->row != unit->clm) return -1;

	for(r = 0; r < unit->row; r++) {
		for(c = 0; c < unit->clm; c++) {
			if(r == c) {
				ARELEM0(unit, r, c) = 1.0;
			}
			else {
				ARELEM0(unit, r, c) = 0.0;
			}
		}
	}

	return 0;
}
*/

/****************************************************************
 *
 *                  INTERNAL HELPER FUNCTIONS
 *
 ***************************************************************/

// from mDet.c  -- helper function for arMatrixDet()
#if 0
static ARFloat
mdet(ARFloat *ap, int dimen, int rowa)
/*  ARFloat  *ap;          input matrix */
/*  int     dimen;        Dimension of linre and row, those must be equal,
                          that is square matrix.       */
/*  int     rowa;         ROW Dimension of matrix A    */
{
    ARFloat det = 1.0;
    ARFloat work;
    int    is = 0;
    int    mmax;
    int    i, j, k;

    for(k = 0; k < dimen - 1; k++) {
        mmax = k;
        for(i = k + 1; i < dimen; i++)
            if (fabs(MATRIX(ap, i, k, rowa)) > fabs(MATRIX(ap, mmax, k, rowa)))
                mmax = i;
        if(mmax != k) {
            for (j = k; j < dimen; j++) {
                work = MATRIX(ap, k, j, rowa);
                MATRIX(ap, k, j, rowa) = MATRIX(ap, mmax, j, rowa);
                MATRIX(ap, mmax, j, rowa) = work;
            }
            is++;
        }
        for(i = k + 1; i < dimen; i++) {
            work = MATRIX(ap, i, k, rowa) / MATRIX(ap, k, k, rowa);
            for (j = k + 1; j < dimen; j++)
                MATRIX(ap, i, j, rowa) -= work * MATRIX(ap, k, j, rowa);
        }
    }
    for(i = 0; i < dimen; i++)
        det *= MATRIX(ap, i, i, rowa);
    for(i = 0; i < is; i++) 
        det *= -1.0;
    return(det);
}
#endif

// from mSelfInv.c -- MATRIX inverse function
static ARFloat*
minv( ARFloat *ap, int dimen, int rowa )
{
        ARFloat *wap, *wcp, *wbp;/* work pointer                 */
        int i,j,n,ip=0,nwork;
        int nos[50];
        ARFloat epsl;
        ARFloat p,pbuf,work;
        //ARFloat  fabs();

        epsl = (ARFloat)1.0e-10;         /* Threshold value      */

        switch (dimen) {
                case (0): return(NULL);                 /* check size */
                case (1): *ap = (ARFloat)1.0 / (*ap);
                          return(ap);                   /* 1 dimension */
        }

        for(n = 0; n < dimen ; n++)
                nos[n] = n;

        for(n = 0; n < dimen ; n++) {
                wcp = ap + n * rowa;

                for(i = n, wap = wcp, p = 0.0; i < dimen ; i++, wap += rowa)
                        if( p < ( pbuf = (ARFloat)fabs(*wap)) ) {
                                p = pbuf;
                                ip = i;
                        }
                if (p <= epsl)
                        return(NULL);

                nwork = nos[ip];
                nos[ip] = nos[n];
                nos[n] = nwork;

                for(j = 0, wap = ap + ip * rowa, wbp = wcp; j < dimen ; j++) {
                        work = *wap;
                        *wap++ = *wbp;
                        *wbp++ = work;
                }

                for(j = 1, wap = wcp, work = *wcp; j < dimen ; j++, wap++)
                        *wap = *(wap + 1) / work;
                *wap = (ARFloat)1.0 / work;

                for(i = 0; i < dimen ; i++) {
                        if(i != n) {
                                wap = ap + i * rowa;
                                for(j = 1, wbp = wcp, work = *wap;
                                                j < dimen ; j++, wap++, wbp++)
                                        *wap = *(wap + 1) - work * (*wbp);
                                *wap = -work * (*wbp);
                        }
                }
        }

        for(n = 0; n < dimen ; n++) {
                for(j = n; j < dimen ; j++)
                        if( nos[j] == n) break;
                nos[j] = nos[n];
                for(i = 0, wap = ap + j, wbp = ap + n; i < dimen ;
                                        i++, wap += rowa, wbp += rowa) {
                        work = *wap;
                        *wap = *wbp;
                        *wbp = work;
                }
        }
        return(ap);
}


}  // namespace Matrix


}  // namespace ARToolKitPlus
// End of ../../src/core/matrix.cxx
// Included ../../src/core/mPCA.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id: mPCA.cxx 162 2006-04-19 21:28:10Z grabner $
 * @file
 * ======================================================================== */


#include <stdio.h>
#include <math.h>
#include <ARToolKitPlus/Tracker.h>


namespace ARToolKitPlus {


#define     VZERO           1e-16
#define     EPS             1e-6
#define     MAX_ITER        100
#define     xmalloc(V,T,S)  { if( ((V) = (T *)malloc( sizeof(T) * (S) ))\
                               == NULL ) {printf("malloc error\n"); exit(1);} }

static int EX( ARMat *input, ARVec *mean );
static int CENTER( ARMat *inout, ARVec *mean );
static int PCA( ARMat *input, ARMat *output, ARVec *ev );
static int x_by_xt( ARMat *input, ARMat *output );
static int xt_by_x( ARMat *input, ARMat *output );
static int EV_create( ARMat *input, ARMat *u, ARMat *output, ARVec *ev );
static int QRM( ARMat *u, ARVec *ev );


/* === matrix definition ===

Input:
  <---- clm (Data dimention)--->
  [ 10  20  30 ] ^
  [ 20  10  15 ] |
  [ 12  23  13 ] row
  [ 20  10  15 ] |(Sample number)
  [ 13  14  15 ] v

Evec:
  <---- clm (Eigen vector)--->
  [ 10  20  30 ] ^
  [ 20  10  15 ] |
  [ 12  23  13 ] row
  [ 20  10  15 ] |(Number of egen vec)
  [ 13  14  15 ] v

Ev:
  <---- clm (Number of eigen vector)--->
  [ 10  20  30 ] eigen value

Mean:
  <---- clm (Data dimention)--->
  [ 10  20  30 ] mean value

=========================== */


AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arMatrixPCA(ARMat *input, ARMat *evec, ARVec *ev, ARVec *mean)
{
    ARMat     *work;
    ARFloat  srow, sum;
    int     row, clm;
    int     check, rval;
    int     i;

    row = input->row;
    clm = input->clm;
    check = (row < clm)? row: clm;
    if( row < 2 || clm < 2 ) return(-1);
    if( evec->clm != input->clm || evec->row != check ) return(-1);
    if( ev->clm   != check )      return(-1);
    if( mean->clm != input->clm ) return(-1);

    work = Matrix::allocDup( input );
    if( work == NULL ) return -1;

    srow = (ARFloat)sqrt((ARFloat)row);
    if( EX( work, mean ) < 0 ) {
        Matrix::free( work );
        return(-1);
    }
    if( CENTER( work, mean ) < 0 ) {
        Matrix::free( work );
        return(-1);
    }
    for(i=0; i<row*clm; i++) work->m[i] /= srow;

    rval = PCA( work, evec, ev );
    Matrix::free( work );

    sum = 0.0;
    for( i = 0; i < ev->clm; i++ ) sum += ev->v[i];
    for( i = 0; i < ev->clm; i++ ) ev->v[i] /= sum;

    return( rval );
}

AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arMatrixPCA2( ARMat *input, ARMat *evec, ARVec *ev )
{
    ARMat   *work;
    // ARFloat  srow; // unreferenced
    ARFloat  sum;
    int     row, clm;
    int     check, rval;
    int     i;

    row = input->row;
    clm = input->clm;
    check = (row < clm)? row: clm;
    if( row < 2 || clm < 2 ) return(-1);
    if( evec->clm != input->clm || evec->row != check ) return(-1);
    if( ev->clm   != check )      return(-1);

    work = Matrix::allocDup( input );
    if( work == NULL ) return -1;

/*
    srow = sqrt((ARFloat)row);
    for(i=0; i<row*clm; i++) work->m[i] /= srow;
*/

    rval = PCA( work, evec, ev );
    Matrix::free( work );

    sum = 0.0;
    for( i = 0; i < ev->clm; i++ ) sum += ev->v[i];
    for( i = 0; i < ev->clm; i++ ) ev->v[i] /= sum;

    return( rval );
}

static int
PCA( ARMat *input, ARMat *output, ARVec *ev )
{
    ARMat     *u;
    ARFloat  *m1, *m2;
    int     row, clm, min;
    int     i, j;

    row = input->row;
    clm = input->clm;
    min = (clm < row)? clm: row;
    if( row < 2 || clm < 2 )        return(-1);
    if( output->clm != input->clm ) return(-1);
    if( output->row != min )        return(-1);
    if( ev->clm != min )            return(-1);

    u = Matrix::alloc( min, min );
    if( u->row != min || u->clm != min ) return(-1);
    if( row < clm ) {
        if( x_by_xt( input, u ) < 0 ) { Matrix::free(u); return(-1); }
    }
    else {
        if( xt_by_x( input, u ) < 0 ) { Matrix::free(u); return(-1); }
    }

    if( QRM( u, ev ) < 0 ) { Matrix::free(u); return(-1); }

    if( row < clm ) {
        if( EV_create( input, u, output, ev ) < 0 ) {
            Matrix::free(u);
            return(-1);
        }
    }
    else{
        m1 = u->m;
        m2 = output->m;
	for( i = 0; i < min; i++){
	    if( ev->v[i] < VZERO ) break;
            for( j = 0; j < min; j++ ) *(m2++) = *(m1++);
        }
	for( ; i < min; i++){
            ev->v[i] = 0.0;
            for( j = 0; j < min; j++ ) *(m2++) = 0.0;
        }
    }

    Matrix::free(u);

    return( 0 );
}

static int
EX( ARMat *input, ARVec *mean )
{
    ARFloat  *m, *v;
    int     row, clm;
    int     i, j;

    row = input->row;
    clm = input->clm;
    if( row <= 0 || clm <= 0 )  return(-1);
    if( mean->clm != clm )      return(-1);

    for( i = 0; i < clm; i++ ) mean->v[i] = 0.0;

    m = input->m;
    for( i = 0; i < row; i++ ) {
        v = mean->v;
        for( j = 0; j < clm; j++ ) *(v++) += *(m++);
    }

    for( i = 0; i < clm; i++ ) mean->v[i] /= row;

    return(0);
}

static int
CENTER( ARMat *inout, ARVec *mean )
{
    ARFloat  *m, *v;
    int     row, clm;
    int     i, j;

    row = inout->row;
    clm = inout->clm;
    if( mean->clm != clm ) return(-1);

    m = inout->m;
    for( i = 0; i < row; i++ ) {
        v = mean->v;
        for( j = 0; j < clm; j++ ) *(m++) -= *(v++);
    }

    return(0);
}

static int
x_by_xt( ARMat *input, ARMat *output )
{
    ARFloat  *in1, *in2, *out;
    int     row, clm;
    int     i, j, k;

    row = input->row;
    clm = input->clm;
    if( output->row != row || output->clm != row ) return(-1);

    out = output->m;
    for( i = 0; i < row; i++ ) {
        for( j = 0; j < row; j++ ) {
            if( j < i ) {
                *out = output->m[j*row+i];
            }
            else {
                in1 = &(input->m[clm*i]);
                in2 = &(input->m[clm*j]);
                *out = 0.0;
                for( k = 0; k < clm; k++ ) {
                    *out += *(in1++) * *(in2++);
                }
            }
            out++;
        }
    }

    return(0);
}

static int
xt_by_x( ARMat *input, ARMat *output )
{
    ARFloat  *in1, *in2, *out;
    int     row, clm;
    int     i, j, k;

    row = input->row;
    clm = input->clm;
    if( output->row != clm || output->clm != clm ) return(-1);

    out = output->m;
    for( i = 0; i < clm; i++ ) {
        for( j = 0; j < clm; j++ ) {
            if( j < i ) {
                *out = output->m[j*clm+i];
            }
            else {
                in1 = &(input->m[i]);
                in2 = &(input->m[j]);
                *out = 0.0;
                for( k = 0; k < row; k++ ) {
                    *out += *in1 * *in2;
                    in1 += clm;
                    in2 += clm;
                }
            }
            out++;
        }
    }

    return(0);
}

static int
EV_create( ARMat *input, ARMat *u, ARMat *output, ARVec *ev )
{
    ARFloat  *m, *m1, *m2;
    ARFloat  sum, work;
    int     row, clm;
    int     i, j, k;

    row = input->row;
    clm = input->clm;
    if( row <= 0 || clm <= 0 ) return(-1);
    if( u->row != row || u->clm != row ) return(-1);
    if( output->row != row || output->clm != clm ) return(-1);
    if( ev->clm != row ) return(-1);

    m = output->m;
    for( i = 0; i < row; i++ ) {
        if( ev->v[i] < VZERO ) break;
        work = (ARFloat)1 / (ARFloat)sqrt(fabs(ev->v[i]));
        for( j = 0; j < clm; j++ ) {
            sum = 0.0;
            m1 = &(u->m[i*row]);
            m2 = &(input->m[j]);
            for( k = 0; k < row; k++ ) {
                sum += *m1 * *m2;
                m1++;
                m2 += clm;
            }
            *(m++) = sum * work;
        }
    }
    for( ; i < row; i++ ) {
        ev->v[i] = 0.0;
        for( j = 0; j < clm; j++ ) *(m++) = 0.0;
    }

    return(0);
}

static int
QRM( ARMat *a, ARVec *dv )
{
    ARVec     *ev, ev1;
    ARFloat  w, t, s, x, y, c;
    ARFloat  *v1, *v2;
    int     dim, iter;
    int     i, j, k, h;

    dim = a->row;
    if( dim != a->clm || dim < 2 ) return(-1);
    if( dv->clm != dim ) return(-1);

    ev = Vector::alloc( dim );
    if( ev == NULL ) return(-1);

    ev1.clm = dim-1;
    ev1.v = &(ev->v[1]);
    if( Vector::tridiagonalize( a, dv, &ev1 ) < 0 ) {
        Vector::free( ev );
        return(-1);
    }

    ev->v[0] = 0.0;
    for( h = dim-1; h > 0; h-- ) {
        j = h;
        while(j>0 && fabs(ev->v[j]) > EPS*(fabs(dv->v[j-1])+fabs(dv->v[j]))) j--;
        if( j == h ) continue;

        iter = 0;
        do{
            iter++;
            if( iter > MAX_ITER ) break;

            w = (dv->v[h-1] - dv->v[h]) / 2;
            t = ev->v[h] * ev->v[h];
            s = (ARFloat)sqrt(w*w+t); 
            if( w < 0 ) s = -s;
            x = dv->v[j] - dv->v[h] + t/(w+s);
            y = ev->v[j+1];
            for( k = j; k < h; k++ ) {
                if( fabs(x) >= fabs(y) ) {
		    if( fabs(x) > VZERO ) {
			t = -y / x;
			c = (ARFloat)1 / (ARFloat)sqrt(t*t+1);
			s = t * c;
		    }
		    else{
			c = 1.0;
			s = 0.0;
		    }
                }
                else{
		    t = -x / y;
		    s = (ARFloat)1.0 / (ARFloat)sqrt(t*t+1);
		    c = t * s;
                }
                w = dv->v[k] - dv->v[k+1];
                t = (w * s + 2 * c * ev->v[k+1]) * s;
                dv->v[k]   -= t;
                dv->v[k+1] += t;
                if( k > j) ev->v[k] = c * ev->v[k] - s * y;
                ev->v[k+1] += s * (c * w - 2 * s * ev->v[k+1]);

                for( i = 0; i < dim; i++ ) {
                    x = a->m[k*dim+i];
                    y = a->m[(k+1)*dim+i];
                    a->m[k*dim+i]     = c * x - s * y;
                    a->m[(k+1)*dim+i] = s * x + c * y;
                }
                if( k < h-1 ) {
                    x = ev->v[k+1];
                    y = -s * ev->v[k+2];
                    ev->v[k+2] *= c;
                }
            }
        } while(fabs(ev->v[h]) > EPS*(fabs(dv->v[h-1])+fabs(dv->v[h])));
    }

    for( k = 0; k < dim-1; k++ ) {
        h = k;
        t = dv->v[h];
        for( i = k+1; i < dim; i++ ) {
            if( dv->v[i] > t ) {
                h = i;
                t = dv->v[h];
            }
        }
        dv->v[h] = dv->v[k];
        dv->v[k] = t;
        v1 = &(a->m[h*dim]);
        v2 = &(a->m[k*dim]);
        for( i = 0; i < dim; i++ ) {
            w = *v1;
            *v1 = *v2;
            *v2 = w;
            v1++;
            v2++;
        }
    }

    Vector::free( ev );
    return(0);
}


}  // namespace ARToolKitPlus
// End of ../../src/core/mPCA.cxx
//#include "../../src/core/paramChangeSize.cxx"
// Included ../../src/core/paramDecomp.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id: paramDecomp.cxx 162 2006-04-19 21:28:10Z grabner $
 * @file
 * ======================================================================== */


#include <stdio.h>
#include <math.h>
#include <ARToolKitPlus/Tracker.h>
#include <ARToolKitPlus/matrix.h>


namespace ARToolKitPlus {


static ARFloat norm( ARFloat a, ARFloat b, ARFloat c );
static ARFloat dot( ARFloat a1, ARFloat a2, ARFloat a3,
                   ARFloat b1, ARFloat b2, ARFloat b3 );

AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arParamDecomp( ARParam *source, ARParam *icpara, ARFloat trans[3][4] )
{
    icpara->xsize          = source->xsize;
    icpara->ysize          = source->ysize;
    icpara->dist_factor[0] = source->dist_factor[0];
    icpara->dist_factor[1] = source->dist_factor[1];
    icpara->dist_factor[2] = source->dist_factor[2];
    icpara->dist_factor[3] = source->dist_factor[3];

    return arParamDecompMat( source->mat, icpara->mat, trans );
}


AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arParamDecompMat( ARFloat source[3][4], ARFloat cpara[3][4], ARFloat trans[3][4] )
{
    int       r, c;
    ARFloat    Cpara[3][4];
    ARFloat    rem1, rem2, rem3;

    if( source[2][3] >= 0 ) {
        for( r = 0; r < 3; r++ ){
	    for( c = 0; c < 4; c++ ){
                Cpara[r][c] = source[r][c];
            }
        }
    }
    else {
        for( r = 0; r < 3; r++ ){
	    for( c = 0; c < 4; c++ ){
                Cpara[r][c] = -(source[r][c]);
            }
        }
    }

    for( r = 0; r < 3; r++ ){
	for( c = 0; c < 4; c++ ){
            cpara[r][c] = 0.0;
	}
    }
    cpara[2][2] = norm( Cpara[2][0], Cpara[2][1], Cpara[2][2] );
    trans[2][0] = Cpara[2][0] / cpara[2][2];
    trans[2][1] = Cpara[2][1] / cpara[2][2];
    trans[2][2] = Cpara[2][2] / cpara[2][2];
    trans[2][3] = Cpara[2][3] / cpara[2][2];
	
    cpara[1][2] = dot( trans[2][0], trans[2][1], trans[2][2],
                       Cpara[1][0], Cpara[1][1], Cpara[1][2] );
    rem1 = Cpara[1][0] - cpara[1][2] * trans[2][0];
    rem2 = Cpara[1][1] - cpara[1][2] * trans[2][1];
    rem3 = Cpara[1][2] - cpara[1][2] * trans[2][2];
    cpara[1][1] = norm( rem1, rem2, rem3 );
    trans[1][0] = rem1 / cpara[1][1];
    trans[1][1] = rem2 / cpara[1][1];
    trans[1][2] = rem3 / cpara[1][1];

    cpara[0][2] = dot( trans[2][0], trans[2][1], trans[2][2],
                       Cpara[0][0], Cpara[0][1], Cpara[0][2] );
    cpara[0][1] = dot( trans[1][0], trans[1][1], trans[1][2],
                       Cpara[0][0], Cpara[0][1], Cpara[0][2] );
    rem1 = Cpara[0][0] - cpara[0][1]*trans[1][0] - cpara[0][2]*trans[2][0];
    rem2 = Cpara[0][1] - cpara[0][1]*trans[1][1] - cpara[0][2]*trans[2][1];
    rem3 = Cpara[0][2] - cpara[0][1]*trans[1][2] - cpara[0][2]*trans[2][2];
    cpara[0][0] = norm( rem1, rem2, rem3 );
    trans[0][0] = rem1 / cpara[0][0];
    trans[0][1] = rem2 / cpara[0][0];
    trans[0][2] = rem3 / cpara[0][0];

    trans[1][3] = (Cpara[1][3] - cpara[1][2]*trans[2][3]) / cpara[1][1];
    trans[0][3] = (Cpara[0][3] - cpara[0][1]*trans[1][3]
                               - cpara[0][2]*trans[2][3]) / cpara[0][0];

    for( r = 0; r < 3; r++ ){
	for( c = 0; c < 3; c++ ){
            cpara[r][c] /= cpara[2][2];
	}
    }

    return 0;
}

static ARFloat norm( ARFloat a, ARFloat b, ARFloat c )
{
    return( (ARFloat)sqrt( a*a + b*b + c*c ) );
}

static ARFloat dot( ARFloat a1, ARFloat a2, ARFloat a3,
		   ARFloat b1, ARFloat b2, ARFloat b3 )
{
    return( a1 * b1 + a2 * b2 + a3 * b3 );
}


}  // namespace ARToolKitPlus
// End of ../../src/core/paramDecomp.cxx
// Included ../../src/core/paramDistortion.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id: paramDistortion.cxx 172 2006-07-25 14:05:47Z daniel $
 * @file
 * ======================================================================== */


#include <stdio.h>
#include <math.h>
#include <ARToolKitPlus/Tracker.h>
#include <ARToolKitPlus/Camera.h>
#include <ARToolKitPlus/param.h>


namespace ARToolKitPlus {


AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arParamObserv2Ideal_std(Camera* pCam, ARFloat ox, ARFloat oy, ARFloat *ix, ARFloat *iy)
{
	pCam->observ2Ideal(ox,oy,ix,iy);
	return(0);
}

AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arParamIdeal2Observ_std(Camera* pCam, ARFloat ix, ARFloat iy, ARFloat *ox, ARFloat *oy)
{
	pCam->ideal2Observ(ix,iy,ox,oy);
	return(0);
}


AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arParamObserv2Ideal_none(Camera* pCam, ARFloat ox, ARFloat oy, ARFloat *ix, ARFloat *iy)
{
	*ix = ox;
	*iy = oy;
	return(0);
}

//
//  these functions store 2 values (x & y) in a single 32-bit unsigned integer
//  each value is stored as 11.5 fixed point
//

inline
void floatToFixed(ARFloat nX, ARFloat nY, unsigned int &nFixed)
{
	short sx = (short)(nX * 32);
	short sy = (short)(nY * 32);

	unsigned short *ux = (unsigned short*)&sx;
	unsigned short *uy = (unsigned short*)&sy;

	nFixed = (*ux << 16) | *uy;
}


inline
void fixedToFloat(unsigned int nFixed, ARFloat& nX, ARFloat& nY)
{
	unsigned short ux = (nFixed >> 16);
	unsigned short uy = (nFixed & 0xffff);

	short *sx = (short*)&ux;
	short *sy = (short*)&uy;

	nX = (*sx)/32.0f;
	nY = (*sy)/32.0f;
}


AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arParamObserv2Ideal_LUT(Camera* pCam, ARFloat ox, ARFloat oy, ARFloat *ix, ARFloat *iy)
{
	if(!undistO2ITable)
		buildUndistO2ITable(pCam);

	int x=(int)ox, y=(int)oy;

	fixedToFloat(undistO2ITable[x+y*arImXsize], *ix,*iy);
	return 0;
}


AR_TEMPL_FUNC void
AR_TEMPL_TRACKER::buildUndistO2ITable(Camera* pCam)
{
	int x,y;
	ARFloat cx,cy, ox,oy;
	unsigned int fixed;
	char* cachename = NULL;
	bool loaded = false;

	if(loadCachedUndist)
	{
		assert(pCam->getFileName());
		cachename = new char[strlen(pCam->getFileName())+5];
		strcpy(cachename, pCam->getFileName());
		strcat(cachename, ".LUT");
	}

	// we have to take care here when using a memory manager that can not free memory
	// (usually this lookup table should only be built once - unless we change camera resolution)
	//
	if(undistO2ITable)
		//delete undistO2ITable;
		artkp_Free(undistO2ITable);

	//undistO2ITable = new unsigned int [arImXsize*arImYsize];
	undistO2ITable = artkp_Alloc<unsigned int>(arImXsize*arImYsize);

	if(loadCachedUndist)
	{
		if(FILE* fp = fopen(cachename, "rb"))
		{
			size_t numBytes = fread(undistO2ITable, 1, arImXsize*arImYsize*sizeof(unsigned int), fp);
			fclose(fp);

			if(numBytes == arImXsize*arImYsize*sizeof(unsigned int))
				loaded = true;
		}
	}

	if(!loaded)
	{
		for(x=0; x<arImXsize; x++)
		{
			for(y=0; y<arImYsize; y++)
			{
				arParamObserv2Ideal_std(pCam, (ARFloat)x, (ARFloat)y, &cx, &cy);
				floatToFixed(cx,cy, fixed);
				fixedToFloat(fixed, ox,oy);
				undistO2ITable[x+y*arImXsize] = fixed;
			}
		}

		if(loadCachedUndist)
			if(FILE* fp = fopen(cachename, "wb"))
			{
				fwrite(undistO2ITable, 1, arImXsize*arImYsize*sizeof(unsigned int), fp);
				fclose(fp);
			}
	}

	delete cachename;
}


AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arParamObserv2Ideal(Camera *pCam, ARFloat ox, ARFloat oy, ARFloat *ix, ARFloat *iy)
{
	pCam->observ2Ideal(ox,oy,ix,iy);
	return(0);
}

AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arParamIdeal2Observ(Camera *pCam, ARFloat ix, ARFloat iy, ARFloat *ox, ARFloat *oy)
{
	pCam->ideal2Observ(ix,iy,ox,oy);
	return(0);
}


}  // namespace ARToolKitPlus


// End of ../../src/core/paramDistortion.cxx
// Included ../../src/core/byteSwap.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Thomas Pintaric
 *
 * $Id: byteSwap.cxx 162 2006-04-19 21:28:10Z grabner $
 * @file
 * ======================================================================== */


#include <ARToolKitPlus/param.h>

namespace ARToolKitPlus {


#ifdef AR_LITTLE_ENDIAN

typedef union {
	int  x;
	unsigned char y[4];
} SwapIntT;

typedef union {
	ARFloat  x;
	unsigned char y[8];
} SwapDoubleT;


static void
byteSwapInt( int *from, int *to )
{
    SwapIntT   *w1, *w2;
    int        i;

    w1 = (SwapIntT *)from;
    w2 = (SwapIntT *)to;
    for( i = 0; i < 4; i++ ) {
        w2->y[i] = w1->y[3-i];
    }

    return;
}

static void
byteSwapDouble( double *from, double *to )
{
    SwapDoubleT   *w1, *w2;
    int           i;

    w1 = (SwapDoubleT *)from;
    w2 = (SwapDoubleT *)to;
    for( i = 0; i < 8; i++ ) {
        w2->y[i] = w1->y[7-i];
    }

    return;
}

static void
byteswap( ARParamDouble *param )
{
    ARParamDouble  wparam;
    int            i, j;

    byteSwapInt( &(param->xsize), &(wparam.xsize) );
    byteSwapInt( &(param->ysize), &(wparam.ysize) );

    for( j = 0; j < 3; j++ ) {
        for( i = 0; i < 4; i++ ) {
            byteSwapDouble( &(param->mat[j][i]), &(wparam.mat[j][i]) );
        }
    }

    for( i = 0; i < 4; i++ ) {
        byteSwapDouble( &(param->dist_factor[i]), &(wparam.dist_factor[i]) );
    }

    *param = wparam;
}

#endif  //AR_LITTLE_ENDIAN

}  // namespace ARToolKitPlus
// End of ../../src/core/byteSwap.cxx
// Included ../../src/core/paramFile.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id: paramFile.cxx 162 2006-04-19 21:28:10Z grabner $
 * @file
 * ======================================================================== */


#include <stdio.h>
#include <stdarg.h>
#include <ARToolKitPlus/Tracker.h>
#include <ARToolKitPlus/param.h>


namespace ARToolKitPlus {

AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arParamSaveDouble(char *filename, int num, ARParamDouble *param, ...)
{
    FILE           *fp;
    va_list        ap;
    ARParamDouble  *param1;
    int            i;

    if( num < 1 ) return -1;

    fp = fopen( filename, "wb" );
    if( fp == NULL ) return -1;

#ifdef AR_LITTLE_ENDIAN
    byteswap( param );
#endif
    if( fwrite( param, sizeof(ARParamDouble), 1, fp ) != 1 ) {
        fclose(fp);
#ifdef AR_LITTLE_ENDIAN
        byteswap( param );
#endif
        return -1;
    }
#ifdef AR_LITTLE_ENDIAN
    byteswap( param );
#endif

    va_start(ap, param);
    for( i = 1; i < num; i++ ) {
        param1 = va_arg(ap, ARParamDouble*);
#ifdef AR_LITTLE_ENDIAN
        byteswap( param1 );
#endif
        if( fwrite( param1, sizeof(ARParamDouble), 1, fp ) != 1 ) {
            fclose(fp);
#ifdef AR_LITTLE_ENDIAN
            byteswap( param1 );
#endif
            return -1;
        }
#ifdef AR_LITTLE_ENDIAN
        byteswap( param1 );
#endif
    }

    fclose(fp);

    return 0;
}

/*
AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arParamSave( char *filename, int num, ARParam *param, ...)
{
	// load really the double version
	//
	ARParamDouble	tmpParam;
	int			    x,y;

	// copy the values to the runtime types
	//
	tmpParam.xsize = param->xsize;
	tmpParam.ysize = param->ysize;
	for(y=0; y<3; y++)
		for(x=0; x<4; x++)
			tmpParam.mat[y][x] = param->mat[y][x];
	for(x=0; x<4; x++)
		tmpParam.dist_factor[x] = param->dist_factor[x];


	arParamSaveDouble(filename, 1, &tmpParam);
    return 0;
}
*/

AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arParamLoadDouble( char *filename, int num, ARParamDouble *param, ...)
{
    FILE           *fp;
    va_list        ap;
    ARParamDouble  *param1;
    int            i;

    if( num < 1 ) return -1;
	
    fp = fopen( filename, "rb" );
    if( fp == NULL ) return -1;

    if( fread( param, sizeof(ARParamDouble), 1, fp ) != 1 ) {
        fclose(fp);
        return -1;
    }
#ifdef AR_LITTLE_ENDIAN
    byteswap( param );
#endif

    va_start(ap, param);
    for( i = 1; i < num; i++ ) {
        param1 = va_arg(ap, ARParamDouble*);
        if( fread( param1, sizeof(ARParamDouble), 1, fp ) != 1 ) {
            fclose(fp);
            return -1;
        }
#ifdef AR_LITTLE_ENDIAN
        byteswap( param1 );
#endif
    }

    fclose(fp);

    return 0;
}

/*
AR_TEMPL_FUNC int
AR_TEMPL_TRACKER::arParamLoad( char *filename, int num, ARParam *param, ...)
{
	// load really the double version
	//
	ARParamDouble	tmpParam;
	int			    x,y;

	if(arParamLoadDouble(filename, 1, &tmpParam)<0)
		return -1;

	// copy the values to the runtime types
	//
	param->xsize = tmpParam.xsize;
	param->ysize = tmpParam.ysize;
	for(y=0; y<3; y++)
		for(x=0; x<4; x++)
			param->mat[y][x] = (ARFloat)tmpParam.mat[y][x];
	for(x=0; x<4; x++)
		param->dist_factor[x] = (ARFloat)tmpParam.dist_factor[x];

    return 0;
}
*/

/*
static int
ARToolKitPlus::Param::display( ARParam *param )
{
    int     i, j;

    printf("--------------------------------------\n");
    printf("SIZE = %d, %d\n", param->xsize, param->ysize);
    printf("Distotion factor = %f %f %f %f\n", param->dist_factor[0],
            param->dist_factor[1], param->dist_factor[2], param->dist_factor[3] );
    for( j = 0; j < 3; j++ ) {
        for( i = 0; i < 4; i++ ) printf("%7.5f ", param->mat[j][i]);
        printf("\n");
    }
    printf("--------------------------------------\n");

    return 0;
}
*/

}  // namespace ARToolKitPlus
// End of ../../src/core/paramFile.cxx
// Included ../../src/core/vector.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id: vector.cxx 162 2006-04-19 21:28:10Z grabner $
 * @file
 * ======================================================================== */


#include <stdlib.h>
#include <stdio.h>
#ifndef __APPLE__
#include <malloc.h>
#else
#include <stdlib.h>
#endif
#include <math.h>

#include <ARToolKitPlus/Tracker.h>
#include <ARToolKitPlus/matrix.h>
#include <ARToolKitPlus/vector.h>


namespace ARToolKitPlus {


namespace Vector {


// from vAlloc.c
static ARVec*
alloc( int clm )
{
    ARVec     *v;

    v = (ARVec *)malloc(sizeof(ARVec));
    if( v == NULL ) return NULL;

    v->v = (ARFloat *)malloc(sizeof(ARFloat) * clm);
    if( v->v == NULL ) {
        free(v);
        return NULL;
    }

    v->clm = clm;

    return v;
}


// from vDisp.c
/*
static int
disp( ARVec *v )
{
    int    c;
    
    if( v == NULL ) return -1;

    printf(" === vector (%d) ===\n", v->clm);
    printf(" |");
    for( c = 0; c < v->clm; c++ ){
	printf( " %10g", v->v[c] );
    }
    printf(" |\n");
    printf(" ===================\n");

    return 0;
}
*/

// from vFree.c
static int
free( ARVec *v )
{
    ::free( v->v );
    ::free( v );

    return 0;
}


// from vHouse.c
static ARFloat
household( ARVec *x )
{
    ARFloat s, t;
    int    i;

    s = (ARFloat)sqrt( innerproduct(x,x) );

    if( s != 0.0 ) {
        if(x->v[0] < 0) s = -s;
        x->v[0] += s;
        t = (ARFloat)1 / (ARFloat)sqrt(x->v[0] * s);
        for( i = 0; i < x->clm; i++ ) {
            x->v[i] *= t;
        }
    }

    return(-s);
}


// from vInnerP.c
static ARFloat
innerproduct( ARVec *x, ARVec *y )
{
    ARFloat   result = 0.0;
    int      i;

    if( x->clm != y->clm ) exit(0);

    for( i = 0; i < x->clm; i++ ) {
        result += x->v[i] * y->v[i];
    }

    return( result );
}


// from vTridiag.c
static int
tridiagonalize( ARMat *a, ARVec *d, ARVec *e )
{
    ARVec     wv1, wv2;
    ARFloat  *v;
    ARFloat  s, t, p, q;
    int     dim;
    int     i, j, k;

    if( a->clm != a->row )   return(-1);
    if( a->clm != d->clm )   return(-1);
    if( a->clm != e->clm+1 ) return(-1);
    dim = a->clm;

    for( k = 0; k < dim-2; k++ ) {
        v = &(a->m[k*dim]);
        d->v[k] = v[k];

        wv1.clm = dim-k-1;
        wv1.v = &(v[k+1]);
        e->v[k] = household(&wv1);
        if( e->v[k] == 0.0 ) continue;

        for( i = k+1; i < dim; i++ ) {
            s = 0.0;
            for( j = k+1; j < i; j++ ) {
                s += a->m[j*dim+i] * v[j];
            }
            for( j = i; j < dim; j++ ) {
                s += a->m[i*dim+j] * v[j];
            }
            d->v[i] = s;
        }

        wv1.clm = wv2.clm = dim-k-1;
        wv1.v = &(v[k+1]);
        wv2.v = &(d->v[k+1]);
        t = innerproduct( &wv1, &wv2 ) / 2;
        for( i = dim-1; i > k; i-- ) {
            p = v[i];
            q = d->v[i] -= t*p;
            for( j = i; j < dim; j++ ) {
                a->m[i*dim+j] -= p*(d->v[j]) + q*v[j];
            }
        }
    }

    if( dim >= 2) {
        d->v[dim-2] = a->m[(dim-2)*dim+(dim-2)];
        e->v[dim-2] = a->m[(dim-2)*dim+(dim-1)];
    }

    if( dim >= 1 ) d->v[dim-1] = a->m[(dim-1)*dim+(dim-1)];

    for( k = dim-1; k >= 0; k-- ) {
        v = &(a->m[k*dim]);
        if( k < dim-2 ) {
            for( i = k+1; i < dim; i++ ) {
                wv1.clm = wv2.clm = dim-k-1;
                wv1.v = &(v[k+1]);
                wv2.v = &(a->m[i*dim+k+1]);
                t = innerproduct( &wv1, &wv2 );
                for( j = k+1; j < dim; j++ ) a->m[i*dim+j] -= t * v[j];
            }
        }
        for( i = 0; i < dim; i++ ) v[i] = 0.0;
        v[k] = 1;
    }

    return(0);
}


}  // namespace Vector


}  // namespace ARToolKitPlus
// End of ../../src/core/vector.cxx

// Included ../../src/CameraImpl.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Thomas Pintaric
 *
 * $Id: CameraImpl.cxx 172 2006-07-25 14:05:47Z daniel $
 * @file
 * ======================================================================== */


#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <ARToolKitPlus/CameraImpl.h>
#include <ARToolKitPlus/byteSwap.h>


namespace ARToolKitPlus {

#define PD_LOOP 3

CameraImpl::CameraImpl()
{}

CameraImpl::~CameraImpl()
{}

bool CameraImpl::
loadFromFile(const char* filename)
{
	FILE           *fp;
	ARParamDouble  param;

	fp = fopen( filename, "rb" );
	if( fp == NULL ) return(false);

	setFileName(filename);

	if(fread(&param, sizeof(ARParamDouble), 1, fp ) != 1 ) 
	{
		fclose(fp);
		return(false);
	}

#ifdef AR_LITTLE_ENDIAN
	byteswap( &param );
#endif

	fclose(fp);

	unsigned int i,j;
	for(i=0; i<4; i++) this->dist_factor[i] = (ARFloat) param.dist_factor[i];
	this->xsize = (int) param.xsize;
	this->ysize = (int) param.ysize;
	for(i=0; i<3; i++)
		for(j=0; j<4; j++)
			this->mat[i][j] = (ARFloat) param.mat[i][j];

	if( (mat[0][1] != 0.0) ||
		(mat[0][3] != 0.0) ||
		(mat[1][0] != 0.0) ||
		(mat[1][3] != 0.0) ||
		(mat[2][0] != 0.0) ||
		(mat[2][1] != 0.0) ||
		(mat[2][2] != 1.0) ||
		(mat[2][3] != 0.0)) return(false);

	return(true);
}



void CameraImpl::
observ2Ideal(ARFloat ox, ARFloat oy, ARFloat *ix, ARFloat *iy)
{
	ARFloat  z02, z0, p, q, z, px, py;
	int     i;

	px = ox - this->dist_factor[0];
	py = oy - this->dist_factor[1];
	p = this->dist_factor[2]/(ARFloat)100000000.0;
	z02 = px*px+ py*py;
	q = z0 = (ARFloat)sqrt(px*px+ py*py);

	for( i = 1; ; i++ ) {
		if( z0 != 0.0 ) {
			z = z0 - (((ARFloat)1.0 - p*z02)*z0 - q) / ((ARFloat)1.0 - (ARFloat)3.0*p*z02);
			px = px * z / z0;
			py = py * z / z0;
		}
		else {
			px = 0.0;
			py = 0.0;
			break;
		}
		if( i == PD_LOOP ) break;

		z02 = px*px+ py*py;
		z0 = (ARFloat)sqrt(px*px+ py*py);
	}

	*ix = px / this->dist_factor[3] + this->dist_factor[0];
	*iy = py / this->dist_factor[3] + this->dist_factor[1];
}

void CameraImpl::
ideal2Observ(ARFloat ix, ARFloat iy, ARFloat *ox, ARFloat *oy)
{
	ARFloat    x, y, d;

	x = (ix - this->dist_factor[0]) * this->dist_factor[3];
	y = (iy - this->dist_factor[1]) * this->dist_factor[3];
	if( x == 0.0 && y == 0.0 ) {
		*ox = this->dist_factor[0];
		*oy = this->dist_factor[1];
	}
	else {
		d = (ARFloat)1.0 - this->dist_factor[2]/(ARFloat)100000000.0 * (x*x+y*y);
		*ox = x * d + this->dist_factor[0];
		*oy = y * d + this->dist_factor[1];
	}
}

Camera* CameraImpl::clone()
{
	CameraImpl* pCam = new CameraImpl();
	pCam->xsize = xsize;
	pCam->ysize = ysize;
	unsigned int i,j;
	for(i=0; i<3; i++) for(j=0; j<4; j++) pCam->mat[i][j] = mat[i][j];
	for(i=0; i<4; i++) pCam->dist_factor[i] = dist_factor[i];
	return((Camera*)pCam);
}

bool CameraImpl::
changeFrameSize(const int frameWidth, const int frameHeight)
{
	if(frameWidth <=0 || frameHeight <=0) return(false);
	const ARFloat scale = (ARFloat)frameWidth / (ARFloat)xsize;
	xsize = frameWidth;
	ysize = frameHeight;

	for(int i = 0; i < 4; i++ )
	{
		mat[0][i] *= scale;
		mat[1][i] *= scale;
	}

	dist_factor[0] *= scale;
	dist_factor[1] *= scale;
	dist_factor[2] /= (scale*scale);

	return(true);
}

void CameraImpl::
logSettings(Logger* logger)
{
	if(logger != NULL)
	{
		logger->artLogEx("ARToolKitPlus: CamSize %d , %d\n", xsize, ysize);
		logger->artLogEx("ARToolKitPlus: Dist.Factor %.2f %.2f %.2f %.2f\n", dist_factor[0],
			dist_factor[1], dist_factor[2], dist_factor[3] );
	}
}


}  // namespace ARToolKitPlus
// End of ../../src/CameraImpl.cxx
// Included ../../src/CameraAdvImpl.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Thomas Pintaric
 *
 * $Id: CameraAdvImpl.cxx 172 2006-07-25 14:05:47Z daniel $
 * @file
 * ======================================================================== */


#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <ARToolKitPlus/CameraAdvImpl.h>


namespace ARToolKitPlus {


CameraAdvImpl::CameraAdvImpl()
{}

CameraAdvImpl::~CameraAdvImpl()
{}

bool CameraAdvImpl::
loadFromFile(const char* filename)
{
	FILE *fp = fopen( filename, "r" );
	if( fp == NULL ) return(false);

	setFileName(filename);

	double _cc[2];
	double _fc[2];
	double _kc[6];
	int n;

	char _s[32];
	sprintf(_s,"%c%i%c%c%c",'%',(int)strlen(CAMERA_ADV_HEADER),'s','\\','n');
	char hdr[MAX_PATH];
	n=fscanf(fp, _s, hdr);
	if(strstr(hdr,CAMERA_ADV_HEADER) == NULL) return(false);

	n=fscanf(fp, "%d%d%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf%d\n",
		&this->xsize, &this->ysize, &_cc[0],&_cc[1],&_fc[0],&_fc[1],
		&_kc[0],&_kc[1],&_kc[2],&_kc[3],&_kc[4],&_kc[5],&undist_iterations); 
	if((n!=13) || ferror(fp))
	{
		return(false);
	}

	unsigned int i,j;
	this->cc[0] = (ARFloat) _cc[0];
	this->cc[1] = (ARFloat) _cc[1];
	this->fc[0] = (ARFloat) _fc[0];
	this->fc[1] = (ARFloat) _fc[1];
	for(i=0; i<6; i++) this->kc[i] = (ARFloat) _kc[i];

	for(i=0; i<3; i++)
		for(j=0; j<4; j++)
			this->mat[i][j] = 0.;

	mat[0][0] = fc[0]; // fc_x
	mat[1][1] = fc[1]; // fc_y
	mat[0][2] = cc[0]; // cc_x
	mat[1][2] = cc[1]; // cc_y
	mat[2][2] = 1.0;

	if (undist_iterations > CAMERA_ADV_MAX_UNDIST_ITERATIONS)
		undist_iterations = CAMERA_ADV_MAX_UNDIST_ITERATIONS;

	fclose(fp);
	return(true);
}


void CameraAdvImpl::
observ2Ideal(ARFloat ox, ARFloat oy, ARFloat *ix, ARFloat *iy)
{
	if(undist_iterations <= 0)
	{
		*ix = ox;
		*iy = oy;
	}
	else
	{
		const ARFloat xd[2] = { (ox - cc[0]) / fc[0], (oy - cc[1]) / fc[1] };
		const ARFloat k1 = kc[0];
		const ARFloat k2 = kc[1];
		const ARFloat k3 = kc[4];
		const ARFloat p1 = kc[2];
		const ARFloat p2 = kc[3];

		ARFloat x[2] = { xd[0], xd[1] };
		for(int kk=0; kk<undist_iterations; kk++)
		{
			const ARFloat x0_sq = (x[0]*x[0]);
			const ARFloat x1_sq = (x[1]*x[1]);
			const ARFloat x0_x1 = (x[0]*x[1]);
			const ARFloat r_2 = x0_sq + x1_sq;
			const ARFloat r_2_sq = (r_2 * r_2);
			const ARFloat k_radial =  1 + k1 * r_2 + k2 * (r_2_sq) + k3 * (r_2*r_2_sq);
			const ARFloat delta_x[2] = {   2*p1*x0_x1 + p2*(r_2 + 2*x0_sq),
				p1 * (r_2 + 2*x1_sq) + 2*p2*x0_x1   };
			x[0] = xd[0] - delta_x[0];
			x[1] = xd[1] - delta_x[1];
			x[0] /= k_radial;
			x[1] /= k_radial;
		}

		*ix = (x[0] * fc[0]) + cc[0];
		*iy = (x[1] * fc[1]) + cc[1];
	}
}

void CameraAdvImpl::
ideal2Observ(ARFloat ix, ARFloat iy, ARFloat *ox, ARFloat *oy)
{
	const ARFloat xu[2] = { (ix - cc[0]) / fc[0], (iy - cc[1]) / fc[1] };

	const ARFloat r2 = (xu[0]*xu[0]) + (xu[1]*xu[1]);
	const ARFloat r4 = r2*r2;
	const ARFloat r6 = r4*r2;
	const ARFloat cdist = 1 + kc[0] * r2 + kc[1] * r4 + kc[4] * r6;

	const ARFloat a1 = 2*xu[0]*xu[1];
	const ARFloat a2 = r2 + 2*(xu[0]*xu[0]);
	const ARFloat a3 = r2 + 2*(xu[1]*xu[1]);

	*ox = (xu[0] * cdist) + (kc[2]*a1 + kc[3]*a2);
	*oy = (xu[1] * cdist) + (kc[2]*a3 + kc[3]*a1);
}

Camera* CameraAdvImpl::clone()
{
	CameraAdvImpl* pCam = new CameraAdvImpl();
	pCam->xsize = xsize;
	pCam->ysize = ysize;
	unsigned int i,j;
	for(i=0; i<3; i++) for(j=0; j<4; j++) pCam->mat[i][j] = mat[i][j];
	for(i=0; i<4; i++) pCam->dist_factor[i] = dist_factor[i];
	pCam->cc[0] = cc[0];
	pCam->cc[1] = cc[1];
	pCam->fc[0] = fc[0];
	pCam->fc[1] = fc[1];
	for(i=0; i<6; i++) pCam->kc[i] = kc[i];
	pCam->undist_iterations = undist_iterations;
	return((Camera*)pCam);
}

bool CameraAdvImpl::
changeFrameSize(const int frameWidth, const int frameHeight)
{
	if(frameWidth <=0 || frameHeight <=0) return(false);
	const ARFloat scale = (ARFloat)frameWidth / (ARFloat)xsize;
	xsize = frameWidth;
	ysize = frameHeight;

	for(int i = 0; i < 4; i++ )
	{
		mat[0][i] *= scale;
		mat[1][i] *= scale;
	}

	cc[0] *= scale; 	cc[1] *= scale;
	fc[0] *= scale; 	fc[1] *= scale;

	return(true);
}

void CameraAdvImpl::
logSettings(Logger* logger)
{
	if(logger != NULL)
	{
		logger->artLogEx("ARToolKitPlus: CamSize %d , %d\n", xsize, ysize);
		logger->artLogEx("ARToolKitPlus: cc = [%.2f  %.2f]  fc = [%.2f  %.2f]\n", 
			cc[0], cc[1], fc[0], fc[1]);
		logger->artLogEx("ARToolKitPlus: kc = [%.4f %.4f %.4f %.4f %.4f %.4f]\n", 
			kc[0], kc[1], kc[2], kc[3], kc[4], kc[5]);
		logger->artLogEx("ARToolKitPlus: undist_iterations = %i\n", undist_iterations); 
	}
}


}  // namespace ARToolKitPlus
// End of ../../src/CameraAdvImpl.cxx
// Included ../../src/CameraFactory.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Thomas Pintaric
 *
 * $Id: CameraFactory.cxx 162 2006-04-19 21:28:10Z grabner $
 * @file
 * ======================================================================== */


#include <ARToolKitPlus/CameraFactory.h>
#include <ARToolKitPlus/CameraImpl.h>
#include <ARToolKitPlus/CameraAdvImpl.h>

namespace ARToolKitPlus {

Camera* CameraFactory::
createCamera(const char* filename)
{
	CameraAdvImpl *c_ptr = new CameraAdvImpl();
	if(c_ptr->loadFromFile(filename)) return((Camera*)c_ptr);
	else delete c_ptr;

	CameraImpl *c2_ptr = new CameraImpl();
	if(c2_ptr->loadFromFile(filename)) return((Camera*)c2_ptr);
	else delete c2_ptr;

	return(NULL);
}

}  // namespace ARToolKitPlus
// End of ../../src/CameraFactory.cxx
// Included ../../src/extra/BCH.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 *
 * See src/extra/BCH_original.txt for details/orig. copyright notice
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Thomas Pintaric
 *
 * $Id: BCH.cxx 162 2006-04-19 21:28:10Z grabner $
 * @file
 * ======================================================================== */


#include <ARToolKitPlus/extra/BCH.h>
#include "assert.h"

#include <vector>

namespace ARToolKitPlus {


static bool
_isBitSet(_64bits n, int which_bit)
{
	return ((n>>which_bit)&1) != 0;
}


static void
_setBit(_64bits &n, int which_bit)
{
	const _64bits one = 1;
	n |= (one<<which_bit);
}

/*
static void
_clearBit(_64bits &n, int which_bit)
{
	const _64bits one = 1;
	n &= ~(one<<which_bit);
}
*/

/*
static void
_copyBit(_64bits &dest_n, const int dest_bit, const _64bits src_n, const int src_bit)
{
	const _64bits one = 1;
	if(((src_n>>src_bit)&1) != 0) dest_n |= (one<<dest_bit);
	else dest_n &= ~(one<<dest_bit);
}
*/

/*
static int
_countOnes(const _64bits src_n)
{
	int cnt = 0;
	for(int i=0; i<64; i++)
		cnt += (int)((src_n>>i)&1);
	return(cnt);
}
*/

static int*
toBitPattern(int b[], _64bits n, int n_bits)
{
	for(int i=0; i<n_bits; i++)
		b[i] = (_isBitSet(n,i) ? 1:0);
	return b;
}

static _64bits
fromBitPattern(int b[], int n_bits)
{
	_64bits n = 0;
	for(int i=0; i<n_bits; i++)
		if(b[i] == 1) _setBit(n,i);
	return(n);
}

/*
static void
printBitPattern(_64bits n, int n_bits)
{
	for(int i=0; i<n_bits; i++)
	{
		if(_isBitSet(n,i)) printf("1");
		else printf("0");
	}
}
*/

/*
static void
printBitArray(int bb[], int bb_length)
{
	for(int i=0; i<bb_length; i++) printf("%i",bb[i]);
}
*/


BCH::BCH()
{
	initialize(BCH_DEFAULT_M, BCH_DEFAULT_LENGTH, BCH_DEFAULT_T);
}

BCH::BCH(int _m, int _length, int _t)
{
	initialize(_m,_length,_t);
}


void BCH::initialize(int _m, int _length, int _t)
{
	int i, ninf;

	m = _m;
	length = _length;
	t = _t;

	p.resize(BCH_MAX_P);
	alpha_to.resize(BCH_MAX_LUT);
	index_of.resize(BCH_MAX_LUT);
	g.resize(BCH_MAX_LUT);

	_elp.resize(BCH_MAX_LUT);
	for(i=0; i<BCH_MAX_LUT; i++) _elp[i].resize(BCH_MAX_LUT);
	_d.resize(BCH_MAX_LUT);
	_l.resize(BCH_MAX_LUT);
	_u_lu.resize(BCH_MAX_LUT);
	_s.resize(BCH_MAX_LUT);
	_root.resize(BCH_MAX_LUT);
	_loc.resize(BCH_MAX_LUT);
	_reg.resize(BCH_MAX_LUT);


	for (i=1; i<m; i++)
		p[i] = 0;
	p[0] = p[m] = 1;
	if (m == 2)			p[1] = 1;
	else if (m == 3)	p[1] = 1;
	else if (m == 4)	p[1] = 1;
	else if (m == 5)	p[2] = 1;
	else if (m == 6)	p[1] = 1;
	else if (m == 7)	p[1] = 1;
	else if (m == 8)	p[4] = p[5] = p[6] = 1;
	else if (m == 9)	p[4] = 1;
	else if (m == 10)	p[3] = 1;
	else if (m == 11)	p[2] = 1;
	else if (m == 12)	p[3] = p[4] = p[7] = 1;
	else if (m == 13)	p[1] = p[3] = p[4] = 1;
	else if (m == 14)	p[1] = p[11] = p[12] = 1;
	else if (m == 15)	p[1] = 1;
	else if (m == 16)	p[2] = p[3] = p[5] = 1;
	else if (m == 17)	p[3] = 1;
	else if (m == 18)	p[7] = 1;
	else if (m == 19)	p[1] = p[5] = p[6] = 1;
	else if (m == 20)	p[3] = 1;

    n = 1;
	for (i = 0; i <= m; i++) 
	{
        n *= 2;
    }
	n = n / 2 - 1;
	ninf = (n + 1) / 2 - 1;

	generate_gf();          /* Construct the Galois Field GF(2**m) */
	gen_poly(t);            /* Compute the generator polynomial of BCH code */
}


void BCH::generate_gf()
/*
 * Generate field GF(2**m) from the irreducible polynomial p(X) with
 * coefficients in p[0]..p[m].
 *
 * Lookup tables:
 *   index->polynomial form: alpha_to[] contains j=alpha^i;
 *   polynomial form -> index form:	index_of[j=alpha^i] = i
 *
 * alpha=2 is the primitive element of GF(2**m) 
 */
{
	int    i, mask;

	mask = 1;
	alpha_to[m] = 0;
	for (i = 0; i < m; i++) {
		alpha_to[i] = mask;
		index_of[alpha_to[i]] = i;
		if (p[i] != 0)
			alpha_to[m] ^= mask;
		mask <<= 1;
	}
	index_of[alpha_to[m]] = m;
	mask >>= 1;
	for (i = m + 1; i < n; i++) {
		if (alpha_to[i - 1] >= mask)
		  alpha_to[i] = alpha_to[m] ^ ((alpha_to[i - 1] ^ mask) << 1);
		else
		  alpha_to[i] = alpha_to[i - 1] << 1;
		index_of[alpha_to[i]] = i;
	}
	index_of[0] = -1;
}


bool BCH::gen_poly(int _t)
/*
 * Compute the generator polynomial of a binary BCH code. Fist generate the
 * cycle sets modulo 2**m - 1, cycle[][] =  (i, 2*i, 4*i, ..., 2^l*i). Then
 * determine those cycle sets that contain integers in the set of (d-1)
 * consecutive integers {1..(d-1)}. The generator polynomial is calculated
 * as the product of linear factors of the form (x+alpha^i), for every i in
 * the above cycle sets.
 */
{
	int	ii, jj, ll, kaux;
	int	test, aux, nocycles, root, noterms, rdncy;
	int             cycle[1024][21], size[1024], min[1024], zeros[1024];

	/* Generate cycle sets modulo n, n = 2**m - 1 */
	cycle[0][0] = 0;
	size[0] = 1;
	cycle[1][0] = 1;
	size[1] = 1;
	jj = 1;			/* cycle set index */

	#ifdef _BCH_VERBOSE_
		if (m > 9)  {
			printf("Computing cycle sets modulo %d\n", n);
			printf("(This may take some time)...\n");
		}
	#endif
	
	do {
		/* Generate the jj-th cycle set */
		ii = 0;
		do {
			ii++;
			cycle[jj][ii] = (cycle[jj][ii - 1] * 2) % n;
			size[jj]++;
			aux = (cycle[jj][ii] * 2) % n;
		} while (aux != cycle[jj][0]);
		/* Next cycle set representative */
		ll = 0;
		do {
			ll++;
			test = 0;
			for (ii = 1; ((ii <= jj) && (!test)); ii++)	
			/* Examine previous cycle sets */
			  for (kaux = 0; ((kaux < size[ii]) && (!test)); kaux++)
			     if (ll == cycle[ii][kaux])
			        test = 1;
		} while ((test) && (ll < (n - 1)));
		if (!(test)) {
			jj++;	/* next cycle set index */
			cycle[jj][0] = ll;
			size[jj] = 1;
		}
	} while (ll < (n - 1));
	nocycles = jj;		/* number of cycle sets modulo n */

	//printf("Enter the error correcting capability, t: ");
	//scanf("%d", &t);
	t = _t;

	d = 2 * t + 1;

	/* Search for roots 1, 2, ..., d-1 in cycle sets */
	kaux = 0;
	rdncy = 0;
	for (ii = 1; ii <= nocycles; ii++) {
		min[kaux] = 0;
		test = 0;
		for (jj = 0; ((jj < size[ii]) && (!test)); jj++)
			for (root = 1; ((root < d) && (!test)); root++)
				if (root == cycle[ii][jj])  {
					test = 1;
					min[kaux] = ii;
				}
		if (min[kaux]) {
			rdncy += size[min[kaux]];
			kaux++;
		}
	}
	noterms = kaux;
	kaux = 1;
	for (ii = 0; ii < noterms; ii++)
		for (jj = 0; jj < size[min[ii]]; jj++) {
			zeros[kaux] = cycle[min[ii]][jj];
			kaux++;
		}

	k = length - rdncy;

    if (k<0)
      {
		#ifdef _BCH_VERBOSE_
			printf("Parameters invalid!\n");
		#endif
         return(false);
      }

	#ifdef _BCH_VERBOSE_
		printf("This is a (%d, %d, %d) binary BCH code\n", length, k, d);
	#endif

	/* Compute the generator polynomial */
	g[0] = alpha_to[zeros[1]];
	g[1] = 1;		/* g(x) = (X + zeros[1]) initially */
	for (ii = 2; ii <= rdncy; ii++) {
	  g[ii] = 1;
	  for (jj = ii - 1; jj > 0; jj--)
	    if (g[jj] != 0)
	      g[jj] = g[jj - 1] ^ alpha_to[(index_of[g[jj]] + zeros[ii]) % n];
	    else
	      g[jj] = g[jj - 1];
	  g[0] = alpha_to[(index_of[g[0]] + zeros[ii]) % n];
	}

	#ifdef _BCH_VERBOSE_
		printf("Generator polynomial:\ng(x) = ");
		for (ii = 0; ii <= rdncy; ii++) {
		  printf("%d", g[ii]);
		  if (ii && ((ii % 50) == 0))
			printf("\n");
		}
		printf("\n");
	#endif
	return(true);
}



void BCH::encode_bch(int *bb, const int *data)
/*
 * Compute redundacy bb[], the coefficients of b(x). The redundancy
 * polynomial b(x) is the remainder after dividing x^(length-k)*data(x)
 * by the generator polynomial g(x).
 */
{
	int    i, j;
	int    feedback;

	for (i = 0; i < length - k; i++)
		bb[i] = 0;
	for (i = k - 1; i >= 0; i--) {
		feedback = data[i] ^ bb[length - k - 1];
		if (feedback != 0) {
			for (j = length - k - 1; j > 0; j--)
				if (g[j] != 0)
					bb[j] = bb[j - 1] ^ feedback;
				else
					bb[j] = bb[j - 1];
			bb[0] = g[0] && feedback;
		} else {
			for (j = length - k - 1; j > 0; j--)
				bb[j] = bb[j - 1];
			bb[0] = 0;
		}
	}
}


int BCH::decode_bch(int *recd)
/*
 * Simon Rockliff's implementation of Berlekamp's algorithm.
 *
 * Assume we have received bits in recd[i], i=0..(n-1).
 *
 * Compute the 2*t syndromes by substituting alpha^i into rec(X) and
 * evaluating, storing the syndromes in s[i], i=1..2t (leave s[0] zero) .
 * Then we use the Berlekamp algorithm to find the error location polynomial
 * elp[i].
 *
 * If the degree of the elp is >t, then we cannot correct all the errors, and
 * we have detected an uncorrectable error pattern. We output the information
 * bits uncorrected.
 *
 * If the degree of elp is <=t, we substitute alpha^i , i=1..n into the elp
 * to get the roots, hence the inverse roots, the error location numbers.
 * This step is usually called "Chien's search".
 *
 * If the number of errors located is not equal the degree of the elp, then
 * the decoder assumes that there are more than t errors and cannot correct
 * them, only detect them. We output the information bits uncorrected.
 */
{
	int i, j, u, q, t2, count = 0, syn_error = 0;
	bool too_many_errors = false;
	//int elp[BCH_MAX_LUT][BCH_MAX_LUT], d[BCH_MAX_LUT], l[BCH_MAX_LUT], u_lu[BCH_MAX_LUT], s[BCH_MAX_LUT];
	//int root[BCH_MAX_SQ], loc[BCH_MAX_SQ], reg[BCH_MAX_SQ];

	t2 = 2 * t;

	/* first form the syndromes */
	#ifdef _BCH_VERBOSE_
		printf("S(x) = ");
	#endif
	for (i = 1; i <= t2; i++) {
		_s[i] = 0;
		for (j = 0; j < length; j++)
			if (recd[j] != 0)
				_s[i] ^= alpha_to[(i * j) % n];
		if (_s[i] != 0)
			syn_error = 1; /* set error flag if non-zero syndrome */
/*
 * Note:    If the code is used only for ERROR DETECTION, then
 *          exit program here indicating the presence of errors.
 */
		/* convert syndrome from polynomial form to index form  */
		_s[i] = index_of[_s[i]];
		#ifdef _BCH_VERBOSE_
			printf("%3d ", _s[i]);
		#endif
	}

	#ifdef _BCH_VERBOSE_
		printf("\n");
	#endif

	if (syn_error) {	/* if there are errors, try to correct them */
		/*
		 * Compute the error location polynomial via the Berlekamp
		 * iterative algorithm. Following the terminology of Lin and
		 * Costello's book :   d[u] is the 'mu'th discrepancy, where
		 * u='mu'+1 and 'mu' (the Greek letter!) is the step number
		 * ranging from -1 to 2*t (see L&C),  l[u] is the degree of
		 * the elp at that step, and u_l[u] is the difference between
		 * the step number and the degree of the elp. 
		 */
		/* initialise table entries */
		_d[0] = 0;			/* index form */
		_d[1] = _s[1];		/* index form */
		_elp[0][0] = 0;		/* index form */
		_elp[1][0] = 1;		/* polynomial form */
		for (i = 1; i < t2; i++) {
			_elp[0][i] = -1;	/* index form */
			_elp[1][i] = 0;	/* polynomial form */
		}
		_l[0] = 0;
		_l[1] = 0;
		_u_lu[0] = -1;
		_u_lu[1] = 0;
		u = 0;
 
		do {
			u++;
			if (_d[u] == -1) {
				_l[u + 1] = _l[u];
				for (i = 0; i <= _l[u]; i++) {
					_elp[u + 1][i] = _elp[u][i];
					_elp[u][i] = index_of[_elp[u][i]];
				}
			} else
				/*
				 * search for words with greatest _u_lu[q] for
				 * which _d[q]!=0 
				 */
			{
				q = u - 1;
				while ((_d[q] == -1) && (q > 0))
					q--;
				/* have found first non-zero _d[q]  */
				if (q > 0) {
				  j = q;
				  do {
				    j--;
				    if ((_d[j] != -1) && (_u_lu[q] < _u_lu[j]))
				      q = j;
				  } while (j > 0);
				}
 
				/*
				 * have now found q such that _d[u]!=0 and
				 * _u_lu[q] is maximum 
				 */
				/* store degree of new elp polynomial */
				if (_l[u] > _l[q] + u - q)
					_l[u + 1] = _l[u];
				else
					_l[u + 1] = _l[q] + u - q;
 
				/* form new elp(x) */
				for (i = 0; i < t2; i++)
					_elp[u + 1][i] = 0;
				for (i = 0; i <= _l[q]; i++)
					if (_elp[q][i] != -1)
						_elp[u + 1][i + u - q] = 
                                   alpha_to[(_d[u] + n - _d[q] + _elp[q][i]) % n];
				for (i = 0; i <= _l[u]; i++) {
					_elp[u + 1][i] ^= _elp[u][i];
					_elp[u][i] = index_of[_elp[u][i]];
				}
			}
			_u_lu[u + 1] = u - _l[u + 1];
 
			/* form (u+1)th discrepancy */
			if (u < t2) {	
			/* no discrepancy computed on last iteration */
			  if (_s[u + 1] != -1)
			    _d[u + 1] = alpha_to[_s[u + 1]];
			  else
			    _d[u + 1] = 0;
			    for (i = 1; i <= _l[u + 1]; i++)
			      if ((_s[u + 1 - i] != -1) && (_elp[u + 1][i] != 0))
			        _d[u + 1] ^= alpha_to[(_s[u + 1 - i] 
			                      + index_of[_elp[u + 1][i]]) % n];
			  /* put _d[u+1] into index form */
			  _d[u + 1] = index_of[_d[u + 1]];	
			}
		} while ((u < t2) && (_l[u + 1] <= t));
 
		u++;
		if (_l[u] <= t) {/* Can correct errors */
			/* put elp into index form */
			for (i = 0; i <= _l[u]; i++)
				_elp[u][i] = index_of[_elp[u][i]];


#ifdef _BCH_VERBOSE_
			printf("sigma(x) = ");
			for (i = 0; i <= _l[u]; i++)
				printf("%3d ", _elp[u][i]);
			printf("\n");
			printf("Roots: ");
#endif

			/* Chien search: find roots of the error location polynomial */
			for (i = 1; i <= _l[u]; i++)
				_reg[i] = _elp[u][i];
			count = 0;
			for (i = 1; i <= n; i++) {
				q = 1;
				for (j = 1; j <= _l[u]; j++)
					if (_reg[j] != -1) {
						_reg[j] = (_reg[j] + j) % n;
						q ^= alpha_to[_reg[j]];
					}
				if (!q) {	/* store root and error
						 * location number indices */
					_root[count] = i;
					_loc[count] = n - i;
					count++;
					#ifdef _BCH_VERBOSE_
						printf("%3d ", n - i);
					#endif
				}
			}
			#ifdef _BCH_VERBOSE_
				printf("\n");
			#endif
			if (count == _l[u])	
			{
				/* no. roots = degree of elp hence <= t errors */
					for (i = 0; i < _l[u]; i++)
					{
						if(_loc[i]<BCH_DEFAULT_LENGTH) recd[_loc[i]] ^= 1;
						else too_many_errors = true;
					}
			}
			else	/* elp has degree >t hence cannot solve */
			{
				return(_l[u]); // number of errors
				#ifdef _BCH_VERBOSE_
					printf("Incomplete decoding: errors detected\n");
				#endif
			}
		}
	}
	if(too_many_errors) return(BCH_DEFAULT_T+1);
	else return(syn_error == 0 ? 0 : _l[u]); // number of errors
}


void BCH::encode(int encoded_bits[BCH_DEFAULT_LENGTH], const _64bits orig_n)
{
	assert(k == BCH_DEFAULT_K && length == BCH_DEFAULT_LENGTH);
	int orig_bits[BCH_DEFAULT_K];
	toBitPattern(orig_bits,orig_n,k);
	encode_bch(encoded_bits,orig_bits);

	for (int i = 0; i < k; i++)
		encoded_bits[i + length - k] = orig_bits[i];

	//printf("[x] encoded_bits: ");
	//printBitArray(encoded_bits, BCH_DEFAULT_LENGTH);
	//printf("\n");

}

bool BCH::decode(int &err_n, _64bits &orig_n, const int encoded_bits[BCH_DEFAULT_LENGTH])
{
	assert(k == BCH_DEFAULT_K && length == BCH_DEFAULT_LENGTH);
	int temp_bits[BCH_DEFAULT_LENGTH];
	for(int i=0; i<BCH_DEFAULT_LENGTH; i++) temp_bits[i] = encoded_bits[i];
	err_n = decode_bch(temp_bits);
	if(err_n > t) return(false);
	orig_n = fromBitPattern(&temp_bits[length - k],BCH_DEFAULT_K);
	return(true);
}

void BCH::encode(_64bits &encoded_n, const _64bits orig_n)
{
	assert(k == BCH_DEFAULT_K && length == BCH_DEFAULT_LENGTH);
	int encoded_bits[BCH_DEFAULT_LENGTH];
	encode(encoded_bits,orig_n);
	encoded_n = fromBitPattern(encoded_bits,BCH_DEFAULT_LENGTH);
}

bool BCH::decode(int &err_n, _64bits &orig_n,    const _64bits encoded_n)
{
	assert(k == BCH_DEFAULT_K && length == BCH_DEFAULT_LENGTH);
	int encoded_bits[BCH_DEFAULT_LENGTH];
	toBitPattern(encoded_bits,encoded_n,BCH_DEFAULT_LENGTH);
	return(decode(err_n, orig_n,encoded_bits));
}


}  // namespace ARToolKitPlus
// End of ../../src/extra/BCH.cxx

// Included ../../src/TrackerImpl.cxx
/* ========================================================================
 * PROJECT: ARToolKitPlus
 * ========================================================================
 * This work is based on the original ARToolKit developed by
 *   Hirokazu Kato
 *   Mark Billinghurst
 *   HITLab, University of Washington, Seattle
 * http://www.hitl.washington.edu/artoolkit/
 *
 * Copyright of the derived and new portions of this work
 *     (C) 2006 Graz University of Technology
 *
 * This framework is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This framework is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this framework; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * For further information please contact 
 *   Dieter Schmalstieg
 *   <schmalstieg@icg.tu-graz.ac.at>
 *   Graz University of Technology, 
 *   Institut for Computer Graphics and Vision,
 *   Inffeldgasse 16a, 8010 Graz, Austria.
 * ========================================================================
 ** @author   Daniel Wagner
 *
 * $Id: TrackerImpl.cxx 172 2006-07-25 14:05:47Z daniel $
 * @file
 * ======================================================================== */


//#include <ARToolKitPlus/Tracker.h>


//#if AR_PATT_SIZE_X!=16 || AR_PATT_SIZE_Y!=16
//#pragma message("CAUTION: AR_PATT_SIZE_X or AR_PATT_SIZE_Y are not standard size")
//#endif

//#pragma message ( "Compiling TrackerImpl.cxx" )


namespace ARToolKitPlus {


AR_TEMPL_FUNC 
AR_TEMPL_TRACKER::TrackerImpl()
{
	int i;

#ifdef _USE_GENERIC_TRIGONOMETRIC_
#  ifdef WIN32
#    pragma message(">>> using SinCos LUT")
#  endif
	Fixed28_Init();
#  else
#    ifdef WIN32
#      pragma message(">>> not using SinCos LUT")
#  endif
#endif

	// set default value to RGB888
	//
	pixelFormat = PIXEL_FORMAT_RGB;
	pixelSize = 3;

	binaryMarkerThreshold = -1;

	autoThreshold.enable = false;
	autoThreshold.numRandomRetries = 2;

	wmarker_num = 0;
	prev_num = 0;
	sprev_num[0] = sprev_num[1] = 0;

	marker_infoTWO = NULL;

	pattern_num = -1;
	for(i=0; i<MAX_LOAD_PATTERNS; i++)
		patf[i] = 0;
	evecf = 0;
	evecBWf = 0;

	// we allocate all large data dynamically
	//
	l_imageL = NULL;
	l_imageL_size = 0;

	workL = artkp_Alloc<int>(WORK_SIZE);
	work2L = artkp_Alloc<int>(WORK_SIZE*7);
	wareaL = artkp_Alloc<int>(WORK_SIZE);
	wclipL = artkp_Alloc<int>(WORK_SIZE*4);
	wposL = artkp_Alloc<ARFloat>(WORK_SIZE*2);

	//workL = new int[WORK_SIZE];
	//work2L = new int[WORK_SIZE*7];
	//wareaL = new int[WORK_SIZE];
	//wclipL = new int[WORK_SIZE*4];
	//wposL = new ARFloat[WORK_SIZE*2];


	// set all right side structures to NULL
	l_imageR = NULL;
	workR = NULL;
	work2R = NULL;
	wareaR = NULL;
	wclipR = NULL;
	wposR = NULL;

	//arDebug                 = 0;
	//arImage                 = NULL;
	arFittingMode           = DEFAULT_FITTING_MODE;
	arImageProcMode         = DEFAULT_IMAGE_PROC_MODE;
	
	arCamera                = NULL;
	loadCachedUndist        = false;
	//arParam;
	arImXsize = arImYsize	= 0;
	arTemplateMatchingMode  = DEFAULT_TEMPLATE_MATCHING_MODE;
	arMatchingPCAMode       = DEFAULT_MATCHING_PCA_MODE;
	arImageL                = NULL;
	//arImageR                = NULL;

	markerMode = MARKER_TEMPLATE;

	RGB565_to_LUM8_LUT = NULL;

	relBorderWidth = 0.25f;

	// undistortion addon by Daniel
	//
	undistMode = UNDIST_STD;
	undistO2ITable = NULL;
	//undistI2OTable = NULL;
	arParamObserv2Ideal_func = &AR_TEMPL_TRACKER::arParamObserv2Ideal_std;
	//arParamIdeal2Observ_func = arParamIdeal2Observ_std;

	vignetting.enabled = false;
	vignetting.corners = 
	vignetting.leftright = 
	vignetting.bottomtop = 0;

	bchProcessor = NULL;

	// RPP integration -- [t.pintaric]
	poseEstimator = POSE_ESTIMATOR_ORIGINAL;
	//poseEstimator_func = &AR_TEMPL_TRACKER::arGetTransMat;
	//multiPoseEstimator_func = &AR_TEMPL_TRACKER::arMultiGetTransMat;

	descriptionString = new char[512];

	profiler.reset();
}


AR_TEMPL_FUNC 
AR_TEMPL_TRACKER::~TrackerImpl()
{
	if(arCamera)
		delete arCamera;
	arCamera = NULL;

	if(bchProcessor)
		delete bchProcessor;
	bchProcessor = NULL;

	if(l_imageL)
		artkp_Free(l_imageL);
	l_imageL = NULL;

	if(workL)
		artkp_Free(workL);
	workL = NULL;

	if(work2L)
		artkp_Free(work2L);
	work2L = NULL;

	if(wareaL)
		artkp_Free(wareaL);
	wareaL = NULL;

	if(wclipL)
		artkp_Free(wclipL);
	wclipL = NULL;

	if(wposL)
		artkp_Free(wposL);
	wposL = NULL;

	if(RGB565_to_LUM8_LUT)
		artkp_Free(RGB565_to_LUM8_LUT);
	RGB565_to_LUM8_LUT = NULL;

	if(undistO2ITable)
		artkp_Free(undistO2ITable);
	undistO2ITable = NULL;

	if(descriptionString)
		delete [] descriptionString;
	descriptionString = NULL;
}


AR_TEMPL_FUNC bool
AR_TEMPL_TRACKER::setPixelFormat(PIXEL_FORMAT nFormat)
{
	PIXEL_FORMAT oldFormat = pixelFormat;

	switch(pixelFormat = nFormat)
	{
	case PIXEL_FORMAT_LUM:
		pixelSize=1;
		return true;

	case PIXEL_FORMAT_RGB565:
		pixelSize=2;
		return true;

	case PIXEL_FORMAT_BGR:
	case PIXEL_FORMAT_RGB:
		pixelSize=3;
		return true;

	case PIXEL_FORMAT_ABGR:
	case PIXEL_FORMAT_BGRA:
	case PIXEL_FORMAT_RGBA:
		pixelSize=4;
		return true;

	default:
		pixelFormat = oldFormat;
		return false;
	}
}


AR_TEMPL_FUNC void
AR_TEMPL_TRACKER::checkImageBuffer()
{
	// we have to take care here when using a memory manager that can not free memory
	// (usually this image buffer should only be built once - unless we change camera resolution)
	//

	int newSize = screenWidth*screenHeight;

	if(newSize==l_imageL_size)
		return;

	if(l_imageL)
		//delete l_imageL;
		artkp_Free(l_imageL);

	l_imageL_size = newSize;

	//l_imageL = new ARInt16[newSize];
	l_imageL = artkp_Alloc<ARInt16>(newSize);
}


AR_TEMPL_FUNC bool
AR_TEMPL_TRACKER::checkPixelFormat()
{
	switch(pixelFormat)
	{
	case PIXEL_FORMAT_LUM:
		return pixelSize==1;

	case PIXEL_FORMAT_RGB565:
		return pixelSize==2;

	case PIXEL_FORMAT_BGR:
	case PIXEL_FORMAT_RGB:
		return pixelSize==3;

	case PIXEL_FORMAT_ABGR:
	case PIXEL_FORMAT_BGRA:
	case PIXEL_FORMAT_RGBA:
		return pixelSize==4;

	default:
		return false;
	}
}


AR_TEMPL_FUNC bool
AR_TEMPL_TRACKER::loadCameraFile(const char* nCamParamFile, ARFloat nNearClip, ARFloat nFarClip)
{
	CameraFactory cf;
	Camera* c_ptr = cf.createCamera(nCamParamFile);
	if(c_ptr == NULL)
	{
		if(logger)
			logger->artLog("ARToolKitPlus: Camera parameter load error!\n");
		return false;
	}

	if(arCamera)
		delete arCamera;
	arCamera = NULL;

	setCamera(c_ptr, nNearClip,nFarClip);
	return true;
}


AR_TEMPL_FUNC void
AR_TEMPL_TRACKER::setCamera(Camera* nCamera)
{
	arCamera = nCamera;

	if(arCamera)
	{
		arCamera->changeFrameSize(screenWidth,screenHeight);
		arInitCparam(arCamera);
		arCamera->logSettings(logger);

		// Comment out if you want to get the matrix for camera calibration
		// printf("MAT=\n");
		// printf("[ %f %f %f;\n",arCamera->mat[0][0],arCamera->mat[0][1],arCamera->mat[0][2]);
		// printf("%f %f %f;\n",arCamera->mat[1][0],arCamera->mat[1][1],arCamera->mat[1][2]);
		// printf("%f %f %f ]\n",arCamera->mat[2][0],arCamera->mat[2][1],arCamera->mat[2][2]);

		buildUndistO2ITable(arCamera);
	}
}


AR_TEMPL_FUNC void
AR_TEMPL_TRACKER::setCamera(Camera* nCamera, ARFloat nNearClip, ARFloat nFarClip)
{
	setCamera(nCamera);

	ARParam gCparam = *((ARParam*)arCamera);

	for(int i = 0; i < 4; i++)
		gCparam.mat[1][i] = (gCparam.ysize-1)*(gCparam.mat[2][i]) - gCparam.mat[1][i];
	convertProjectionMatrixToOpenGLStyle(&gCparam, nNearClip,nFarClip, gl_cpara);
}


AR_TEMPL_FUNC ARFloat
AR_TEMPL_TRACKER::calcOpenGLMatrixFromMarker(ARMarkerInfo* nMarkerInfo, ARFloat nPatternCenter[2], ARFloat nPatternSize, ARFloat *nOpenGLMatrix)
{
	ARFloat tmpTrans[3][4];

	ARFloat err = executeSingleMarkerPoseEstimator(nMarkerInfo, nPatternCenter, nPatternSize, tmpTrans);
	convertTransformationMatrixToOpenGLStyle(tmpTrans, nOpenGLMatrix);

	return err;
}


AR_TEMPL_FUNC void
AR_TEMPL_TRACKER::convertTransformationMatrixToOpenGLStyle(ARFloat para[3][4], ARFloat gl_para[16])
{
    int     i, j;

    for( j = 0; j < 3; j++ ) {
        for( i = 0; i < 4; i++ ) {
            gl_para[i*4+j] = para[j][i];
        }
    }
    gl_para[0*4+3] = gl_para[1*4+3] = gl_para[2*4+3] = 0.0;
    gl_para[3*4+3] = 1.0;
}


AR_TEMPL_FUNC bool
AR_TEMPL_TRACKER::convertProjectionMatrixToOpenGLStyle( ARParam *param, ARFloat gnear, ARFloat gfar, ARFloat m[16] )
{
	return convertProjectionMatrixToOpenGLStyle2( param->mat, param->xsize, param->ysize, gnear, gfar, m );
}


AR_TEMPL_FUNC bool
AR_TEMPL_TRACKER::convertProjectionMatrixToOpenGLStyle2( ARFloat cparam[3][4], int width, int height, ARFloat gnear, ARFloat gfar, ARFloat m[16] )
{
    ARFloat   icpara[3][4];
    ARFloat   trans[3][4];
    ARFloat   p[3][3], q[4][4];
    int      i, j;

    if(arParamDecompMat(cparam, icpara, trans) < 0) {
        printf("gConvGLcpara: Parameter error!!\n");
        return false;
    }

    for( i = 0; i < 3; i++ ) {
        for( j = 0; j < 3; j++ ) {
            p[i][j] = icpara[i][j] / icpara[2][2];
        }
    }
    q[0][0] = (2.0f * p[0][0] / width);
    q[0][1] = (2.0f * p[0][1] / width);
    q[0][2] = ((2.0f * p[0][2] / width)  - 1.0f);
    q[0][3] = 0.0f;

    q[1][0] = 0.0f;
    q[1][1] = (2.0f * p[1][1] / height);
    q[1][2] = ((2.0f * p[1][2] / height) - 1.0f);
    q[1][3] = 0.0f;

    q[2][0] = 0.0f;
    q[2][1] = 0.0f;
    q[2][2] = (gfar + gnear)/(gfar - gnear);
    q[2][3] = -2.0f * gfar * gnear / (gfar - gnear);

    q[3][0] = 0.0f;
    q[3][1] = 0.0f;
    q[3][2] = 1.0f;
    q[3][3] = 0.0f;

    for( i = 0; i < 4; i++ ) {
        for( j = 0; j < 3; j++ ) {
            m[i+j*4] = q[i][0] * trans[0][j]
                     + q[i][1] * trans[1][j]
                     + q[i][2] * trans[2][j];
        }
        m[i+3*4] = q[i][0] * trans[0][3]
                 + q[i][1] * trans[1][3]
                 + q[i][2] * trans[2][3]
                 + q[i][3];
    }

	return true;
}


AR_TEMPL_FUNC bool
AR_TEMPL_TRACKER::calcCameraMatrix(const char* nCamParamFile, int nWidth, int nHeight, ARFloat nNear, ARFloat nFar, ARFloat *nMatrix)
{
	CameraFactory cf;
	Camera* pCam = cf.createCamera(nCamParamFile);
	if(pCam == NULL)
	{
		return(false);
	}

	pCam->changeFrameSize(AR_TEMPL_TRACKER::screenWidth,AR_TEMPL_TRACKER::screenHeight);

	int i;
    for(i = 0; i < 4; i++ )
        pCam->mat[1][i] = (pCam->ysize-1)*(pCam->mat[2][i]) - pCam->mat[1][i];

	ARFloat glcpara[16];

    if(!convertProjectionMatrixToOpenGLStyle((ARParam*)pCam, nNear,nFar, glcpara))
		return false;

	// convert to float (in case of ARFloat is def'ed to doubled
	for(i = 0; i < 16; i++ )
		nMatrix[i] = (ARFloat)glcpara[i];

	return(true);
}


AR_TEMPL_FUNC void
AR_TEMPL_TRACKER::changeCameraSize(int nWidth, int nHeight)
{
	screenWidth = nWidth;
	screenHeight = nHeight;

	arCamera->changeFrameSize(nWidth,nHeight);
	arInitCparam(arCamera);

	if(logger)
		logger->artLogEx("ARToolKitPlus: Changed CamSize %d, %d", arCamera->xsize, arCamera->ysize);
}


AR_TEMPL_FUNC void
AR_TEMPL_TRACKER::setUndistortionMode(UNDIST_MODE nMode)
{
	undistMode = nMode;
	switch(undistMode)
	{
	case UNDIST_NONE:
		arParamObserv2Ideal_func = &AR_TEMPL_TRACKER::arParamObserv2Ideal_none;
		//arParamIdeal2Observ_func = arParamIdeal2Observ_none;
		break;

	case UNDIST_STD:
		arParamObserv2Ideal_func = &AR_TEMPL_TRACKER::arParamObserv2Ideal_std;
		//arParamIdeal2Observ_func = arParamIdeal2Observ_std;
		break;

	case UNDIST_LUT:
		arParamObserv2Ideal_func = &AR_TEMPL_TRACKER::arParamObserv2Ideal_LUT;
		//arParamIdeal2Observ_func = arParamIdeal2Observ_LUT;
		break;
	}
}


AR_TEMPL_FUNC bool
AR_TEMPL_TRACKER::setPoseEstimator(POSE_ESTIMATOR nMode)
{
	poseEstimator = nMode;
/*	switch(poseEstimator)
	{
	case POSE_ESTIMATOR_ORIGINAL:
		poseEstimator_func = &AR_TEMPL_TRACKER::arGetTransMat;
		multiPoseEstimator_func = &AR_TEMPL_TRACKER::arMultiGetTransMat;		// will use arGetTransMat internally
		return true;

	case POSE_ESTIMATOR_ORIGINAL_CONT:
		poseEstimator_func = &AR_TEMPL_TRACKER::arGetTransMatCont2;
		multiPoseEstimator_func = &AR_TEMPL_TRACKER::arMultiGetTransMat;		// will use arGetTransMatCont2 internally
		return true;

	case POSE_ESTIMATOR_RPP:
		if(rppSupportAvailabe())
		{
			poseEstimator_func = &AR_TEMPL_TRACKER::rppGetTransMat;
			multiPoseEstimator_func = &AR_TEMPL_TRACKER::rppMultiGetTransMat;
			return true;
		}

		if(logger)
			logger->artLog("ARToolKitPlus: Failed to set RPP pose estimator - RPP disabled during build\n");
		return false;
	}

	return false;*/

	return true;
}


AR_TEMPL_FUNC ARFloat
AR_TEMPL_TRACKER::executeSingleMarkerPoseEstimator(ARMarkerInfo *marker_info, ARFloat center[2], ARFloat width, ARFloat conv[3][4])
{
	switch(poseEstimator)
	{
	case POSE_ESTIMATOR_ORIGINAL:
		return arGetTransMat(marker_info, center, width, conv);

	case POSE_ESTIMATOR_ORIGINAL_CONT:
		return arGetTransMatCont2(marker_info, center, width, conv);

	case POSE_ESTIMATOR_RPP:
		if(rppSupportAvailabe())
		{
			return rppGetTransMat(marker_info, center, width, conv);
		}
		if(logger)
			logger->artLog("ARToolKitPlus: Failed to set RPP pose estimator - RPP disabled during build\n");
		return -1.0f;
	}

	return -1.0f;
}


AR_TEMPL_FUNC ARFloat
AR_TEMPL_TRACKER::executeMultiMarkerPoseEstimator(ARMarkerInfo *marker_info, int marker_num, ARMultiMarkerInfoT *config)
{
	switch(poseEstimator)
	{
	case POSE_ESTIMATOR_ORIGINAL:
		return arMultiGetTransMat(marker_info, marker_num, config);

	case POSE_ESTIMATOR_ORIGINAL_CONT:
		return arMultiGetTransMat(marker_info, marker_num, config);

	case POSE_ESTIMATOR_RPP:
		if(rppSupportAvailabe())
		{
			return rppMultiGetTransMat(marker_info, marker_num, config);
		}
		if(logger)
			logger->artLog("ARToolKitPlus: Failed to set RPP pose estimator - RPP disabled during build\n");
		return -1.0f;
	}

	return -1.0f;
}



AR_TEMPL_FUNC void
AR_TEMPL_TRACKER::activateVignettingCompensation(bool nEnable, int nCorners, int nLeftRight, int nTopBottom)
{
	vignetting.enabled = nEnable;
	vignetting.corners = nCorners;
	vignetting.leftright = nLeftRight;
	vignetting.bottomtop = nTopBottom;
}


AR_TEMPL_FUNC void
AR_TEMPL_TRACKER::setMarkerMode(MARKER_MODE nMarkerMode)
{
	markerMode = nMarkerMode;
}


static void
convertPixel16To24(unsigned short nPixel, unsigned char& nRed, unsigned char& nGreen, unsigned char& nBlue)
{
	const unsigned short RED_MASK    = 0x1F << 11;
	const unsigned short GREEN_MASK  = 0x3F << 5;
	const unsigned short BLUE_MASK   = 0x1F;

	nRed =   (unsigned char)((nPixel&RED_MASK) >> 8);
	nGreen = (unsigned char)((nPixel&GREEN_MASK) >> 3);
	nBlue =  (unsigned char)((nPixel&BLUE_MASK) << 3);
}


AR_TEMPL_FUNC void
AR_TEMPL_TRACKER::checkRGB565LUT()
{
	int i;

	if(RGB565_to_LUM8_LUT)
		return;

	//RGB565_to_LUM8_LUT = new unsigned char[LUM_TABLE_SIZE];
	RGB565_to_LUM8_LUT = artkp_Alloc<unsigned char>(LUM_TABLE_SIZE);


#ifdef SMALL_LUM8_TABLE
	for(i=0; i<(LUM_TABLE_SIZE >> 6); i++)
	{
		unsigned char red,green,blue;

		convertPixel16To24(i<<6, red, green, blue);
		RGB565_to_LUM8_LUT[i] =  (unsigned char)(((red<<1) + (green<<2) + green + blue)>>3);
	}
#else
	for(i=0; i<LUM_TABLE_SIZE; i++)
	{
		unsigned char red,green,blue;

		convertPixel16To24(i, red, green, blue);
		RGB565_to_LUM8_LUT[i] =  (unsigned char)(((red<<1) + (green<<2) + green + blue)>>3);
	}
#endif //SMALL_LUM8_TABLE
}


// cleanup function called when program exits
//
AR_TEMPL_FUNC void
AR_TEMPL_TRACKER::cleanup()
{
	if(marker_infoTWO)
		//delete [] marker_infoTWO;
		artkp_Free(marker_infoTWO);
	marker_infoTWO = NULL;
}


AR_TEMPL_FUNC const char*
AR_TEMPL_TRACKER::getDescription()
{
	char* pixelformats[] = { "NONE", "ABGR", "BGRA", "BGR", "RGBA", "RGB", "RGB565", "LUM"  };
	int f = getPixelFormat();

	char *compilerstr = new char[256];

#ifdef __INTEL_COMPILER
	sprintf(compilerstr, "Intel C++ v%d.%d", __INTEL_COMPILER/100, __INTEL_COMPILER%100);
#   pragma message ( ">> ARToolKitPlus: compiling with Intel Compiler" )
#elif _MSC_VER
	sprintf(compilerstr, "MS C++ v%d.%d", _MSC_VER/100, _MSC_VER%100);
//#   pragma message ( ">> ARToolKitPlus: compiling with Microsoft Compiler" )
#elif __GNUC__
	sprintf(compilerstr, "GCC %d.%d.%d", __GNUC__, __GNUC_MINOR__,  __GNUC_PATCHLEVEL__);
#else
	sprintf(compilerstr, "Unknown Compiler");
#endif

	assert(strlen(compilerstr)<256);

//
// Under WinCE it is very important to use the Intel XScale compiler.
// This will double the speed of ARToolKit
//
#if defined(_WIN32_WCE) && defined(NDEBUG)

#  ifdef __INTEL_COMPILER
#    pragma message ( ">>>    WinCE Targeted with Intel Compiler" )
#  else
#    pragma message("  ")
#    pragma message("  >>>    WinCE PERFORMANCE WARNING: Release builds of ARToolKitPlus should be done with the Intel XScale Compiler !!!")
#    pragma message("  ")
#  endif

//#  ifndef _USEGPP_
//#    pragma message("  ")
//#    pragma message("  >>>    PERFORMANCE WARNING: Release builds of ARToolKitPlus should be done with the Intel GPP !!!")
//#    pragma message("  ")
//#  else
//#   pragma message ( ">> ARToolKitPlus: compiling with Intel GPP" )
//#  endif

#endif //defined(_WIN32_WCE) && defined(NDEBUG)

	sprintf(descriptionString,
			"ARToolKitPlus v%d.%d: built %s %s (%s); %s; %s precision; %dx%d marker; %s pixelformat; %scustom memory manager; RPP support %savailable.",
			VERSION_MAJOR, VERSION_MINOR,
			__DATE__, __TIME__,
			compilerstr,
#if defined(_USEFIXED_) || defined(_USEGPP_)
			"fixed-point",
#else
			"floating-point",
#endif
			usesSinglePrecision() ? "single" : "double",
			PATTERN_WIDTH,PATTERN_HEIGHT,
			f<=PIXEL_FORMAT_LUM ? pixelformats[f] : pixelformats[0],
#ifdef _ARTKP_NO_MEMORYMANAGER_
			"no ",
#else
			"",
#endif
			rppSupportAvailabe() ? "" : "not "
			);

	delete [] compilerstr;

	assert(strlen(descriptionString)<512);
	return descriptionString;
}


static bool usesSinglePrecision()
{
	return sizeof(ARFloat)==4;
}


AR_TEMPL_FUNC void*
AR_TEMPL_TRACKER::operator new(size_t size)
{
#ifndef _ARTKP_NO_MEMORYMANAGER_
	if(memManager)
		return memManager->getMemory(size);
	else
#endif //_ARTKP_NO_MEMORYMANAGER_
		return malloc(size);
}


AR_TEMPL_FUNC void
AR_TEMPL_TRACKER::operator delete(void *rawMemory)
{
	if(!rawMemory)
		return;

#ifndef _ARTKP_NO_MEMORYMANAGER_
	if(memManager)
		memManager->releaseMemory(rawMemory);
	else
#endif //_ARTKP_NO_MEMORYMANAGER_
		free(rawMemory);
}


AR_TEMPL_FUNC size_t
AR_TEMPL_TRACKER::getDynamicMemoryRequirements()
{
	// requirements for allocations in the constructor
	//
	size_t size = sizeof(unsigned int)*(WORK_SIZE +			// workL = new int[WORK_SIZE];
										WORK_SIZE*7 +		// work2L = new int[WORK_SIZE*7];
										WORK_SIZE +			// wareaL = new int[WORK_SIZE];
										WORK_SIZE*4 +		// wclipL = new int[WORK_SIZE*4];
										WORK_SIZE*2);		// wposL = new ARFloat[WORK_SIZE*2];

	// requirements for the image buffer (arImageL)
	//
	size += sizeof(ARUint8)*MAX_BUFFER_WIDTH*MAX_BUFFER_HEIGHT;


	// requirements for allocation of marker_infoTWO
	//
	size += sizeof(ARMarkerInfo2)*MAX_IMAGE_PATTERNS;


	// requirements for allocation of l_imageL
	//
	size += sizeof(ARInt16)*MAX_BUFFER_WIDTH*MAX_BUFFER_HEIGHT;


	// requirements for the lens undistortion table (undistO2ITable)
	//
	size += sizeof(unsigned int)*MAX_BUFFER_WIDTH*MAX_BUFFER_HEIGHT;


	// requirements for the RGB565 to gray table RGB565_to_LUM8_LUT
	//
	size += sizeof(unsigned char)*LUM_TABLE_SIZE;


	return size;
}


}  // namespace ARToolKitPlus
// End of ../../src/TrackerImpl.cxx
//#include "../../src/extra/harrisCornerDetector.cxx"
//#include "../../src/extra/cornerRefinement.cxx"


#endif //__ARTOOLKIT_TRACKERIMPL_HEADERFILE__
